/* (C) 2006 Richard Mills / The Linux Academy Ltd.
 *
 * Provide authentication settings and methods.
 *
 * $Id: AuthSettings.cc,v 1.4 2007-04-17 13:44:04 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <iostream>
#include <sstream>
#include <iomanip>
#include "AuthSettings.hh"
#include "Utils.hh"
#include "AuthDigest.hh"


namespace HTTPTunnel {


// Construction/destruction
AuthSettings::AuthSettings(AuthScheme authScheme, std::string realm)
: m_authScheme(NONE), m_user(""), m_passwd(""), m_realm(), m_credentials(),
  m_digestNonce(), m_digestQop(), m_digestCNonce(), m_digestCNonceCount(1)
{
	// Setup authentication
	setAuthScheme(authScheme, realm);
}

AuthSettings::AuthSettings(AuthScheme authScheme, std::string user,
	std::string passwd)
: m_authScheme(NONE), m_user(user), m_passwd(passwd), m_realm(), m_credentials(),
  m_digestNonce(), m_digestQop(), m_digestCNonce(), m_digestCNonceCount(1)
{
	// Setup authentication
	setAuthScheme(authScheme);
}

AuthSettings::AuthSettings(const AuthSettings& that)
: m_authScheme(that.m_authScheme), m_user(that.m_user), m_passwd(that.m_passwd),
  m_realm(that.m_realm), m_credentials(that.m_credentials),
  m_digestNonce(that.m_digestNonce), m_digestQop(that.m_digestQop),
  m_digestCNonce(that.m_digestCNonce), m_digestCNonceCount(that.m_digestCNonceCount)
{
}

AuthSettings::~AuthSettings()
{
}


// Assignment
AuthSettings& AuthSettings::operator=(const AuthSettings& that)
{
	// Self assignment
	if(this != &that) {
		m_authScheme = that.m_authScheme;
		m_user = that.m_user;
		m_passwd = that.m_passwd;
		m_realm = that.m_realm;
		m_credentials = that.m_credentials;
		m_digestNonce = that.m_digestNonce;
		m_digestQop = that.m_digestQop;
		m_digestCNonce = that.m_digestCNonce;
		m_digestCNonceCount = that.m_digestCNonceCount;
	}

	return *this;
}


/*************/
/* Utilities */
/*************/


// Construct tunneling request/response headers
std::string AuthSettings::getRequestHeader(std::string uri)
{
	std::stringstream res;

	// Using authorization?
	if(m_authScheme == BASIC) {
		// Set basic authentication fields
		res << getAuthReqestHeaderID() << ": Basic " << m_credentials << "\r\n";
	}
	else if(m_authScheme == DIGEST) {
		// Get client nonce count
		std::stringstream cnoncec;
		cnoncec << std::setw(8) << std::setfill('0') << m_digestCNonceCount++;

		// Calculate disgest repsonse value
		std::string response = getDigestResponse("MD5", m_user, m_passwd, m_realm,
			m_digestNonce, m_digestCNonce, cnoncec.str(), m_digestQop, uri);

		// Set digest authentication fields
		res << getAuthReqestHeaderID() << ": Digest ";
		res << "username=\"" << m_user << "\"";
		res << ", realm=\"" << m_realm << "\"";
		res << ", nonce=\"" << m_digestNonce << "\"";
		res << ", uri=\"" << uri << "\"";
		res << ", algorithm=\"MD5\"";
		res << ", qop=\"" << m_digestQop << "\"";
		res << ", cnonce=\"" << m_digestCNonce << "\"";
		res << ", nc=" << cnoncec.str();
		res << ", response=\"" << response << "\"";
		res << "\r\n";
	}

	return res.str();
}

// Construct tunneling request/response headers
std::string AuthSettings::getResponseHeader(std::string uri)
{
	return "";
}

// Calculate the response fields for digest authentication
std::string AuthSettings::getDigestResponse(std::string alg, std::string user,
	std::string passwd, std::string realm, std::string nonce, std::string cnonce,
	std::string cnoncec, std::string qop, std::string uri)
{
	// Construct digest HA1
	MD5Digest::HASHHEX HA1;
	MD5Digest::HASHHEX HA2 = "";
	MD5Digest::HASHHEX HHResponse;

	MD5Digest::DigestCalcHA1(alg.c_str(), user.c_str(), realm.c_str(),
		passwd.c_str(), nonce.c_str(), cnonce.c_str(), HA1);
	MD5Digest::DigestCalcResponse(HA1, nonce.c_str(), cnoncec.c_str(),
		cnonce.c_str(), qop.c_str(), "GET", uri.c_str(), HA2, HHResponse);

	return HHResponse;
}


/*************/
/* Accessors */
/*************/


// Get authentication scheme in use
AuthSettings::AuthScheme AuthSettings::getAuthScheme(void) const
{
	return m_authScheme;
}

// Do we have a user/passwd specified?
bool AuthSettings::gotUserPasswd(void) const
{
	return (m_user.size() > 0 && m_passwd.size() > 0);
}

// Get proxy authentication user
std::string AuthSettings::getUser(void) const
{
	return m_user;
}

// Get proxy authentication passwd
std::string AuthSettings::getPasswd(void) const
{
	return m_passwd;
}


/************/
/* Mutators */
/************/


// Set the authentication scheme
void AuthSettings::setAuthScheme(AuthScheme scheme)
{
	setAuthScheme(scheme, m_realm, m_digestQop, m_digestNonce);
}

void AuthSettings::setAuthScheme(AuthScheme scheme, std::string realm)
{
	setAuthScheme(scheme, realm, m_digestQop, m_digestNonce);
}

void AuthSettings::setAuthScheme(AuthScheme scheme, std::string realm, std::string qop)
{
	setAuthScheme(scheme, realm, qop, m_digestNonce);
}

void AuthSettings::setAuthScheme(AuthScheme scheme, std::string realm, std::string qop,
	std::string nonce)
{
	m_authScheme = scheme;
	m_realm = realm;

	// What authentication scheme?
	if(m_authScheme == BASIC) {
		// Create basic credentials
		m_credentials = Utils::Base64Encode(m_user + ":" + m_passwd);
	}
	else if(m_authScheme == DIGEST) {
		// Set digest fields
		m_digestQop = qop;
		m_digestNonce = nonce;

		// Create initial digest CNonce
		m_digestCNonceCount = 1;
		std::stringstream cnonce;
		cnonce << std::setw(8) << std::hex << std::setfill('0') << rand();
		m_digestCNonce = cnonce.str();
	}
}


// Set user name & password
void AuthSettings::setUserPasswd(std::string user, std::string passwd)
{
	// Save user & passwd
	m_user = user;
	m_passwd = passwd;
}


// Generate a new random nonce
void AuthSettings::newNonce(void)
{
	std::stringstream nonce;
	nonce << std::setw(8) << std::hex << std::setfill('0') << rand();
	m_digestNonce = nonce.str();
}


}	// namespace

