/* (C) 2006 Richard Mills / The Linux Academy Ltd.
 *
 * Client side Tunnel.
 *
 * $Id: ClientTunnel.cc,v 1.13 2007-08-04 11:21:43 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include "ClientTunnel.hh"
#include "Logger.hh"
#include "Debug.hh"
#include "MutexVar.hh"
#include "HTTPMsg.hh"
#include "TunnelConfig.hh"
#include "Authenticator.hh"
#include "ServerAuthSettings.hh"
#include "GlobalConfig.hh"
#include "ClientSideConfig.hh"
#include "Connection.hh"


// Global references
extern HTTPTunnel::Logger* g_pLogger;
extern HTTPTunnel::GlobalConfig* g_pConfig;


namespace HTTPTunnel {


// Construction/destruction
ClientTunnel::ClientTunnel(TunnelConfig* tcfg, Tunneler* tunneler, TunnelState state)
: Tunnel(tcfg, tunneler, RIGHT, state), m_noActiveTunnels(0), m_httpConnStable(false),
  m_lastRecvAckSeqNo(1)
{
	DPRINTI("ClientTunnel::ClientTunnel()");

	// Use more lenient ack timeout values in client
	m_minAckTimeout = (time_t)(m_minAckTimeout * 1.5);
	m_maxAckTimeout = (time_t)(m_maxAckTimeout * 1.5);
	m_ackTimeout = m_maxAckTimeout;
}

ClientTunnel::~ClientTunnel()
{
	#ifdef DEBUG_TUN
	std::stringstream str;
	str << "ClientTunnel::~ClientTunnel() - " << m_tid;
	DPRINTI(str.str());
	#endif
}


/************/
/* Mutators */
/************/


// Inform us of how many tunnels in total are currently up
void ClientTunnel::tunnelsCount(unsigned int tot)
{
	Threads::MutexVar mv(m_mtx);

	// Lengthen poll timeout if no. of tunnels grows large
	m_noActiveTunnels = tot;
	scalePollTimeout();
}


/********************/
/* Public Interface */
/********************/


// Process some newly arrived tunnel msg from one of our connections
void ClientTunnel::processTunnelMsg(const HTTPMsg* msg)
{
	DPRINTV("ClientTunnel::processTunnelMsg() - " + m_tid.toString());

	Threads::MutexVar mv(m_mtx);

	// Check whether HTTP connection is clear
	const HTTPMsgParms* parms = msg->getParms();
	m_httpConnStable = (parms->getRespCode() == 200);

	// Update our quotas?
	if(parms->getState() == AUTHREQ) {
		Authenticator* authenticator = getAuthenticator();
		authenticator->setUserQuotas(m_user, parms->getQuotas());
	}

	// Base class
	Tunnel::processTunnelMsg(msg);
}


// Set whether the tunnel user is authenticated
void ClientTunnel::setUserAuthenticated(bool auth)
{
	Threads::MutexVar mv(m_mtx);
	Authenticator* authenticator = getAuthenticator();
	authenticator->setUserAuthenticated(m_user, auth, this);
}


/***********************/
/* Protected Interface */
/***********************/


// Create complete HTTP msg parameters for sending down tunnel - Caller hold mutex
HTTPMsgParms* ClientTunnel::makeTunnelMsg(const HTTPMsgParms* recvd,
	unsigned int maxPayload, bool retry) throw(MemoryError)
{
	// If we have to use header based tunneling, then limit the payload size
	if(!m_tunCfg->getUseHeaders())
		maxPayload = MAX_URL_HTTP_PAYLOAD;

	// Get basic parameters from base class
	HTTPMsgParms* parms = Tunnel::makeTunnelMsg(recvd, maxPayload, retry);

	// Don't keep hammering away with RECONNECT attempts, scale back poll period
	if(m_tunnelState == RECONNECT) {
		m_pollTimeout = MAX_POLL_TIMEOUT;
	}

	return parms;
}


// Successfully sent a new message down the tunnel - Caller hold mutex
void ClientTunnel::sentNewTunnelMsg(HTTPMsgParms* parms)
{
	// We always want to remember in the client
	Tunnel::sentNewTunnelMsg(parms, true);
}


// Scale poll timeout - Caller hold mutex
void ClientTunnel::scalePollTimeout(void)
{
	#ifdef DEBUG_TUN
	unsigned int oldPollTimeout = m_pollTimeout;
	#endif

	// Take account of ack timeout
	float ackScale = (float)(m_ackTimeout - m_minAckTimeout) /
		(float)(m_maxAckTimeout - m_minAckTimeout);

	// Take account of number of currently active tunnels
	#ifndef PLATFORM_OSX
	float atScale = (float)(m_noActiveTunnels / TUNNEL_COUNT_POLL_THRESHOLD) *
		TUNNEL_COUNT_POLL_STEP;
	#else
	float atScale = (float)(m_noActiveTunnels / TUNNEL_COUNT_POLL_THRESHOLD) *
		0.2;
	#endif
	if(atScale > 1)
		atScale = 1;

	// Combine all scaling factors
	float scale = ackScale + atScale;
	if(scale > 1)
		scale = 1;

	// Scale poll timeout
	float range = (MAX_POLL_TIMEOUT - MIN_POLL_TIMEOUT) * scale;
	m_pollTimeout = (int)(MIN_POLL_TIMEOUT + range);

	#ifdef DEBUG_TUN
	if(oldPollTimeout != m_pollTimeout) {
		std::stringstream nfo;
		nfo << "ClientTunnel::scalePollTimeout() - Now " << m_pollTimeout;
		nfo << " for tunnel " << getTidStr();
		DPRINTV(nfo.str());
	}
	#endif
}


// Have we got a stable HTTP connection to other end?
bool ClientTunnel::httpStable(void) const
{
	return m_httpConnStable;
}


// Can we send a new tunnel msg? - Caller holds mutex
bool ClientTunnel::clearToSendNewMsg(Connection* conn)
{
	bool sendNew = false;

	try {
		// Check for unacked management msgs and consider resending them early
		checkUnackedMgmtMsgs();

		// New message or resend of old message?
		sendNew = !resendUnackedMsgs(conn);
		if(sendNew) {
			// How many outstanding unacked msgs do we already have?
			if(m_unackedMsgQueue.size() > MAX_UNACKED_MSG) {
				// If we're in a special connection state, go ahead with send anyway
				if(m_tunnelState == CONNECTED || m_tunnelState == CLOSING) {
					DPRINTI("Too many unacked msgs on tunnel, not sending another");
					sendNew = false;
				}
			}
		}
	}
	catch(ResendError& e) {
		g_pLogger->warn("Error resending msg: " + e.what());
	}

	return sendNew;
}


// Determine whether an ack implies we should resend a msg - Caller holds mutex
void ClientTunnel::checkForResendRequest(const SeqNo& ackNo)
{
	// Is the sender requesting we resend an unacked msg?
	if(ackNo == m_lastRecvAckSeqNo) {
		// Is the msg still unacked?
		std::vector<SentMsgParms*>::iterator pit = std::find_if(m_unackedMsgQueue.begin(),
			m_unackedMsgQueue.end(), CmpSentParmsSeqNo(ackNo));
		if(pit != m_unackedMsgQueue.end()) {
			// Should we consider resending this msg early?
			SentMsgParms* sentMsgParms = *pit;
			sentMsgParms->m_ackCount++;
			if(sentMsgParms->m_ackCount >= (MAX_UNACKED_MSG / 2)) {
				DPRINTV("ClientTunnel::checkForResendRequest() - Looks like resend request");
				sentMsgParms->m_ackCount = 0;
				needToResendMsg(ackNo);
			}
		}
	}

	// Remember for next time
	m_lastRecvAckSeqNo = ackNo;
}


// Check whether any connections need to be blocked/unblocked - Caller holds mutex
void ClientTunnel::checkBlockedConnections(void)
{
	// Check raw connections
	std::vector<Tunnel::Selector*> filt;
	filt.push_back(new SelLeftConn(this));
	std::vector<Connection*> conns;
	getConnsFromSel(filt, conns);

	// Do we want to block or unblock?
	bool unblock = false;
	bool block = (m_rawDataQueueSize >= MAX_RAW_DATA_FILL);
	if(!block) {
		// Only unblock if we're well below full
		unsigned int buffFill = (unsigned int)(((float)m_rawDataQueueSize /
			(float)MAX_RAW_DATA_FILL) * 100);
		unblock = (buffFill <= RAW_DATA_UNBLOCK_LEVEL);
	}

	// Check each connection
	std::vector<Connection*>::iterator cit;
	for(cit = conns.begin(); cit != conns.end(); ++cit) {
		// Get connection
		Connection* conn = *cit;

		// Need to change blocked state?
		if(block)
			conn->setBlocked(true);
		else if(unblock)
			conn->setBlocked(false);
	}
}


}	// namespace

