/* (C) 2005 Richard Mills / The Linux Academy Ltd.
 *
 * Tunnel controller for client end of tunnel.
 *
 * $Id: ClientTunneler.cc,v 1.75 2007-07-08 13:51:36 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <vector>
#include <algorithm>
#include <iostream>
#include <sstream>
#include <memory>
#include "ClientTunneler.hh"
#include "Enums.hh"
#include "Debug.hh"
#include "Connection.hh"
#include "RawConnection.hh"
#include "HTTPConnection.hh"
#include "GlobalConfig.hh"
#include "ClientSideConfig.hh"
#include "ProxySettings.hh"
#include "ServerAuthSettings.hh"
#include "TunnelID.hh"
#include "Exceptions.hh"
#include "Logger.hh"
#include "ClientTunnel.hh"
#include "MutexVar.hh"
#include "HTTPMsg.hh"
#include "HTTPResponseMsg.hh"
#include "ShutdownTunnelEvent.hh"
#include "HTTPMsgParms.hh"
#include "CheckAckEvent.hh"
#include "PollTimeoutEvent.hh"
#include "ConnectionErrorEvent.hh"
#include "SendMsgEvent.hh"
#include "Utils.hh"
#include "Authenticator.hh"
#include "ConnectionManager.hh"
#include "Server.hh"
#include "Sock.hh"


// Global references
extern HTTPTunnel::Logger* g_pLogger;
extern HTTPTunnel::GlobalConfig* g_pConfig;


namespace HTTPTunnel {


// Construction/destruction
ClientTunneler::ClientTunneler(Server* server)
: Tunneler(server), m_serverTestConnID(0)
{
	DPRINTV("ClientTunneler::ClientTunneler()");

	// Test connection to server
	//launchTestServerConnection();
}

ClientTunneler::~ClientTunneler()
{
	DPRINTV("ClientTunneler::~ClientTunneler()");

	// Shutdown event scheduling
	if(!shuttingDown())
		shutdownSchedulers();

	// Clear up server test connection
	destroyTestServerConnection();
}


/**********/
/* Events */
/**********/


// Notification of a new socket connection
void ClientTunneler::newConnectionArrived(Sock* insock)
	throw(CommsError)
{
	DPRINTI("ClientTunneler::newConnectionArrived()");

	///////////////////////////////////////////////////////////////
	// In ClientTunneler a new connection always means a new tunnel
	///////////////////////////////////////////////////////////////

	// Do our users quotas allow another tunnel?
	Authenticator* authenticator = m_server->getConfigMutable()->getAuthenticator();
	std::string userName = authenticator->getDefaultUserName();
	bool userQuotaOk = authenticator->getUserCanCreateTunnel(userName);
	if(userQuotaOk) {
		// Create new tunnel config
		TunnelConfig* tcfg = createTunnelConfig();

		// Lock connection details
		Threads::MutexVar cmv(m_connectionsMtx);

		// Create raw server connection to handle incoming traffic
		Connection* rawconn = new RawConnection(insock, this, *tcfg->getTunnelID());

		// Create HTTP client connection to handle tunneled traffic
		HTTPConnection* httpconn = new HTTPConnection(this);

		// Save connection details
		addConnection(rawconn);
		addConnection(httpconn);

		// Create tunnel
		Tunnel* tun = new ClientTunnel(tcfg, this, Tunnel::CONNECT);
		tun->addLeftConnection(rawconn);
		tun->addRightConnection(httpconn);

		// Schedule connect out on HTTP side
		const ClientSideConfig* clientCfg = getClientSideConfig();
		m_connectionMgr->addProxyConnect(httpconn, 0, clientCfg->getNextHopIP(),
			clientCfg->getNextHopPort(), Enums::TCP, clientCfg->getSSL(),
			clientCfg->getCertificate());

		cmv.unlock();

		// Store Tunnel object in our list of active tunnels
		Threads::MutexVar tmv(m_tunnelsMtx);
		m_tunnels.push_back(tun);
		noTunnelsChanged();
		tmv.unlock();
	}
	else {
		// Tunnel creation quota exceeded
		std::stringstream err;
		err << "Tunnel creation quota exceeded for user " << userName;
		throw CommsError(err.str());
	}
}


// Notification that a tunnel has insufficient left/right connections to be viable
void ClientTunneler::unviableTunnel(const TunnelID& tid)
{
	DPRINTI("ClientTunneler::unviableTunnel() - " + tid.toString());

	// Lookup tunnel
	Threads::MutexVar cmv(m_connectionsMtx);
	Threads::MutexVar tmv(m_tunnelsMtx);
	Tunnel* tun = getTunFromPred(PredTunID(&tid));
	if(tun) {
		// Get number of remaining left side connections for tunnel
		unsigned int left = tun->noLeftConnections();
		if(left == 0) {
			/* In the client, no remaining left connections means we must
			 * abandon this tunnel and close it down.
			 */

			#ifdef DEBUG_EVENTS
			std::stringstream nfo;
			nfo << "Scheduling ShutdownTunnelEvent for tunnel id ";
			nfo << tid.toString();
			nfo << " because client tunneler has no remaining left connections";
			g_pLogger->info("##########");
			g_pLogger->info(nfo.str());
			g_pLogger->info("##########");
			#endif

			// Schedule tunnel shutdown
			DPRINTI(tid.toString() + " - No connections, not preserving");
			scheduleEvent(new ShutdownTunnelEvent(0, this, tid, false));
		}

		// Get number of remaining right side connections for tunnel
		unsigned int right = tun->noRightConnections();
		if(right == 0) {
			/* In the client, no remaining right connections means that if
			 * we want to try and preserve this tunnel we should try and
			 * re-establish a connection to the server.
			 */

			// Do we want to try and preserve this tunnel?
			if(!tun->errorLevelExceeded(Tunnel::RIGHT)) {
				DPRINTI(tid.toString() + " - Trying to preserve");

				// Delay preservation if tunnel wasn't already connected
				Tunnel::TunnelState oldTunState = tun->getState();
				unsigned int preserveDelay = 0;
				if(oldTunState != Tunnel::CONNECTED) {
					DPRINTV("Delaying preserve connect");
					preserveDelay = PRESERVE_RECONNECT_DELAY;
				}

				// Let tunnel know
				tun->reconnecting();

				// Create replacement HTTP client connection to server
				HTTPConnection* httpconn = new HTTPConnection(this);

				// Save connection details
				addConnection(httpconn);
				tun->addRightConnection(httpconn);

				// Schedule initial connect out HTTP side
				const ClientSideConfig* clientCfg = getClientSideConfig();
				m_connectionMgr->addProxyConnect(httpconn, Time(0,preserveDelay),
					clientCfg->getNextHopIP(), clientCfg->getNextHopPort(),
					Enums::TCP, clientCfg->getSSL(), clientCfg->getCertificate());
			}
			else {
				// Schedule tunnel shutdown
				DPRINTI(tid.toString() + " - Too many errors, not preserving");

				#ifdef DEBUG_EVENTS
				std::stringstream nfo;
				nfo << "Scheduling ShutdownTunnelEvent for tunnel id ";
				nfo << tid.toString();
				nfo << " because client tunneler has too many errors on tunnel";
				g_pLogger->info("##########");
				g_pLogger->info(nfo.str());
				g_pLogger->info("##########");
				#endif

				scheduleEvent(new ShutdownTunnelEvent(0, this, tid, false));
			}
		}
	}
	else {
		std::stringstream err;
		err << "ClientTunneler::unviableTunnel() - Unknown tunnel " << tid;
		g_pLogger->warn(err.str());
	}
}


// Notification of new HTTP message arrival
void ClientTunneler::messageArrived(unsigned int connID, const HTTPMsg* msg)
	throw(LookupManagementError)
{
	#ifdef DEBUG_TUN
	std::stringstream nfo;
	nfo << "ClientTunneler::messageArrived() - On connection " << connID;
	DPRINTV(nfo.str());
	#endif

	bool fatalConnectionError = false;

	// Check for HTTP error messages
	const HTTPResponseMsg* respMsg = dynamic_cast<const HTTPResponseMsg*>(msg);
	unsigned short respCode = respMsg->getResponseCode();
	if(respCode == 407 || respCode == 401) {
		// Authentication required
		authenticationFailed(connID, respMsg, respCode);
	}
	else {
		// Standard tunneling msg
		try {
			// Find out which Tunnel this message is from
			const HTTPMsgParms* parms = msg->getParms();
			TunnelID tid(parms->getTid());
			Threads::MutexVar tmv(m_tunnelsMtx);
			ClientTunnel* tun = dynamic_cast<ClientTunnel*>(getTunFromPred(PredTunID(&tid)));

			// Have we found the tunnel for this msg?
			if(tun) {
				// If response code is normal, we know we are authenticated
				if(respCode == 200) {
					tun->setUserAuthenticated(true);
				}

				// Handle any received tunnel payload
				tun->processTunnelMsg(msg);

				// Check reported tunnel state
				if(parms->getState() == Tunnel::CLOSED) {
					// Notification of tunnel closure, schedule tunnel shutdown
					DPRINTI("Received tunnel shutdown request from server");

					#ifdef DEBUG_EVENTS
					std::stringstream nfo;
					nfo << "Scheduling ShutdownTunnelEvent for tunnel ";
					nfo << tid.toString();
					nfo << " because client tunneler received shutdown request from server";
					g_pLogger->info("##########");
					g_pLogger->info(nfo.str());
					g_pLogger->info("##########");
					#endif

					scheduleEvent(new ShutdownTunnelEvent(0, this, tid, true));
				}
				else if(parms->getState() == Tunnel::AUTHREQ) {
					// Send back authentication response
					authenticationRequest(tun);
				}
			}
			else {
				// Unknown tunnel
				std::stringstream wrn;
				wrn << "ClientTunneler::messageArrived() - Unknown tunnel " << tid;
				g_pLogger->warn(wrn.str());
				fatalConnectionError = true;
			}
		}
		catch(InvalidTunnelID& e) {
			g_pLogger->error("Error parsing tunnel ID: " + e.what());
			fatalConnectionError = true;
		}
	}

	// Did something very serious fail?
	if(fatalConnectionError) {
		#ifdef DEBUG_TUN
		std::stringstream dbg;
		dbg << "Fatal error processing msg on connection " << connID;
		DPRINTV(dbg.str());
		#endif

		// No point keeping this connection around, kill it now
		Threads::MutexVar cmv(m_connectionsMtx);
		std::vector<Connection*>::iterator cit = getConnection(connID);
		destroyConnection(cit);
	}
}


// Notification of new raw data arrival
void ClientTunneler::rawDataArrived(unsigned int connID, const TunnelID& tid,
	const RawBuff& raw) throw(LookupManagementError)
{
	// Lookup tunnel
	Threads::MutexVar tmv(m_tunnelsMtx);
	Tunnel* tun = getTunFromPred(PredTunID(&tid));
	if(tun) {
		// Inform tunnel of raw data arrival
		tun->processRawData(raw);

		// Try sending this data on to the server
		tun->sendTunnelMsg(false, false, 0);
	}
	else {
		std::stringstream err;
		err << "ClientTunneler::rawDataArrived() - On connection " << connID;
		err << " Unknown tunnel " << tid;
		throw LookupManagementError(err.str());
	}
}


// Notification of a new UDP 'connection'
void ClientTunneler::newUDPConnectionArrived(Sock* insock) throw(CommsError)
{
	////////////////////////////////////////////////////////////////////
	// We currently only support fixed UDP tunnelling, so if we already
	// have a tunnel created we can ignore this 'new connection'.
	////////////////////////////////////////////////////////////////////

	// Do we already have a UDP tunnel to use?
	Threads::MutexVar cmv(m_connectionsMtx);
	Threads::MutexVar tmv(m_tunnelsMtx);
	if(m_tunnels.size() == 0) {
		DPRINTI("ClientTunneler::newUDPConnectionArrived()");

		// Create new tunnel config
		TunnelConfig* tcfg = createTunnelConfig();

		// Create raw server connection to handle incoming traffic
		Connection* rawconn = new RawConnection(insock, this, *tcfg->getTunnelID());

		// Create HTTP client connection to handle tunneled traffic
		HTTPConnection* httpconn = new HTTPConnection(this);

		// Save connection details
		addConnection(rawconn);
		addConnection(httpconn);

		// Create tunnel
		Tunnel* tun = new ClientTunnel(tcfg, this, Tunnel::CONNECT);
		tun->addLeftConnection(rawconn);
		tun->addRightConnection(httpconn);

		// Schedule connect out on HTTP side
		const ClientSideConfig* clientCfg = getClientSideConfig();
		m_connectionMgr->addProxyConnect(httpconn, 0, clientCfg->getNextHopIP(),
			clientCfg->getNextHopPort(), Enums::TCP, clientCfg->getSSL(),
			clientCfg->getCertificate());

		// Store Tunnel object in our list of active tunnels
		m_tunnels.push_back(tun);
		noTunnelsChanged();
	}
}


// Check we have received an ack for some sent message
void ClientTunneler::checkAckReceived(void)
{
	std::vector<Tunnel*>::iterator tit;
	Threads::MutexVar tmv(m_tunnelsMtx);

	#ifdef DEBUG_TUN
	if(m_tunnels.size() > 0)
		DPRINTV("ClientTunneler::checkAckReceived()");
	#endif

	// Check acks for all tunnels
	for(tit = m_tunnels.begin(); tit != m_tunnels.end(); ++tit) {
		// Let tunnel check we have received any due acks, & resend if not
		Tunnel* tun = *tit;
		if(tun->checkAckTimeout()) {
			tun->sendTunnelMsg(false, false, 0);
		}
	}

	// Schedule next ack check event
	scheduleEvent(new CheckAckEvent(ACK_CHECK_PERIOD, this));
}


// Notification of a poll timeout on a tunnel
void ClientTunneler::pollTimeout(const TunnelID& tid)
{
	// Lookup tunnel
	Threads::MutexVar tmv(m_tunnelsMtx);
	Tunnel* tun = getTunFromPred(PredTunID(&tid));
	if(tun) {
		// Send a poll msg to the server side
		tun->pollTunnel();
	}
	else {
		std::stringstream err;
		err << "ClientTunneler::pollTimeout() - Unknown tunnel " << tid;
		g_pLogger->warn(err.str());
	}
}


// Notification that a tunnel has changed state
void ClientTunneler::tunnelStateChanged(const TunnelID& tid,
	Tunnel::TunnelState state)
{
	DPRINTI("ClientTunneler::tunnelStateChanged() - " + tid.toString());

	// Has tunnel become CONNECTED?
	if(state == Tunnel::CONNECTED) {
		// Lookup tunnel
		Threads::MutexVar tmv(m_tunnelsMtx);
		Tunnel* tun = getTunFromPred(PredTunID(&tid));
		if(tun) {
			// Do we have a poll timeout scheduled for this tunnel?
			if(!isScheduled<PollTimeoutEvent, TunnelID>(tid)) {
				// Schedule first poll timeout for newly connected tunnel
				scheduleEvent(new PollTimeoutEvent(Time(0, tun->getPollTimeout()),
					this, tid));
			}
		}
		else {
			std::stringstream err;
			err << "ClientTunneler::tunnelStateChanged() - Unknown tunnel " << tid;
			g_pLogger->warn(err.str());
		}
	}
}


// Notification of a fatal error on a socket connection
void ClientTunneler::connectionError(ConnectionErrorEvent* error)
{
	// Get values for error
	unsigned int connID = error->getConnectionID();

	#ifdef DEBUG_TUN
	std::stringstream nfo;
	Tunnel::TunnelErrorType errType = error->getErrorType();
	nfo << "ClientTunneler::connectionError() - Connection id " << connID;
	nfo << ", type " << errType;
	DPRINTI(nfo.str());
	#endif

	Threads::MutexVar cmv(m_connectionsMtx);

	// Is this our server test connection?
	if(connID == m_serverTestConnID) {
		// Relaunch test connection
		launchTestServerConnection();
	}
	else {
		// Let base class handle it
		Tunneler::connectionError(error);
	}
}


// Notification that our test connection to the server has completed
void ClientTunneler::serverTestConnectResult(ConnectResultEvent* result)
{
	DPRINTI("ClientTunneler::serverTestConnectResult()");

	if(m_serverTestConnID > 0) {
		try {
			// Lookup connection
			Threads::MutexVar cmv(m_connectionsMtx);
			Threads::MutexVar tmv(m_tunnelsMtx);
			std::vector<Connection*>::iterator cit;
			cit = getConnection(m_serverTestConnID);
			Connection* conn = *cit;

			// Let connection manager clear up after connect attempt
			m_connectionMgr->delProxyConnect(conn);

			// What was the result?
			ConnectResultEvent::ConnectResultType resType = result->getType();
			if(resType == ConnectResultEvent::SUCCESS) {
				// Connected
				DPRINTV("Server test connection up");
			}
			else {
				// Keep trying
				DPRINTV("Server test connection failed");
				m_connectionMgr->addProxyConnect(conn, TEST_RECONNECT_DELAY,
					result->getHost(), result->getPort(), result->getProtocol(),
					result->getSSL(), result->getSSLCert());
			}
		}
		catch(LookupManagementError& e) {
			g_pLogger->warn("Error handling server test connect result: " + e.what());
		}
	}
	else {
		g_pLogger->warn("No test server connection to handle connect result");
	}
}


/***********************/
/* Protected Interface */
/***********************/


// The number of tunnels in existance has changed - Caller holds tunnels mutex
void ClientTunneler::noTunnelsChanged(void)
{
	// Tell all ClientTunnels how many tunnels we now manage
	unsigned int noTunnels = m_tunnels.size();
	std::vector<Tunnel*>::iterator tit;
	for(tit = m_tunnels.begin(); tit != m_tunnels.end(); ++tit) {
		ClientTunnel* tun = dynamic_cast<ClientTunnel*>(*tit);
		tun->tunnelsCount(noTunnels);
	}
}


// Create config for tunnel to server
TunnelConfig* ClientTunneler::createTunnelConfig(void)
{
	// Create tunnelID
	ClientSideConfig* clientCfg = getClientSideConfigMutable();
	TunnelID* tid = new TunnelID(incTunCount(), clientCfg->getFixedDestIP(),
		clientCfg->getFixedDestPort(), g_pConfig->getHostName(), g_pConfig->getPid(),
		clientCfg->getProtocol());

	// Create tunnel config
	return new TunnelConfig(tid, clientCfg->getProxySettings(),
		clientCfg->getAuthenticator()->getDefaultAuthSettings(),
		clientCfg->getFixedDestIP(), clientCfg->getFixedDestPort());
}


/*********************/
/* Private Interface */
/*********************/


// Respond to an authentication request from the server - Mutexes held by caller
void ClientTunneler::authenticationRequest(Tunnel* tun)
{
	DPRINTI("ClientTunneler::authenticationRequest()");

	// Send back authentication response
	tun->sendTunnelMsg(true, true, 0);
}


// Notification of an authentication failure
void ClientTunneler::authenticationFailed(unsigned int connID,
	const HTTPResponseMsg* respMsg, unsigned short respCode)
{
	#ifdef DEBUG_TUN
	std::stringstream nfo;
	nfo << "ClientTunneler::authenticationFailed() - Connection id " << connID;
	DPRINTI(nfo.str());
	#endif

	bool proxy = false;
	bool server = false;

	try {
		// Lookup connection (also lock tunnels to prevent race with tunnel shutdown)
		Threads::MutexVar cmv(m_connectionsMtx);
		Threads::MutexVar tmv(m_tunnelsMtx);
		std::vector<Connection*>::iterator cit = getConnection(connID);
		Connection* conn = *cit;

		// Lookup carried tunnels (can't rely on having tunnel ID from msg)
		std::vector<Tunnel*> tunnels;
		std::vector<Tunnel*>::iterator tit;
		getAllTunsFromPred(PredTunConnection(conn), tunnels);
		for(tit = tunnels.begin(); tit != tunnels.end(); ++tit) {
			ClientTunnel* tun = dynamic_cast<ClientTunnel*>(*tit);

			// Reconfigure whatever is necessary for the required auth
			if(respCode == 407) {
				proxy = true;
				setupProxyServerAuthConfig(respMsg->getAuthArgs(), tun->getProxySettings(),
					connID, tun->getTidStr());
			}
			else if(respCode == 401) {
				server = true;
				tun->setUserAuthenticated(false);
				setupProxyServerAuthConfig(respMsg->getAuthArgs(), tun->getServerAuthSettings(),
					connID, tun->getTidStr());
			}

			// Schedule chance to resend messages using authentication
			scheduleEvent(new SendMsgEvent(Time(0, RESEND_AUTH_DELAY), connID, true,
				true, *tun->getConfig()->getTunnelID(), this));
		}
	}
	catch(TunnelError& e) {
		g_pLogger->warn("Error handling authentication: " + e.what());

		// Schedule fatal error
		if(proxy)
			scheduleEvent(new ConnectionErrorEvent(connID, Tunnel::PROXY_AUTH_ERROR, this));
		else
			scheduleEvent(new ConnectionErrorEvent(connID, Tunnel::SERVER_AUTH_ERROR, this));
	}
}


// Setup config for required proxy/server authentication
void ClientTunneler::setupProxyServerAuthConfig(const std::vector<std::string>& authArgs,
	AuthSettings* settings, unsigned int connID, std::string tidStr)
	throw(AuthProcessingError)
{
	// Pick authentication scheme
	bool basic = false, digest = false;
	std::vector<std::string>::const_iterator ait;
	ait = std::find(authArgs.begin(), authArgs.end(), "Basic");
	if(ait != authArgs.end()) {
		// Use basic auth
		basic = true;
	}
	else {
		ait = std::find(authArgs.begin(), authArgs.end(), "Digest");
		if(ait != authArgs.end()) {
			// Use digest auth
			digest = true;
		}
		else {
			throw AuthProcessingError("No suitable authentication scheme suggested");
		}
	}

	// Modify config
	if(basic) {
		// Check we haven't already tried this
		if(settings->getAuthScheme() != AuthSettings::BASIC) {
			// Get parameters for basic authentication
			std::string realm;
			if(Utils::getParmValue(authArgs, "realm", realm)) {
				settings->setAuthScheme(AuthSettings::BASIC, realm);
			}
			else {
				std::stringstream err;
				err << "Failed to get all basic authentication parameters for tunnel ";
				err << tidStr;
				throw AuthProcessingError(err.str());
			}
		}
		else {
			std::stringstream err;
			err << "Basic authentication failed for tunnel " << tidStr;
			throw AuthProcessingError(err.str());
		}
	}
	else {
		// Get parameters for digest authentication
		std::string realm, nonce, qop;
		if(Utils::getParmValue(authArgs, "realm", realm) && Utils::getParmValue(authArgs,
		"nonce", nonce) && Utils::getParmValue(authArgs, "qop", qop)) {
			settings->setAuthScheme(AuthSettings::DIGEST, realm, qop, nonce);
		}
		else {
			std::stringstream err;
			err << "Failed to get all digest authentication parameters for tunnel ";
			err << tidStr;
			throw AuthProcessingError(err.str());
		}
	}
}


// Launch test server connection
void ClientTunneler::launchTestServerConnection(void)
{
	DPRINTV("ClientTunneler::launchTestServerConnection()");

	Threads::MutexVar cmv(m_connectionsMtx);
	Threads::MutexVar tmv(m_tunnelsMtx);

	// Tidy up any existing test connection
	destroyTestServerConnection();

	// Create new test connection
	HTTPConnection* httpconn = new HTTPConnection(this, true);
	addConnection(httpconn);
	m_serverTestConnID = httpconn->getID();

	// Schedule connect out to server
	const ClientSideConfig* clientCfg = getClientSideConfig();
	m_connectionMgr->addProxyConnect(httpconn, 0, clientCfg->getNextHopIP(),
		clientCfg->getNextHopPort(), Enums::TCP, clientCfg->getSSL(),
		clientCfg->getCertificate());
}


// Destroy test server connection
void ClientTunneler::destroyTestServerConnection(void)
{
	Threads::MutexVar cmv(m_connectionsMtx);
	Threads::MutexVar tmv(m_tunnelsMtx);

	if(m_serverTestConnID > 0) {
		DPRINTV("Destroying server test connection");
		try {
			// Lookup connection
			std::vector<Connection*>::iterator cit;
			cit = getConnection(m_serverTestConnID);

			// Destroy it
			destroyConnection(cit);
			m_serverTestConnID = 0;
		}
		catch(LookupManagementError& e) {
			g_pLogger->warn("Error destroying server test connection: " + e.what());
		}
	}
}


}	// namespace

