/* (C) 2005 Richard Mills / The Linux Academy Ltd.
 *
 * Tunnel controller for client end of tunnel.
 *
 * $Id: ClientTunneler.hh,v 1.41 2007-07-08 11:36:51 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#ifndef _CLIENT_TUNNELER_HH_
#define _CLIENT_TUNNELER_HH_

#include <vector>
#include "Tunneler.hh"
#include "Exceptions.hh"

namespace HTTPTunnel {

class Tunnel;
class Connection;
class HTTPResponseMsg;
class AuthSettings;
class ClientSideConfig;


class ClientTunneler : public Tunneler {
public:
   // Construction/destruction
   ClientTunneler(Server* server);
   virtual ~ClientTunneler();


	/**********/
	/* Events */
	/**********/

	// Notification of a new socket connection
	virtual void newConnectionArrived(Sock* insock) throw(CommsError);

	// Notification that a tunnel has insufficient left/right connections to be viable
	virtual void unviableTunnel(const TunnelID& tid);

	// Notification of new HTTP message arrival
	virtual void messageArrived(unsigned int connID, const HTTPMsg* msg)
		throw(LookupManagementError);

	// Notification of new tunneled/raw data arrival
	virtual void rawDataArrived(unsigned int connID, const TunnelID& tid,
		const RawBuff& raw) throw(LookupManagementError);

	// Notification of a new UDP 'connection'
	virtual void newUDPConnectionArrived(Sock* insock) throw(CommsError);

	// Check we have received an ack for some sent message
	virtual void checkAckReceived(void);

	// Notification of a poll timeout on a tunnel
	virtual void pollTimeout(const TunnelID& tid);

	// Notification that a tunnel has changed state
	virtual void tunnelStateChanged(const TunnelID& tid, Tunnel::TunnelState state);

	// Notification of a fatal error on a socket connection
	virtual void connectionError(ConnectionErrorEvent* error);

	// Notification that our test connection to the server has completed
	virtual void serverTestConnectResult(ConnectResultEvent* result);


protected:

	/*************/
	/* Constants */
	/*************/

	// Time to pause (milli-seconds) between preservation reconnects
	static const unsigned int			PRESERVE_RECONNECT_DELAY	= 250;

	// Delay (milli-seconds) before trying to resend authentication msg
	static const unsigned int			RESEND_AUTH_DELAY				= 100;


	/***********/
	/* Methods */
	/***********/

	// The number of tunnels in existance has changed
	virtual void noTunnelsChanged(void);

	// Create config for tunnel to server
	TunnelConfig* createTunnelConfig(void);


private:

	/*************/
	/* Constants */
	/*************/

	// Time between server test reconnection attempts
	static const unsigned int			TEST_RECONNECT_DELAY			= 10;


	/***********/
	/* Methods */
	/***********/

	// Respond to an authentication request from the server
	void authenticationRequest(Tunnel* tun);

	// Notification of an authentication failure
	void authenticationFailed(unsigned int connID, const HTTPResponseMsg* respMsg,
		unsigned short respCode);

	// Setup config for required proxy/server authentication
	void setupProxyServerAuthConfig(const std::vector<std::string>& authArgs,
		AuthSettings* settings, unsigned int connID, std::string tidStr)
		throw(AuthProcessingError);

	// Launch test server connection
	void launchTestServerConnection(void);

	// Destroy test server connection
	void destroyTestServerConnection(void);


	/********/
	/* Data */
	/********/

	// Our server test connection ID
	unsigned int					m_serverTestConnID;

};


}	// namespace

#endif

