/* (C) 2005 Richard Mills / The Linux Academy Ltd.
 *
 * An object capable of receiving and executing events.
 *
 * $Id: EventReceiver.cc,v 1.13 2007-04-17 13:44:04 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <algorithm>
#include <iostream>
#include <sstream>
#include "EventReceiver.hh"
#include "Event.hh"
#include "TimedEvent.hh"
#include "MutexVar.hh"
#include "Debug.hh"
#include "Callback.hh"
#include "Logger.hh"


// Referenec to global logger
extern HTTPTunnel::Logger* g_pLogger;


namespace HTTPTunnel {


// Construction/destruction
EventReceiver::EventReceiver()
: m_threadMgr(), m_eventThread(0), m_events(),
  m_eventsCond(), m_eventsMtx(), m_eventsRunning(true), m_shutdownMtx(),
  m_shutdownCond()
{
	// Launch main event scheduler loop
	Threads::ZeroArgClassCallback<EventReceiver> cb1(&EventReceiver::processEvents,
		this);
	m_eventThread = m_threadMgr.newThread(cb1);
}

EventReceiver::~EventReceiver()
{
	DPRINTV("EventReceiver::~EventReceiver()");

	#ifdef DEBUG_TUN
	// Take event queues mutexs
	Threads::MutexVar emv(m_eventsMtx);

	// Check shutdown has been performed by derived class
	if(m_eventsRunning) {
		g_pLogger->error("EventReceiver not shutdown by derived class");
		exit(1);
	}

	// Check we don't have any events left unprocessed
	if(m_events.size() != 0) {
		g_pLogger->error("EventReceiver has unprocessed events in destructor");
		exit(1);
	}
	#endif
}


/*************/
/* Accessors */
/*************/


// Are we shutting down?
bool EventReceiver::shuttingDown(void) const
{
	return (!m_eventsRunning);
}


/********************/
/* Public Interface */
/********************/


// Schedule a new event
void EventReceiver::scheduleEvent(Event* event)
{
	// Take event queue mutex
	Threads::MutexVar emv(m_eventsMtx);

	#ifdef DEBUG_EVENTS_SCHEDULING
	unsigned int queuedPos = 0;
	#endif

	// Are we allowed to schedule new events?
	if(m_eventsRunning) {
		// Insert event at right place in queue
		std::vector<Event*>::iterator eit = std::upper_bound(m_events.begin(),
			m_events.end(), event, Event::PredRunTime());
		m_events.insert(eit, event);

		// Signal timed event scheduling thread it has some work
		m_eventsCond.signal();

		#ifdef DEBUG_EVENTS_SCHEDULING
		for(eit = m_events.begin(); eit != m_events.end(); ++eit, ++queuedPos) {
			if(*eit == event)
				break;
		}
		#endif
	}
	else {
		// Can't schedule new event
		g_pLogger->warn("Can't schedule event because we are shutting down");
		delete event;
		event = 0;
	}

	#ifdef DEBUG_EVENTS_SCHEDULING
	if(m_events.size() > 1) {
		std::stringstream dbg;
		dbg << "### ";
		dbg << "Scheduled new event at queue pos " << queuedPos;
		dbg << ", event queue size " << m_events.size();

		if(m_events.size() > queuedPos + 1) {
			dbg << std::endl;
			dbg << "This event time = " << event->getRunTime() << ", ";
			dbg << "Next event time = " << m_events[queuedPos + 1]->getRunTime();
		}

		if(event) {
			dbg << std::endl;
			dbg << "Event details: ";
			event->details(dbg);
		}
		//std::vector<Event*>::iterator deit;
		//for(deit = m_events.begin(); deit != m_events.end(); ++deit) {
		//	(*deit)->details(dbg);
		//}
		DPRINTV(dbg.str());
	}
	#endif
}


/***********************/
/* Protected Interface */
/***********************/


// Instruct the event scheuling thread to terminate
void EventReceiver::shutdownSchedulers(void)
{
	DPRINTV("EventReceiver::shutdownSchedulers()");

	// Main scheduler
	shutdownMainScheduler();
}


/*********************/
/* Private Interface */
/*********************/


// Process events on the event queue - Thread entry point
void EventReceiver::processEvents(void)
{
	try {
		DPRINTV("EventReceiver::processEvents()");

		// Take event queue mutex
		Threads::MutexVar emv(m_eventsMtx);

		// Run until told to stop
		while(m_eventsRunning) {

			// Process next event on the queue
			if(m_events.size() > 0) {
				// Get first event
				Event* event = m_events.front();

				// Is it time to run this event?
				Time runTime = event->getRunTime();
				if(runTime.elapsed()) {
					// Run & delete event
					m_events.erase(m_events.begin());
					emv.unlock();
					event->execute();
					delete event;
					emv.lock();
				}
				else {
					// Wait until event scheduled time or until new event scheduled
					m_eventsCond.timedwait(emv, runTime.getTimeVal());
				}
			}
			else {
				// Should we accept any more events?
				if(m_eventsRunning) {
					// Wait for more events
					m_eventsCond.wait(emv);
				}
			}
		}

		// Clear up any still pending events
		while(m_events.size() > 0) {
			Event* event = m_events.front();
			m_events.erase(m_events.begin());
			delete event;
		}

		// Unlock timed events
		emv.unlock();

		// Signal we have finished
		Threads::MutexVar smv(m_shutdownMtx);
		DPRINTV("EventReceiver::processEvents() - Event thread returning");
		m_shutdownCond.signal();
		smv.unlock();
	}
	catch(TunnelError& e) {
		g_pLogger->error("TunnelError while processing events: " + e.what());
	}
	catch(Threads::ThreadError& e) {
		g_pLogger->error("ThreadError while processing events: " + e.what());
	}
	catch(std::exception& e) {
		g_pLogger->error("Unknown exception while processing events");
	}
}


// Shutdown the main event scheduler thread
void EventReceiver::shutdownMainScheduler(void)
{
	DPRINTV("EventReceiver::shutdownMainScheduler()");

	// Take event queue mutex
	Threads::MutexVar emv(m_eventsMtx);

	// Take shutdown mutex
	Threads::MutexVar smv(m_shutdownMtx);

	// Indicate to the scheduler we are shutting down
	m_eventsRunning = false;

	// Signal the scheduler to empty its event queue
	m_eventsCond.signal();
	emv.unlock();

	// Wait until the scheduler returns
	m_shutdownCond.wait(smv);
}


}	// namespace

