/* (C) 2005 Richard Mills / The Linux Academy Ltd.
 *
 * A convenient place to store all global configuration.
 *
 * $Id: GlobalConfig.cc,v 1.28 2007-07-08 11:36:51 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <string>
#include <iostream>
#include <iomanip>
#include <sstream>
#include <fstream>
#include <unistd.h>
#include "GlobalConfig.hh"
#include "ClientSideConfig.hh"
#include "ServerSideConfig.hh"
#include "ProxySettings.hh"
#include "ClientServer.hh"
#include "ServerServer.hh"
#include "MutexVar.hh"


namespace HTTPTunnel {


// Construction/destruction
GlobalConfig::GlobalConfig()
: m_mtx(), m_daemon(true), m_hostName(), m_debugLevel(2), m_pid(getpid()),
  m_pidFile(""), m_realm("FireDrill"), m_servers(), m_serverConfigs()
{
	// Create pseudo unique local hostname
	std::stringstream hostnamestr;
	char hostname[OUR_HOST_NAME_MAX + 1];
	if(gethostname(hostname, sizeof(hostname)) == 0) {
		hostnamestr << hostname;
	}

	// Try to ensure pseudeo-uniqueness by appending some random number
	srand(time(0));
	hostnamestr << std::setw(4) << std::setfill('0') << rand();
	m_hostName = hostnamestr.str();
}

GlobalConfig::~GlobalConfig()
{
	Threads::MutexVar mv(m_mtx);

	// Stop local servers
	destroyLocalServers();
}


/*************/
/* Utilities */
/*************/


// Serialize to a config file
void GlobalConfig::dumpToFile(std::string name) throw(ConfigError)
{
	Threads::MutexVar mv(m_mtx);

	// Open file for writing
	std::ofstream file(name.c_str());
	if(file) {
		/* Header */

		file << "##########################" << std::endl;
		file << "## " << STRNAME << " config" << std::endl;
		file << "## Version " << VERSION << std::endl;
		file << "##########################" << std::endl;

		/* Global options */

		file << std::endl;
		file << "#################" << std::endl;
		file << "## Global options" << std::endl;
		file << "#################" << std::endl;
		file << std::endl;
		file << "[GLOBAL]" << std::endl;

		// Daemon?
		if(!getDaemon()) {
			file << "nodaemon" << std::endl;
		}

		// Debug level
		file << "debug " << getDebug() << std::endl;

		// PID file
		std::string pidfile = getPIDFile();
		if(pidfile.size() > 0) {
			file << "pidfile " << pidfile << std::endl;
		}

		/* Servers config */

		file << std::endl;

		// Get config for all servers
		std::vector<SideConfig*>::const_iterator cit;
		for(cit = m_serverConfigs.begin(); cit != m_serverConfigs.end(); ++cit) {
			// Serialize config from this server
			(*cit)->serialize(file);
		}
	}
	else {
		std::stringstream err;
		err << "Failed to open config file " << name << " for writing";
		throw ConfigError(err.str());
	}
}


// Perform any required late-configuration
void GlobalConfig::doLateConfiguration(void) throw(ConfigError)
{
	Threads::MutexVar mv(m_mtx);

	try {
		// Do late configuration for each server config
		std::vector<SideConfig*>::iterator cit;
		for(cit = m_serverConfigs.begin(); cit != m_serverConfigs.end(); ++cit) {
			(*cit)->doLateConstruction();
		}
	}
	catch(TunnelError& e) {
		std::stringstream err;
		err << "Error performing config late configuration: " << e.what();
		throw ConfigError(err.str());
	}
}


/**********************/
/* Accessors/Mutators */
/**********************/


// Configure as a server
void GlobalConfig::configureServer(bool ssl, std::string cert, std::string certKey,
	std::string lhost, unsigned int lport, std::string firewallIPs,
	std::string firewallPorts, bool auth) throw(AuthError)
{
	Threads::MutexVar mv(m_mtx);
	m_serverConfigs.push_back(new ServerSideConfig(ssl, cert, certKey, lhost,
		lport, firewallIPs, firewallPorts, auth));
}

// Reconfigure a server
void GlobalConfig::reconfigureServer(unsigned int no, bool ssl, std::string cert,
	std::string certKey, std::string lhost, unsigned int lport, std::string firewallIPs,
	std::string firewallPorts, bool auth) throw(TunnelError)
{
	Threads::MutexVar mv(m_mtx);

	// Clear up existing server
	destroyNamedLocalServer(no);

	// Create replacement config
	m_serverConfigs.insert(m_serverConfigs.begin() + no, new ServerSideConfig(ssl,
		cert, certKey, lhost, lport, firewallIPs, firewallPorts, auth));

	// Start replacement server
	createNamedLocalServer(no);
}

// Configure default server
unsigned int GlobalConfig::configureDefaultServer(void) throw(AuthError)
{
	Threads::MutexVar mv(m_mtx);

	// Pick some sensible defaults
	configureServer(false, "", "", "", 8080, "", "", false);

	// Return index for new config
	return (m_serverConfigs.size() - 1);
}


// Configure as a client
void GlobalConfig::configureClient(Enums::Protocol prot, bool ssl, std::string cert,
	std::string lhost, unsigned int lport, std::string shost, unsigned int sport,
	std::string suser, std::string spasswd, std::string dhost, unsigned int dport)
{
	Threads::MutexVar mv(m_mtx);
	m_serverConfigs.push_back(new ClientSideConfig(prot, ssl, cert, lhost, lport,
		shost, sport, suser, spasswd, dhost, dport));
}

void GlobalConfig::configureClient(Enums::Protocol prot, bool ssl, std::string cert,
	std::string lhost, unsigned int lport, std::string shost, unsigned int sport,
	std::string suser, std::string spasswd, std::string dhost, unsigned int dport,
	std::string phost, unsigned int pport, std::string puser, std::string ppasswd)
{
	Threads::MutexVar mv(m_mtx);
	m_serverConfigs.push_back(new ClientSideConfig(prot, ssl, cert, lhost, lport,
		shost, sport, suser, spasswd, dhost, dport, phost, pport, puser, ppasswd));
}

void GlobalConfig::configureClient(Enums::Protocol prot, bool ssl, std::string cert,
	std::string lhost, unsigned int lport, std::string shost, unsigned int sport,
	std::string suser, std::string spasswd)
{
	Threads::MutexVar mv(m_mtx);
	m_serverConfigs.push_back(new ClientSideConfig(prot, ssl, cert, lhost, lport,
		shost, sport, suser, spasswd));
}

void GlobalConfig::configureClient(Enums::Protocol prot, bool ssl, std::string cert,
	std::string lhost, unsigned int lport, std::string shost, unsigned int sport,
	std::string suser, std::string spasswd, std::string phost, unsigned int pport,
	std::string puser, std::string ppasswd)
{
	Threads::MutexVar mv(m_mtx);
	m_serverConfigs.push_back(new ClientSideConfig(prot, ssl, cert, lhost, lport,
		shost, sport, suser, spasswd, phost, pport, puser, ppasswd));
}

// Reconfigure a client
void GlobalConfig::reconfigureClient(unsigned int no, Enums::Protocol prot, bool ssl,
	std::string cert, std::string lhost, unsigned int lport, std::string shost,
	unsigned int sport, std::string suser, std::string spasswd, std::string dhost,
	unsigned int dport) throw(TunnelError)
{
	Threads::MutexVar mv(m_mtx);

	// Clear up existing server
	destroyNamedLocalServer(no);

	// Create replacement config
	m_serverConfigs.insert(m_serverConfigs.begin() + no, new ClientSideConfig(prot,
		ssl, cert, lhost, lport, shost, sport, suser, spasswd, dhost, dport));

	// Start replacement server
	createNamedLocalServer(no);
}

void GlobalConfig::reconfigureClient(unsigned int no, Enums::Protocol prot, bool ssl,
	std::string cert, std::string lhost, unsigned int lport, std::string shost,
	unsigned int sport, std::string suser, std::string spasswd, std::string dhost,
	unsigned int dport, std::string phost, unsigned int pport, std::string puser,
	std::string ppasswd) throw(TunnelError)
{
	Threads::MutexVar mv(m_mtx);

	// Clear up existing server
	destroyNamedLocalServer(no);

	// Create replacement config
	m_serverConfigs.insert(m_serverConfigs.begin() + no, new ClientSideConfig(prot,
		ssl, cert, lhost, lport, shost, sport, suser, spasswd, dhost, dport, phost,
		pport, puser, ppasswd));

	// Start replacement server
	createNamedLocalServer(no);
}

void GlobalConfig::reconfigureClient(unsigned int no, Enums::Protocol prot, bool ssl,
	std::string cert, std::string lhost, unsigned int lport, std::string shost,
	unsigned int sport, std::string suser, std::string spasswd) throw(TunnelError)
{
	Threads::MutexVar mv(m_mtx);

	// Clear up existing server
	destroyNamedLocalServer(no);

	// Create replacement config
	m_serverConfigs.insert(m_serverConfigs.begin() + no, new ClientSideConfig(prot,
		ssl, cert, lhost, lport, shost, sport, suser, spasswd));

	// Start replacement server
	createNamedLocalServer(no);
}

void GlobalConfig::reconfigureClient(unsigned int no, Enums::Protocol prot, bool ssl,
	std::string cert, std::string lhost, unsigned int lport, std::string shost,
	unsigned int sport, std::string suser, std::string spasswd, std::string phost,
	unsigned int pport, std::string puser, std::string ppasswd) throw(TunnelError)
{
	Threads::MutexVar mv(m_mtx);

	// Clear up existing server
	destroyNamedLocalServer(no);

	// Create replacement config
	m_serverConfigs.insert(m_serverConfigs.begin() + no, new ClientSideConfig(prot,
		ssl, cert, lhost, lport, shost, sport, suser, spasswd, phost, pport, puser,
		ppasswd));

	// Start replacement server
	createNamedLocalServer(no);
}

// Configure default client
unsigned int GlobalConfig::configureDefaultClient(void)
{
	Threads::MutexVar mv(m_mtx);

	// Pick some sensible defaults
	configureClient(Enums::TCP, false, "", "", 1080, "localhost", 8080, "", "");

	// Return index for new config
	return (m_serverConfigs.size() - 1);
}


// Get local host name
std::string GlobalConfig::getHostName(void) const
{
	return m_hostName;
}


// Get/set daemon flag
void GlobalConfig::setDaemon(bool flag)
{
	m_daemon = flag;
}
bool GlobalConfig::getDaemon(void) const
{
	return m_daemon;
}

// Get/set debug level
void GlobalConfig::setDebug(unsigned int debug)
{
	m_debugLevel = debug;
}
unsigned int GlobalConfig::getDebug(void) const
{
	return m_debugLevel;
}

// Get our process ID
unsigned int GlobalConfig::getPid(void) const
{
	return m_pid;
}

// Get/set PID file name
void GlobalConfig::setPIDFile(std::string pidfile)
{
	m_pidFile = pidfile;
}
std::string GlobalConfig::getPIDFile(void) const
{
	return m_pidFile;
}


// Get/set the default realm
std::string GlobalConfig::getRealm(void) const
{
	return m_realm;
}
void GlobalConfig::setRealm(std::string realm)
{
	m_realm = realm;
}


// Create local servers
void GlobalConfig::createLocalServers(void) throw(TunnelError)
{
	Threads::MutexVar mv(m_mtx);

	// Create server for each set of config
	for(unsigned int cit = 0; cit < m_serverConfigs.size(); ++cit)
		createNamedLocalServer(cit);
}

// Create named local server
void GlobalConfig::createNamedLocalServer(unsigned int no) throw(TunnelError)
{
	Threads::MutexVar mv(m_mtx);

	// Get config for named server
	SideConfig* cfg = m_serverConfigs[no];

	// Create server
	if(cfg->isClient()) {
		// Create client side server
		ClientSideConfig* ccfg = dynamic_cast<ClientSideConfig*>(cfg);
		m_servers.insert(m_servers.begin() + no, new ClientServer(ccfg));
	}
	else {
		// Create server side server
		ServerSideConfig* scfg = dynamic_cast<ServerSideConfig*>(cfg);
		m_servers.insert(m_servers.begin() + no, new ServerServer(scfg));
	}
}


// Destroy local servers
void GlobalConfig::destroyLocalServers(void)
{
	Threads::MutexVar mv(m_mtx);

	// Stop local servers
	while(m_servers.size() > 0)
		destroyNamedLocalServer(0);
}

// Destroy specified local server/config
void GlobalConfig::destroyNamedLocalServer(unsigned int no)
{
	Threads::MutexVar mv(m_mtx);

	// Remove server config
	std::vector<SideConfig*>::iterator cit = m_serverConfigs.begin() + no;
	SideConfig* cfg = *cit;
	m_serverConfigs.erase(cit);

	// Got a corresponding server for this config?
	if(m_servers.size() > no) {
		std::vector<Server*>::iterator sit = m_servers.begin() + no;
		Server* svr = *sit;
		if(svr->getConfig() == cfg) {
			// Clear up server
			delete svr;
			m_servers.erase(sit);
		}
	}

	// Delete server config
	delete cfg;
}


// Have we got at least one client or server configured?
bool GlobalConfig::confCliSvr(void)
{
	return (m_serverConfigs.size() > 0);
}


}	// namespace HTTPTunnel

