/* (C) 2005 Richard Mills / The Linux Academy Ltd.
 *
 * Handle reading/writing for one end of an HTTP tunnel
 * connection.
 *
 * $Id: HTTPConnection.cc,v 1.41 2007-07-07 18:05:11 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include "HTTPConnection.hh"
#include "Tunneler.hh"
#include "HTTPReader.hh"
#include "HTTPWriter.hh"
#include "Debug.hh"
#include "Tunnel.hh"
#include "SendMsgEvent.hh"
#include "ServerTestConnectResultEvent.hh"
#include "GlobalConfig.hh"
#include "Sock.hh"


// Global references
extern HTTPTunnel::GlobalConfig* g_pConfig;


namespace HTTPTunnel {


// Construction/destruction
HTTPConnection::HTTPConnection(Sock* sock, Tunneler* tunneler)
: Connection(sock, tunneler), m_clientAddr(sock->getConnectedClientAddr()),
  m_serverTest(false)
{
	DPRINTV("HTTPConnection::HTTPConnection(server)");

	// Construct reader/writer
	createReaderWriter();
}

HTTPConnection::HTTPConnection(Tunneler* tunneler, bool serverTest)
: Connection(tunneler), m_clientAddr(), m_serverTest(serverTest)
{
	DPRINTV("HTTPConnection::HTTPConnection(client)");
}

HTTPConnection::~HTTPConnection()
{
	#ifdef DEBUG_TUN
	std::stringstream dbg;
	dbg << "HTTPConnection::~HTTPConnection() - connection " << getID();
	DPRINTV(dbg.str());
	#endif

	// We are closing down
	Threads::MutexVar cmv(m_connMtx);
	m_closing = true;
}


/********************/
/* Public Interface */
/********************/


// Let us know the result of a connection attempt
void HTTPConnection::connectResult(ConnectResultEvent::ConnectResultType result,
	Sock* sock) throw(CommsError)
{
	#ifdef DEBUG_TUN
	std::stringstream dbg;
	dbg << "HTTPConnection::connectResult() - connection " << getID();
	DPRINTV(dbg.str());
	#endif

	// Are we a full connection or just a server test?
	if(m_serverTest) {
		serverTestConnectResult(result, sock);
	}
	else {
		// Base class
		Connection::connectResult(result, sock);

		// Success?
		if(result == ConnectResultEvent::SUCCESS) {
			// Now send initial HTTP GET to setup HTTP connection
			poll(true);
		}
	}
}


// Poll the other end of this connection with an empty GET message
void HTTPConnection::poll(bool force) throw(CommsError)
{
	Threads::MutexVar mv(m_connMtx);

	// Can't poll unless connected & active
	if(connected() && active()) {
		// Only bother polling if we actually currently carry a tunnel
		if(m_carriedTunnels.size() > 0) {
			// Force the sending of a (probably empty) message
			Tunnel* tun = m_carriedTunnels.front();
			m_tunneler->scheduleEvent(new SendMsgEvent(Time(0,0), getID(), true,
				force, *tun->getConfig()->getTunnelID(), m_tunneler));
		}
	}
	else {
		throw CommsError("Error polling server: Not connected or inactive");
	}
}


/*************/
/* Accessors */
/*************/


// Get address of connected client at server
std::string HTTPConnection::getClientAddr(void) const
{
	return m_clientAddr;
}


/***********************/
/* Protected Interface */
/***********************/


// Construct reader/writer
void HTTPConnection::createReaderWriter(void)
{
	// Create HTTP reader & writer
	Threads::MutexVar mv(m_connMtx);
	m_reader = new HTTPReader(this);
	m_writer = new HTTPWriter(this);

	// Only truly connected once reader & writer are created
	setConnected();
}


// Got result of server test connection
void HTTPConnection::serverTestConnectResult(
	ConnectResultEvent::ConnectResultType result, Sock* sock)
{
	Threads::MutexVar mv(m_connMtx);

	// Update state
	m_connecting = false;
	m_sock = sock;

	// Let tunneler know connect attempt is over
	m_tunneler->scheduleEvent(new ServerTestConnectResultEvent(m_tunneler,
		getID(), getDestIP(), getDestPort(), getDestProtocol(), getDestSSL(),
		getDestSSLCert(), result));

	// Success?
	if(result == ConnectResultEvent::SUCCESS) {
		// Create required reader/writer
		createReaderWriter();
	}
}


}	// namespace


