/* (C) 2007 Richard Mills / The Linux Academy Ltd.
 *
 * Encapsulate resolved host IP addresses.
 *
 * $Id$
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include "HostAddr.hh"


namespace HTTPTunnel {


// Construction/destruction
HostAddr::HostAddr(struct hostent* phost) throw(MemoryError)
: m_addresses(), m_resolved(false)
{
	// Got any resolved addresses?
	m_resolved = (phost->h_length > 0);

	// Iterate over hostent addresses
	for(int i = 0; i < phost->h_length; ++i) {
		unsigned char* addr = (unsigned char*)(phost->h_addr_list[i]);
		RawBuff* addrBuf = new RawBuff(addr, sizeof(struct in_addr));
		m_addresses.push_back(addrBuf);
	}
}

HostAddr::HostAddr(std::string addrStr) throw(MemoryError)
: m_addresses(), m_resolved(false)
{

}

HostAddr::~HostAddr()
{
	// Tidy up addresses
	while(m_addresses.size() > 0) {
		std::vector<RawBuff*>::iterator ait = m_addresses.begin();
		delete *ait;
		m_addresses.erase(ait);
	}
}


/*************/
/* Accessors */
/*************/


// Have we got a resolved IP address?
bool HostAddr::resolved(void) const
{
	return m_resolved;
}


// Get resolved IP address
const struct in_addr* HostAddr::getAddr(unsigned int no)
{
	// Got the requested address?
	if(m_addresses.size() > no) {
		RawBuff* addrBuf = m_addresses[no];
		return (const struct in_addr*)(addrBuf->begin());
	}

	// Not found
	return 0;
}


// Get resolved IP address as a string
std::string HostAddr::getAddrStr(unsigned int no)
{
	std::string addrStr;

	// Got the requested address?
	if(m_addresses.size() > no) {
		// Turn IP address back into string
		char address[INET_ADDRSTRLEN];
		const struct in_addr* inaddr = getAddr(no);
		inet_ntop(AF_INET, inaddr, address, sizeof(address));
		addrStr = address;
	}

	return addrStr;
}


/*********************/
/* Private Interface */
/*********************/


// Resolve an address to a host entry
int HostAddr::getHostByName(std::string addr, struct hostent* host)
{
	#if !defined(PLATFORM_OSX) && !defined(PLATFORM_CYGWIN)
	char auxBuff[AUX_HOST_BUFF_SIZE];
	size_t auxBuffLen = sizeof(auxBuff);
	int gethosterrno = 0;
	#endif

	#ifdef PLATFORM_SOLARIS
	struct hostent* gethostres = gethostbyname_r(addr.c_str(), host,
		auxBuff, auxBuffLen, &gethosterrno);
	return (gethostres == 0)? -1 : 0;

	#elif defined(PLATFORM_OSX) || defined(PLATFORM_CYGWIN)
	Threads::MutexVar mv(c_mtx);
	struct hostent *gethostres = gethostbyname(addr.c_str());
	if(gethostres)
		*host = *gethostres;
	return (gethostres == 0)? -1 : 0;

	#else
	struct hostent* phost = 0;
	int gethostres = gethostbyname_r(addr.c_str(), host, auxBuff,
		auxBuffLen, &phost, &gethosterrno);
	return (gethostres == 0)? 0 : -1;
	#endif
}


}	// namespace

