/* (C) 2006 Richard Mills / The Linux Academy Ltd.
 *
 * A convenient place to store all global configuration for
 * a server tunneler.
 *
 * $Id: ServerSideConfig.cc,v 1.13 2007-05-26 13:10:56 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include "ServerSideConfig.hh"
#include "Enums.hh"
#include "ServerAuthSettings.hh"
#include "GlobalConfig.hh"
#include "Firewall.hh"

#ifdef AUTH_DB
#include "DBServerAuthenticator.hh"
#endif
#include "DefaultServerAuthenticator.hh"


// Global references
extern HTTPTunnel::GlobalConfig* g_pConfig;


namespace HTTPTunnel {


// Construction/destruction
ServerSideConfig::ServerSideConfig(bool ssl, std::string cert, std::string certKey,
	std::string lhost, unsigned int lport, std::string firewallIPs,
	std::string firewallPorts, bool auth) throw(AuthError)
: SideConfig(Enums::TCP, ssl, cert, lhost, lport), m_useAuth(auth), m_firewall(0),
  m_certKey(certKey)
{
	// Create firewall
	m_firewall = new Firewall(firewallIPs, firewallPorts);
}

ServerSideConfig::~ServerSideConfig()
{
	// Destroy firewall
	if(m_firewall) {
		delete m_firewall;
		m_firewall = 0;
	}
}


/*************/
/* Accessors */
/*************/


// Server config?
bool ServerSideConfig::isServer(void) const
{
	return true;
}

// Get firewall
Firewall* ServerSideConfig::getFirewall(void)
{
	return m_firewall;
}

// Get SSL private key
std::string ServerSideConfig::getCertificateKey(void) const
{
	return m_certKey;
}


/************/
/* Mutators */
/************/


// Turn authentication on/off
void ServerSideConfig::setAuthenticate(bool auth) throw(AuthError)
{
	m_useAuth = auth;
}


/********************/
/* Public Interface */
/********************/


// Do any required late-construction of data
void ServerSideConfig::doLateConstruction(void) throw(TunnelError)
{
	// Create authenticator
	createAuthenticator();
}


// Serialize to a file
void ServerSideConfig::serialize(std::ofstream& file)
{
	// Server header
	file << "#######################" << std::endl;
	file << "## Server configuration" << std::endl;
	file << "#######################" << std::endl;
	file << std::endl;
	file << "[Server]" << std::endl;

	// Local listening address
	file << "listen ";
	if(getListenIP().size() > 0) {
		file << getListenIP() << " ";
	}
	file << getListenPort() << std::endl;

	// Server authentication
	if(m_useAuth) {
		file << "svrauth" << std::endl;
	}

	// SSL certificate & key
	if(m_ssl) {
		file << "ssl " << getCertificate() << " " << getCertificateKey();
		file << std::endl;
	}

	// Firewall settings
	m_firewall->serialize(file);

	file << std::endl;
}


/***********************/
/* Protected Interface */
/***********************/


// Create authenticator
void ServerSideConfig::createAuthenticator(void) throw(AuthError)
{
	// Create new one
	if(m_useAuth) {

		#ifdef AUTH_DB
		if(m_authenticator) {
			// Destroy old default authenticator
			delete m_authenticator;
		}

		// Create new authenticator
		m_authenticator = new DBServerAuthenticator();
		#else

		if(!m_authenticator) {
			// Create new default authenticator
			m_authenticator = new DefaultServerAuthenticator();
		}
		#endif
	}
	else {
		if(!m_authenticator) {
			// Create new default authenticator
			m_authenticator = new DefaultServerAuthenticator();
		}
	}

	// Configure server auth settings
	ServerAuthSettings* authSettings = m_authenticator->getDefaultAuthSettings();
	if(m_useAuth) {
		// Use digest authentication
		authSettings->setAuthScheme(AuthSettings::DIGEST, g_pConfig->getRealm(), "auth");
	}
	else {
		// Use no authentication
		authSettings->setAuthScheme(AuthSettings::NONE);
	}
}


}	// namespace

