/* (C) 2006 Richard Mills / The Linux Academy Ltd.
 *
 * Server side Tunnel.
 *
 * $Id: ServerTunnel.cc,v 1.9 2007-08-04 11:21:43 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <sstream>
#include "ServerTunnel.hh"
#include "Logger.hh"
#include "MutexVar.hh"
#include "Debug.hh"
#include "GlobalConfig.hh"
#include "Authenticator.hh"
#include "Utils.hh"
#include "HTTPMsg.hh"
#include "Connection.hh"
#include "RawConnection.hh"
#include "HTTPConnection.hh"


// Global references
extern HTTPTunnel::Logger* g_pLogger;
extern HTTPTunnel::GlobalConfig* g_pConfig;


namespace HTTPTunnel {


// Construction/destruction
ServerTunnel::ServerTunnel(TunnelConfig* tcfg, Tunneler* tunneler, TunnelState state)
: Tunnel(tcfg, tunneler, LEFT, state)
{
	DPRINTI("ServerTunnel::ServerTunnel()");
}

ServerTunnel::~ServerTunnel()
{
	#ifdef DEBUG_TUN
	std::stringstream str;
	str << "ServerTunnel::~ServerTunnel() - " << m_tid;
	DPRINTI(str.str());
	#endif

	Threads::MutexVar mv(m_mtx);

	// Tell user they own one less tunnel
	Authenticator* authenticator = getAuthenticator();
	authenticator->userTunnelClosed(m_user, this);
}


/*************/
/* Accessors */
/*************/


// Get source addresses of connections for this tunnel
std::string ServerTunnel::getSrcConnectionAddresses(void)
{
	std::stringstream res;

	Threads::MutexVar mv(m_mtx);

	// Get addresses for left side connections
	std::vector<Connection*>::iterator cit;
	bool comma = false;
	for(cit = m_leftConn.begin(); cit != m_leftConn.end(); ++cit) {
		// Should only be HTTPConnections in server side tunnel
		HTTPConnection* conn = dynamic_cast<HTTPConnection*>(*cit);
		if(conn) {
			// Do we have a client address?
			std::string cliAddr = conn->getClientAddr();
			if(cliAddr.size() > 0) {
				if(comma)
					res << ",";
				else
					comma = true;
				res << cliAddr;
			}
		}
	}

	return res.str();
}

// Get destination addresses of connections for this tunnel
std::string ServerTunnel::getDstConnectionAddresses(void)
{
	std::stringstream res;

	Threads::MutexVar mv(m_mtx);

	// Get addresses for right side connections
	std::vector<Connection*>::iterator cit;
	bool comma = false;
	for(cit = m_rightConn.begin(); cit != m_rightConn.end(); ++cit) {
		// Should only be RawConnections in server side tunnel
		RawConnection* conn = dynamic_cast<RawConnection*>(*cit);
		if(conn) {
			// Do we have a destination address?
			std::string destAddr = conn->getDestIP();
			unsigned int destPort = conn->getDestPort();
			if(destPort > 0) {
				if(comma)
					res << ",";
				else
					comma = true;
				res << destAddr << ":" << destPort;
			}
		}
	}

	return res.str();
}


/********************/
/* Public Interface */
/********************/


// Process some newly arrived tunnel msg from one of our connections
void ServerTunnel::processTunnelMsg(const HTTPMsg* msg)
{
	DPRINTV("ServerTunnel::processTunnelMsg() - " + m_tid.toString());

	Threads::MutexVar mv(m_mtx);

	// Do we need to try and authenticate client?
	if(!userAuthenticated()) {
		DPRINTV("Client not authenticated");
		authenticateClient(msg->getAuthArgs());
	}

	// Is the client authenticated now?
	if(userAuthenticated()) {
		// Base class
		Tunnel::processTunnelMsg(msg);
	}
}


/***********************/
/* Protected Interface */
/***********************/


// Create complete HTTP msg parameters for sending down tunnel - Caller hold mutex
HTTPMsgParms* ServerTunnel::makeTunnelMsg(const HTTPMsgParms* recvd,
	unsigned int maxPayload, bool retry) throw(MemoryError)
{
	// Get basic parameters from base class
	HTTPMsgParms* parms = Tunnel::makeTunnelMsg(recvd, maxPayload, retry);

	// Do we need to authenticate the client?
	if(!userAuthenticated()) {
		parms->setRespCodePhrase(401, "Unauthorized");
	}

	// Is the client version ok?
	if(recvd->getState() == CONNECT) {
		std::string cliVer = recvd->getCliVer();
		std::vector<std::string> verFields;
		Utils::SplitString(cliVer, ".", verFields);
		if(verFields.size() == 3) {
			unsigned int cliMajor = atoi(verFields[0].c_str());
			unsigned int cliMinor = atoi(verFields[1].c_str());
			unsigned int cliPoint = atoi(verFields[2].c_str());
			bool cliMajorOk = (cliMajor >= MIN_CLIVER_MAJOR);
			bool cliMinorOk = (cliMinor >= MIN_CLIVER_MINOR);
			bool cliPointOk = (cliPoint >= MIN_CLIVER_POINT);
			if(!cliMajorOk || !cliMinorOk || !cliPointOk) {
				// Client too old, so reject them
				std::stringstream nfo;
				nfo << "Client version " << cliVer << " is too old, rejecting them";
				g_pLogger->info(nfo.str());

				// Tell client they are forbidden
				parms->setRespCodePhrase(403, "Forbidden");
			}
		}
		else {
			// Reject client
			std::stringstream wrn;
			wrn << "Bad format for client version identifier: " << cliVer;
			g_pLogger->warn(wrn.str());

			// Tell client they are forbidden
			parms->setRespCodePhrase(403, "Forbidden");
		}
	}

	return parms;
}


// Successfully sent a new message down the tunnel - Caller hold mutex
void ServerTunnel::sentNewTunnelMsg(HTTPMsgParms* parms)
{
	// Only remember if the sent msg is part of a normal authenticated tunnel session
	bool remember = userAuthenticated();
	Tunnel::sentNewTunnelMsg(parms, remember);
}


// Can we send a new tunnel msg? - Caller holds mutex
bool ServerTunnel::clearToSendNewMsg(Connection* conn)
{
	bool sendNew = false;

	try {
		// Check for unacked management msgs and consider resending them early
		checkUnackedMgmtMsgs();

		// New message or resend of old message?
		sendNew = !resendUnackedMsgs(conn);
	}
	catch(ResendError& e) {
		g_pLogger->warn("Error resending msg: " + e.what());
	}

	return sendNew;
}


// Determine whether an ack implies we should resend a msg - Caller holds mutex
void ServerTunnel::checkForResendRequest(const SeqNo& ackNo)
{
}


// Check whether any connections need to be blocked/unblocked - Caller holds mutex
void ServerTunnel::checkBlockedConnections(void)
{
	// Check raw connections
	std::vector<Tunnel::Selector*> filt;
	filt.push_back(new SelRightConn(this));
	std::vector<Connection*> conns;
	getConnsFromSel(filt, conns);

	// Do we want to block or unblock?
	bool unblock = false;
	bool block = (m_rawDataQueueSize >= MAX_RAW_DATA_FILL);
	if(!block) {
		// Only unblock if we're well below full
		unsigned int buffFill = (unsigned int)(((float)m_rawDataQueueSize /
			(float)MAX_RAW_DATA_FILL) * 100);
		unblock = (buffFill <= RAW_DATA_UNBLOCK_LEVEL);
	}

	// Check each connection
	std::vector<Connection*>::iterator cit;
	for(cit = conns.begin(); cit != conns.end(); ++cit) {
		// Get connection
		Connection* conn = *cit;

		// Need to change blocked state?
		if(block)
			conn->setBlocked(true);
		else if(unblock)
			conn->setBlocked(false);
	}
}


/*********************/
/* Private Interface */
/*********************/


// Try to authenticate a client - Caller holds mutex
void ServerTunnel::authenticateClient(const std::vector<std::string>& authArgs)
{
	// Get supplied user name
	std::string user;
	if(Utils::getParmValue(authArgs, "username", user)) {
		// Use supplied user name
		m_user = user;
	}

	// First check quotas for number of tunnels
	Authenticator* authenticator = getAuthenticator();
	if(authenticator->getUserCanCreateTunnel(m_user)) {
		// Try to authenticate
		authenticator->authenticate(m_user, this, authArgs);

		#ifdef DEBUG_TUN
		std::stringstream nfo;
		if(userAuthenticated())
			nfo << "Client auth OK";
		else
			nfo << "Client auth failed";
		DPRINTV(nfo.str());
		#endif
	}
	else {
		// User trying to exceed tunnel number quota
		std::stringstream nfo;
		nfo << "User " << m_user << " exceeding tunnel count quota, ";
		nfo << "not allowing them to authenticate";
		g_pLogger->info(nfo.str());
		recordConnectionError(FATAL_QUOTAS_ERROR);
	}
}


// Is this tunnel user authenticated? - Caller holds mutex
bool ServerTunnel::userAuthenticated(void)
{
	Authenticator* authenticator = getAuthenticator();
	return authenticator->getUserAuthenticated(m_user, this);
}


}	// namespace

