/* (C) 2005 Richard Mills / The Linux Academy Ltd.
 *
 * Tunnel controller for server end of tunnel.
 *
 * $Id: ServerTunneler.cc,v 1.71 2007-06-30 16:43:23 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <sstream>
#include "Tunneler.hh"
#include "ServerTunneler.hh"
#include "Debug.hh"
#include "Connection.hh"
#include "RawConnection.hh"
#include "HTTPConnection.hh"
#include "GlobalConfig.hh"
#include "ServerSideConfig.hh"
#include "Logger.hh"
#include "ServerTunnel.hh"
#include "HTTPMsg.hh"
#include "HTTPRequestMsg.hh"
#include "ShutdownTunnelEvent.hh"
#include "ShutdownConnectionEvent.hh"
#include "CheckAckEvent.hh"
#include "HTTPMsgParms.hh"
#include "ErrorMsgParms.hh"
#include "Authenticator.hh"
#include "ConnectionManager.hh"
#include "Sock.hh"


// Global references
extern HTTPTunnel::Logger* g_pLogger;
extern HTTPTunnel::GlobalConfig* g_pConfig;


namespace HTTPTunnel {


// Construction/destruction
ServerTunneler::ServerTunneler(Server* server)
: Tunneler(server)
{
	DPRINTV("ServerTunneler::ServerTunneler()");
}

ServerTunneler::~ServerTunneler()
{
	DPRINTV("ServerTunneler::~ServerTunneler()");

	// Shutdown event scheduling
	if(!shuttingDown())
		shutdownSchedulers();
}


/**********/
/* Events */
/**********/


// Notification of a new socket connection
void ServerTunneler::newConnectionArrived(Sock* insock)
	throw(CommsError)
{
	DPRINTI("ServerTunneler::newConnectionArrived()");

	//////////////////////////////////////////////////////////////////
	// In ServerTunneler we can't really do much until we are told
	// whether this is a connection from a new tunnel, or a new pipe
	// within an existing tunnel.
	//////////////////////////////////////////////////////////////////

	// Create HTTP connection to handle incoming traffic
	Threads::MutexVar cmv(m_connectionsMtx);
	HTTPConnection* httpconn = new HTTPConnection(insock, this);
	addConnection(httpconn);
}


// Notification that a tunnel has insufficient left/right connections to be viable
void ServerTunneler::unviableTunnel(const TunnelID& tid)
{
	DPRINTI("ServerTunneler::unviableTunnel() - " + tid.toString());

	// Lookup tunnel
	Threads::MutexVar tmv(m_tunnelsMtx);
	Tunnel* tun = getTunFromPred(PredTunID(&tid));
	if(tun) {
		bool preserving = false;

		// Get number of remaining left side connections for tunnel
		unsigned int left = tun->noLeftConnections();
		if(left == 0) {
			/* In the server, no remaining left connections means that if
			 * we want to try and preserve this tunnel we should wait for
			 * a re-connection from the client.
			 */

			// Do we want to try and preserve this tunnel?
			if(!tun->errorLevelExceeded(Tunnel::LEFT)) {
				// Schedule timeout for client to reconnect
				DPRINTI(tid.toString() + " - Trying to preserve");
				preserving = true;

				#ifdef DEBUG_EVENTS
				std::stringstream nfo;
				nfo << "Scheduling future ShutdownTunnelEvent for tunnel id ";
				nfo << tid.toString();
				nfo << " because server tunneler has no remaining left connections";
				g_pLogger->info("##########");
				g_pLogger->info(nfo.str());
				g_pLogger->info("##########");
				#endif

				scheduleEvent(new ShutdownTunnelEvent(TUNNEL_RECONNECT_TIMEOUT,
					this, tid, false));
			}
			else {
				// Schedule tunnel shutdown
				DPRINTI(tid.toString() + " - Too many errors, not preserving");

				#ifdef DEBUG_EVENTS
				std::stringstream nfo;
				nfo << "Scheduling ShutdownTunnelEvent for tunnel id ";
				nfo << tid.toString();
				nfo << " because server tunneler has no remaining left connections";
				g_pLogger->info("##########");
				g_pLogger->info(nfo.str());
				g_pLogger->info("##########");
				#endif

				scheduleEvent(new ShutdownTunnelEvent(0, this, tid, false));
			}
		}

		// Get number of remaining right side connections for tunnel
		unsigned int right = tun->noRightConnections();
		if(right == 0 && !preserving) {
			/* In the server, no remaining right connections means we should
			 * abandon this tunnel and close it down.
			 */

			#ifdef DEBUG_EVENTS
			std::stringstream nfo;
			nfo << "Scheduling ShutdownTunnelEvent for tunnel id ";
			nfo << tid.toString();
			nfo << " because server tunneler has no remaining right connections";
			g_pLogger->info("##########");
			g_pLogger->info(nfo.str());
			g_pLogger->info("##########");
			#endif

			// Schedule tunnel shutdown
			DPRINTI(tid.toString() + " - No connections, not preserving");
			scheduleEvent(new ShutdownTunnelEvent(0, this, tid, false));
		}
	}
	else {
		std::stringstream err;
		err << "ServerTunneler::unviableTunnel() - Unknown tunnel " << tid;
		g_pLogger->warn(err.str());
	}
}


// Notification of new HTTP message arrival
void ServerTunneler::messageArrived(unsigned int connID, const HTTPMsg* msg)
	throw(LookupManagementError)
{
	#ifdef DEBUG_TUN
	std::stringstream nfo;
	nfo << "ServerTunneler::messageArrived() - On connection " << connID;
	DPRINTV(nfo.str());
	#endif

	bool fatalConnectionError = false;

	try {
		// Find out which Tunnel this message is from
		const HTTPMsgParms* parms = msg->getParms();
		Tunnel::TunnelState state = parms->getState();
		TunnelID tid(parms->getTid());
		Threads::MutexVar cmv(m_connectionsMtx);
		Threads::MutexVar tmv(m_tunnelsMtx);
		Tunnel* tun = getTunFromPred(PredTunID(&tid));

		// Do we already know about the tunnel this msg claims it is from?
		if(tun) {
			// Inform tunnel of payload arrival
			tun->processTunnelMsg(msg);

			// Lookup connection
			std::vector<Connection*>::iterator cit = getConnection(connID);
			Connection* conn = *cit;

			// What state does this message report the tunnel as?
			if(state == Tunnel::CONNECT) {
				// A new pipe connection within this tunnel
				DPRINTI("Got new pipe connection within tunnel");
				newTunnelPipeConnectionRequest(tun, conn);
			}
			else if(state == Tunnel::AUTHRESP) {
				// Check authentication response
				authenticationResponse(tun, tid);
			}
			// Is there a notification of tunnel closure?
			else if(state == Tunnel::CLOSED) {
				// Schedule tunnel shutdown
				DPRINTI("Received tunnel shutdown request from client");
				scheduleEvent(new ShutdownTunnelEvent(0, this, tid, true));

				#ifdef DEBUG_EVENTS
				std::stringstream nfo;
				nfo << "Scheduling ShutdownTunnelEvent for tunnel id ";
				nfo << tid.toString();
				nfo << " because server tunneler received close request from client";
				g_pLogger->info("##########");
				g_pLogger->info(nfo.str());
				g_pLogger->info("##########");
				#endif
			}
			else if(state == Tunnel::CLOSING) {
				// Treat like a normal connected tunnel for now
				DPRINTV("Peer reports tunnel as CLOSING");
			}

			// Check for reconnection events
			if(state == Tunnel::RECONNECT) {
				// Reconnection for existing tunnel
				DPRINTI("Got explicit reconnection request from client");
				tunnelReconnectionRequest(tun, conn, state);
			}
			else {
				// Check for new connection from proxy
				if(!tun->ownsConnection(conn)) {
					// Treat like a reconnection
					DPRINTI("Got reconnection request, probably from proxy");
					tunnelReconnectionRequest(tun, conn, state);
				}
			}

			// Send back any queued data/acks to the client
			tun->respondToTunnelMsg(false, conn, parms);
		}

		else {
			// Unknown tunnel ID; what state does this message report the tunnel as?
			if(state == Tunnel::CONNECT) {
				// Lookup connection
				std::vector<Connection*>::iterator cit = getConnection(connID);
				Connection* conn = *cit;

				// Connection for a new tunnel
				tun = newTunnelConnectionRequest(conn, new TunnelID(tid), parms);
				if(tun) {
					// Inform tunnel of payload arrival, & send back any queued data/acks
					tun->processTunnelMsg(msg);
					tun->respondToTunnelMsg(false, conn, parms);
				}
				else {
					// Failed to create tunnel
					fatalConnectionError = true;
				}
			}
			else {
				std::stringstream wrn;
				wrn << "Unexpected state " << state << " in msg for unknown tunnel - ";
				wrn << tid.toString();
				g_pLogger->warn(wrn.str());
				fatalConnectionError = true;
			}
		}
	}
	catch(InvalidTunnelID& e) {
		g_pLogger->error("Error parsing tunnel ID: " + e.what());
		fatalConnectionError = true;
	}

	// Did something very serious fail?
	if(fatalConnectionError) {
		#ifdef DEBUG_TUN
		std::stringstream dbg;
		dbg << "Fatal error processing msg on connection " << connID;
		DPRINTV(dbg.str());
		#endif

		// No point keeping this connection around, kill it now
		Threads::MutexVar cmv(m_connectionsMtx);
		std::vector<Connection*>::iterator cit = getConnection(connID);
		sendFatalErrorResponse(*cit, msg);
		destroyConnection(cit);
	}
}


// Notification of new raw data arrival
void ServerTunneler::rawDataArrived(unsigned int connID, const TunnelID& tid,
	const RawBuff& raw) throw(LookupManagementError)
{
	// Lookup tunnel
	Threads::MutexVar tmv(m_tunnelsMtx);
	Tunnel* tun = getTunFromPred(PredTunID(&tid));
	if(tun) {
		// Inform tunnel of raw data arrival
		tun->processRawData(raw);
	}
	else {
		std::stringstream err;
		err << "ServerTunneler::rawDataArrived() - On connection " << connID;
		err << " Unknown tunnel " << tid;
		throw LookupManagementError(err.str());
	}
}


// Check we have received an ack for some sent message
void ServerTunneler::checkAckReceived(void)
{
	std::vector<Tunnel*>::iterator tit;
	Threads::MutexVar tmv(m_tunnelsMtx);

	#ifdef DEBUG_TUN
	if(m_tunnels.size() > 0)
		DPRINTV("ServerTunneler::checkAckReceived()");
	#endif

	// Check acks for all tunnels
	for(tit = m_tunnels.begin(); tit != m_tunnels.end(); ++tit) {
		// Let tunnel check we have received any due acks
		Tunnel* tun = *tit;
		tun->checkAckTimeout();
	}

	// Schedule next ack check event
	scheduleEvent(new CheckAckEvent(ACK_CHECK_PERIOD, this));
}


/***********************/
/* Protected Interface */
/***********************/


// The number of tunnels in existance has changed - Caller holds tunnels mutex
void ServerTunneler::noTunnelsChanged(void)
{
	// Not bothered in server
}


/*********************/
/* Private Interface */
/*********************/


// Create a new carried tunnel in response to a received message - Mutexes held
// by caller
Tunnel* ServerTunneler::newTunnelConnectionRequest(Connection* conn, TunnelID* tid,
	const HTTPMsgParms* parms)
{
	#ifdef DEBUG_TUN
	std::stringstream nfo;
	nfo << "ServerTunneler::newTunnelConnectionRequest() - " << *tid;
	DPRINTI(nfo.str());

	// New connection for new tunnel
	std::string cliVer = parms->getCliVer();
	nfo.str("");
	nfo << "New tunnel connection for new tunnel from client version " << cliVer;
	DPRINTI(nfo.str());
	#endif

	// Create tunnel config
	ServerSideConfig* serverCfg = getServerSideConfigMutable();
	TunnelConfig* tcfg = new TunnelConfig(tid, parms->getTunDest(),
		serverCfg->getAuthenticator()->getDefaultAuthSettings());

	// Create tunnel
	Tunnel* tun = new ServerTunnel(tcfg, this, Tunnel::AUTHREQ);
	tun->addLeftConnection(conn);

	// Store tunnel object in our list of active tunnels
	m_tunnels.push_back(tun);

	return tun;
}

// Reconfigure an existing tunnel in response to a reconnection message
void ServerTunneler::tunnelReconnectionRequest(Tunnel* tun, Connection* conn,
	Tunnel::TunnelState reportedState)
{
	DPRINTI("ServerTunneler::tunnelReconnectionRequest() - " + tun->getTidStr());

	// Let tunnel know it has a new incoming connection
	tun->addLeftConnection(conn);

	// Is it worth keeping this tunnel?
	if(reportedState != Tunnel::CLOSED) {
		// Cancel any pending timeout
		cancelEvents<ShutdownTunnelEvent, TunnelID>(*tun->getConfig()->getTunnelID());

		#ifdef DEBUG_EVENTS
		std::stringstream nfo;
		nfo << "Cancelled any ShutdownTunnelEvent for tunnel ";
		nfo << tun->getTidStr();
		nfo << " because server tunneler got tunnel reconnection";
		g_pLogger->info("##########");
		g_pLogger->info(nfo.str());
		g_pLogger->info("##########");
		#endif
	}
}

// Reconfigure an existing tunnel in response to a new pipe connection
void ServerTunneler::newTunnelPipeConnectionRequest(Tunnel* tun, Connection* conn)
{
	DPRINTI("ServerTunneler::newTunnelPipeConnectionRequest() - " + tun->getTidStr());

	// Check we don't already know about this connection
	if(!tun->ownsConnection(conn)) {
		// Let tunnel know it has a new connection to manage
		tun->addLeftConnection(conn);
	}
}


// Handle an authentication response from a client - Mutexes held by caller
void ServerTunneler::authenticationResponse(Tunnel* tun, const TunnelID& tid)
{
	DPRINTI("ServerTunneler::authenticationResponse()");

	/* Now we have authenticated the connecting client, setup raw connection
	 * out to tunnel destination.
	 */

	// Create raw client connection to handle outgoing traffic
	RawConnection* rawconn = new RawConnection(this, tid);
	addConnection(rawconn);
	tun->addRightConnection(rawconn);

	// Schedule initial connect out raw side
	const TunnelConfig* tcfg = tun->getConfig();
	m_connectionMgr->addProxyConnect(rawconn, 0, tcfg->getDestIP(),
		tcfg->getDestPort(), tcfg->getDestProtocol(), false, "");
}


// Send back a tunnel error response to a client
void ServerTunneler::sendFatalErrorResponse(Connection* conn, const HTTPMsg* msg)
{
	DPRINTV("ServerTunneler::sendFatalErrorResponse()");

	ErrorMsgParms* respParms = 0;

	try {
		// Create suitable msg parameters for response
		const HTTPMsgParms* recvd = msg->getParms();
		respParms = new ErrorMsgParms(recvd);

		// Send error response back to client
		conn->write(respParms);
	}
	catch(TunnelError& e) {
		g_pLogger->warn("Failed to send error response msg: " + e.what());
	}

	// Free up response message parameters
	if(respParms)
		delete respParms;
}


}	// namespace

