/* (C) 2003 Kevin Manning.
 * (C) 2005 Richard Mills / The Linux Academy Ltd.
 *
 * Thread manager class.
 *
 * $Id: ThreadManager.cc,v 1.2 2007-04-17 13:44:05 rich Exp $
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <iostream>
#include <vector>
#include <algorithm>
#include <poll.h>
#include <unistd.h>
#include "ThreadManager.hh"
#include "Callback.hh"
#include "Thread.hh"
#include "MutexVar.hh"

namespace Threads {

// Initialise global data
Mutex ThreadManager::c_mtx;
bool ThreadManager::c_poolIds[MAX_THREADS];


/*****************/
/* ThreadDetails */
/*****************/


// Construction/destruction
ThreadDetails::ThreadDetails(Thread* thread)
: m_thread(thread)
{
}

ThreadDetails::~ThreadDetails()
{
}

// Get/set thread
Thread* ThreadDetails::getThread(void) const
{
	return m_thread;
}
void ThreadDetails::setThread(Thread* thread)
{
	m_thread = thread;
}



/*****************/
/* ThreadManager */
/*****************/


// Construction/Destruction
ThreadManager::ThreadManager(int maxThreads) throw(MutexError)
: m_threadDetails(), m_maxThreads(maxThreads), m_mtx(), m_cv_thread_finished()
{
	#ifdef DEBUG_THREADS
	std::cerr << "ThreadManager::ThreadManager() - max threads ";
	std::cerr << m_maxThreads << std::endl;
	#endif
}

ThreadManager::~ThreadManager()
{
	#ifdef DEBUG_THREADS
	std::cerr << "Entered ThreadManager::~ThreadManager()" << std::endl;
	#endif

	// Wait for all threads to finish
	waitForAll();

	#ifdef DEBUG_THREADS
	std::cerr << "Leaving ThreadManager::~ThreadManager()" << std::endl;
	#endif
}


// Create a new thread and add it to the pool
Thread* ThreadManager::newThread(Callback& callBack) throw(ThreadError)
{
	#ifdef DEBUG_THREADS
	std::cerr << "ThreadManager::newThread()" << std::endl;
	#endif

	// Limit reached ?
	MutexVar mv(m_mtx);
	if(m_threadDetails.size() >= m_maxThreads)
		return 0;

	// Find a free pool id
	unsigned int tc;
	MutexVar cmv(c_mtx);
	for(tc = 0; tc < MAX_THREADS; tc++) {
		if(c_poolIds[tc] == false) {
			c_poolIds[tc] = true;
			break;
		}
	}
	cmv.unlock();

	// Create new thread
	Thread* newThread = new Thread(this, callBack, tc + 256);
	m_threadDetails.push_back(new ThreadDetails(newThread));

	return newThread;
}


/*********************/
/* Private Interface */
/*********************/


// One of our threads has started
void ThreadManager::started(Thread* thread)
{
	#ifdef DEBUG_THREADS
	std::cerr << "ThreadManager::started()" << std::endl;
	#endif
}


// One of our threads has finished
void ThreadManager::finished(Thread* thread)
{
	#ifdef DEBUG_THREADS
	std::cerr << "ThreadManager::finished()" << std::endl;
	#endif

	// Take mutex
	MutexVar mv(m_mtx);

	// Remove thread from list of active threads
	std::vector<ThreadDetails*>::iterator i = find_if(m_threadDetails.begin(),
		m_threadDetails.end(), PredGetDetails(thread));
	if(i != m_threadDetails.end()) {
		delete *i;
		m_threadDetails.erase(i);
	}

	// Signal thread completion to any waiters
	mv.unlock();
	m_cv_thread_finished.broadcast();

	// Free up this threads pool id
	MutexVar cmv(c_mtx);
	unsigned int tc = thread->getThreadCount() - 256;
	c_poolIds[tc] = false;
	cmv.unlock();
}


// Wait for all currently running threads to finish
void ThreadManager::waitForAll(void)
{
	#ifdef DEBUG_THREADS
	std::cerr << "ThreadManager::waitForAll()" << std::endl;
	#endif

	// Take mutex
	MutexVar mv(m_mtx);

	// Still waiting for any threads?
	while(m_threadDetails.size() > 0) {
		#ifdef DEBUG_THREADS
		std::cerr << "ThreadManager::waitForAll() - Waiting for thread" << std::endl;
		#endif

		// Wait for signal thread has finished
		m_cv_thread_finished.wait(mv);
	}
}


}	// namespace Threads

