/* (C) 2006 Richard Mills / The Linux Academy Ltd.
 *
 * Represent a time in seconds & milli-seconds.
 *
 * $Id: Time.cc,v 1.6 2006-12-29 15:08:44 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <iostream>
#include <sstream>
#include <sys/time.h>
#include "Time.hh"


namespace HTTPTunnel {


// Construction/destruction
Time::Time(time_t secs, unsigned int millis)
: m_timeVal()
{
	// Get current time
	struct timeval now;
	gettimeofday(&now, 0);
	time_t s = now.tv_sec;
	suseconds_t m = now.tv_usec;

	// Fill out timeval
	s += secs;
	m += millis * MILLISCALE;
	while(m >= MILLISCALE * MILLISCALE) {
		s += 1;
		m -= MILLISCALE * MILLISCALE;
	}
	m_timeVal.tv_sec = s;
	m_timeVal.tv_usec = m;
}

Time::Time(time_t secs, unsigned int millis, bool abs)
: m_timeVal()
{
	// Fill out timeval
	m_timeVal.tv_sec = secs;
	m_timeVal.tv_usec = (millis * MILLISCALE);
}

Time::Time(const Time& that)
: m_timeVal(that.m_timeVal)
{
}

Time::~Time()
{
}


// Assignment
Time& Time::operator=(const Time& that)
{
	if(this != &that) {
		m_timeVal = that.m_timeVal;
	}

	return *this;
}


// Subtraction
Time Time::operator-(const Time& that) const
{
	// Break into component parts
	time_t thisSecs = getSecs();
	time_t thatSecs = that.getSecs();
	suseconds_t thisMillis = getMillis();
	suseconds_t thatMillis = that.getMillis();

	// Subtract seconds (ignore underflow, that must be >= this)
	thisSecs -= thatSecs;

	// Subtract millis
	if(thatMillis > thisMillis) {
		// Underflow
		thisMillis = 1000 - (thatMillis - thisMillis);
		thisSecs -= 1;
	}
	else {
		thisMillis -= thatMillis;
	}

	// Construct & return result
	return Time(thisSecs, thisMillis, true);
}


// Equality operator
bool Time::operator==(const Time& that) const
{
	return (m_timeVal.tv_sec == that.m_timeVal.tv_sec &&
		m_timeVal.tv_usec == that.m_timeVal.tv_usec);
}

// Less than operator
bool Time::operator<(const Time& that) const
{
	bool res = false;

	if(m_timeVal.tv_sec < that.m_timeVal.tv_sec) {
		res = true;
	}
	else {
		if(m_timeVal.tv_sec == that.m_timeVal.tv_sec) {
			res = (m_timeVal.tv_usec < that.m_timeVal.tv_usec);
		}
	}

	return res;
}


// Get current time
Time Time::now(void)
{
	// Get current time
	struct timeval now;
	gettimeofday(&now, 0);
	return Time(now.tv_sec, (now.tv_usec / MILLISCALE), true);
}


/********************/
/* Public Interface */
/********************/


/*************/
/* Accessors */
/*************/


// Get raw internal time val
const struct timeval& Time::getTimeVal(void) const
{
	return m_timeVal;
}

// Get seconds component
time_t Time::getSecs(void) const
{
	return (m_timeVal.tv_sec);
}

// Get milli-seconds component
unsigned int Time::getMillis(void) const
{
	return (m_timeVal.tv_usec / MILLISCALE);
}

// Get total milli-seconds
unsigned int Time::getTotalMillis(void) const
{
	unsigned int tot = getMillis();
	tot += getSecs() * 1000;
	return tot;
}


/************/
/* Mutators */
/************/


// Subtract some fixed interval
void Time::subInterval(time_t secs, unsigned int millis)
{
	// Subtract seconds
	m_timeVal.tv_sec -= secs;

	// Subtract millis
	unsigned int curMillis = getMillis();
	if(curMillis < millis) {
		// Deal with underflow
		curMillis = 1000 - (millis - curMillis);
		m_timeVal.tv_sec -= 1;
	}
	else {
		curMillis -= millis;
	}
	m_timeVal.tv_usec = curMillis * MILLISCALE;
}


/*************/
/* Utilities */
/*************/


// Has this time elapsed?
bool Time::elapsed(void) const
{
	bool res = true;

	// Get current time
	struct timeval now;
	gettimeofday(&now, 0);

	// Have we passed the current time
	if(m_timeVal.tv_sec > now.tv_sec) {
		// Seconds have not passed
		res = false;
	}
	else {
		if(m_timeVal.tv_sec == now.tv_sec) {
			// Seconds are equal
			if(m_timeVal.tv_usec > now.tv_usec) {
				// Microseconds have not passed
				res = false;
			}
		}
	}

	return res;
}


// Format as a string
std::string Time::toString(void) const
{
	std::stringstream str;
	str << "Secs: " << getSecs() << ", Mili: " << getMillis();
	return str.str();
}


/******************/
/* Helper Methods */
/******************/


// Output to a stream
std::ostream& operator<<(std::ostream& str, const Time& time)
{
	str << time.toString();
	return str;
}


}	// namespace

