/* (C) 2005 Richard Mills / The Linux Academy Ltd.
 *
 * Somewhere to store all information about a unique Tunnel.
 *
 * $Id: Tunnel.cc,v 1.97 2007-08-12 14:42:57 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <vector>
#include <algorithm>
#include <iostream>
#include <sstream>
#include "Enums.hh"
#include "Tunnel.hh"
#include "Tunneler.hh"
#include "Connection.hh"
#include "Logger.hh"
#include "Mutex.hh"
#include "MutexVar.hh"
#include "Debug.hh"
#include "Payload.hh"
#include "GlobalConfig.hh"
#include "ClientSideConfig.hh"
#include "ServerSideConfig.hh"
#include "ProxySettings.hh"
#include "ServerAuthSettings.hh"
#include "TunnelConfig.hh"
#include "TunnelUnviableEvent.hh"
#include "TunnelStateEvent.hh"
#include "ShutdownTunnelEvent.hh"
#include "ConnectionErrorEvent.hh"
#include "PollTimeoutEvent.hh"
#include "UDPSilentEvent.hh"
#include "HTTPMsg.hh"
#include "RawBuff.hh"
#include "HTTPMsgParms.hh"
#include "Authenticator.hh"
#include "Utils.hh"
#include "Server.hh"


// Global references
extern HTTPTunnel::Logger* g_pLogger;
extern HTTPTunnel::GlobalConfig* g_pConfig;


namespace HTTPTunnel {


// Construction/destruction
Tunnel::Tunnel(TunnelConfig* tcfg, Tunneler* tunneler, Direction tunDir,
	TunnelState state)
: m_mtx(), m_leftConn(), m_rightConn(), m_tunCfg(tcfg), m_tid(*tcfg->getTunnelID()),
  m_minAckTimeout(MIN_ACK_TIMEOUT), m_maxAckTimeout(MAX_ACK_TIMEOUT),
  m_ackTimeout(MAX_ACK_TIMEOUT), m_user(), m_pollTimeout(MIN_POLL_TIMEOUT),
  m_tunnelState(state), m_unackedMsgQueue(), m_lastTunnelAnySendTime(Time::now()),
  m_tunnelDataQueue(), m_rawDataQueue(), m_rawDataQueueSize(0), m_oldConns(),
  m_tunneler(tunneler), m_nextSendSeqNo(1), m_nextRecvSeqNo(1), m_resendQueue(),
  m_errorCount(0), m_leftEverConnected(false), m_rightEverConnected(false),
  m_tunnelDirection(tunDir), m_shutdownScheduled(false), m_ackRTTs(), m_avgAckRTT(0),
  m_lastTunnelSendTime(0), m_sentCloseMsg(false), m_lastTunnelConnectedSendTime(Time::now()),
  m_logKey(0), m_lastRawRecvTime(Time::now()), m_lastTunRecvTime(Time::now())
{
	#ifdef DEBUG_TUN
	std::stringstream str;
	str << "Tunnel::Tunnel() - Creating tunnel " << m_tid;
	DPRINTI(str.str());
	#endif

	// Start with default username
	Authenticator* authenticator = getAuthenticator();
	m_user = authenticator->getDefaultUserName();

	// What type of data are we tunneling?
	if(m_tid.getProtocol() == Enums::UDP) {
		// Start running UDP silent timeout events
		m_tunneler->scheduleEvent(new UDPSilentEvent(m_tunneler, m_tid));
	}
}

Tunnel::~Tunnel()
{
	DPRINTI("Tunnel::~Tunnel()");

	Threads::MutexVar mv(m_mtx);

	// Tidy up all remaining connections
	std::vector<Connection*>::iterator cit;
	for(cit = m_leftConn.begin(); cit != m_leftConn.end(); ++cit) {
		// Tell connection it no longer carries us
		Connection* conn = *cit;
		conn->removeCarriedTunnel(this);
	}
	for(cit = m_rightConn.begin(); cit != m_rightConn.end(); ++cit) {
		// Tell connection it no longer carries us
		Connection* conn = *cit;
		conn->removeCarriedTunnel(this);
	}

	// Free up data queues
	while(m_tunnelDataQueue.size() > 0) {
		std::vector<Payload*>::iterator it = m_tunnelDataQueue.begin();
		delete *it;
		m_tunnelDataQueue.erase(it);
	}
	while(m_unackedMsgQueue.size() > 0) {
		std::vector<SentMsgParms*>::iterator it = m_unackedMsgQueue.begin();
		delete *it;
		m_unackedMsgQueue.erase(it);
	}
	while(m_rawDataQueue.size() > 0) {
		std::vector<RawBuff*>::iterator it = m_rawDataQueue.begin();
		delete *it;
		m_rawDataQueue.erase(it);
	}

	// Free config
	delete m_tunCfg;
}


/*************/
/* Selectors */
/*************/


// Select only left side connections
void Tunnel::SelLeftConn::operator()(std::vector<Connection*>& sel)
{
	// Move all right side connections to the end of the list
	std::vector<Connection*>::iterator leftend;
	leftend = std::remove_if(sel.begin(), sel.end(), PredIsRight(m_tun));
	// Erase all right side connections
	sel.erase(leftend, sel.end());
}

// Select only right side connections
void Tunnel::SelRightConn::operator()(std::vector<Connection*>& sel)
{
	// Move all left side connections to the end of the list
	std::vector<Connection*>::iterator rightend;
	rightend = std::remove_if(sel.begin(), sel.end(), PredIsLeft(m_tun));
	// Erase all left side connections
	sel.erase(rightend, sel.end());
}

// Sort according to most recently first connected connection
void Tunnel::SelRecentConn::operator()(std::vector<Connection*>& sel)
{
	// Sort according to connection time, then reverse
	std::sort(sel.begin(), sel.end(), Connection::PredConnectTime());
	std::reverse(sel.begin(), sel.end());
}

// Exclude the specified connection
void Tunnel::SelNotConn::operator()(std::vector<Connection*>& sel)
{
	// Remove all occurances of specified connection
	sel.erase(std::remove(sel.begin(), sel.end(), m_conn), sel.end());
}

// Select only connected connections
void Tunnel::SelConnectedConn::operator()(std::vector<Connection*>& sel)
{
	// Move all non-connected connections to the end of the list
	std::vector<Connection*>::iterator end;
	end = std::remove_if(sel.begin(), sel.end(), std::not1(Connection::PredConnected()));
	// Erase all non-connected connections
	sel.erase(end, sel.end());
}


/*************/
/* Accessors */
/*************/


// Get our config
const TunnelConfig* Tunnel::getConfig(void) const
{
	return m_tunCfg;
}

// Get our proxy settings for modification
ProxySettings* Tunnel::getProxySettings(void)
{
	return m_tunCfg->getProxySettings();
}

// Get our serverauth settings for modification
ServerAuthSettings* Tunnel::getServerAuthSettings(void)
{
	return m_tunCfg->getServerAuthSettings();
}


// Get our tunnel ID as a string
std::string Tunnel::getTidStr(void) const
{
	return m_tid.toString();
}

// Get our owning tunneler
Tunneler* Tunnel::getTunneler(void)
{
	return m_tunneler;
}

// Does the specified connection belong to this Tunnel?
bool Tunnel::ownsConnection(const Connection* conn)
{
	Threads::MutexVar mv(m_mtx);

	// Need to check each managed list
	bool res = (leftConnection(conn) || rightConnection(conn));
	return res;
}

// Is the specified connection from the left side?
bool Tunnel::leftConnection(const Connection* conn)
{
	Threads::MutexVar mv(m_mtx);

	bool res = (std::find(m_leftConn.begin(), m_leftConn.end(), conn) !=
		m_leftConn.end());
	return res;
}

// Is the specified connection from the right side?
bool Tunnel::rightConnection(const Connection* conn)
{
	Threads::MutexVar mv(m_mtx);

	bool res = (std::find(m_rightConn.begin(), m_rightConn.end(), conn) !=
		m_rightConn.end());
	return res;
}

// Get the number of left side connections for this tunnel
unsigned int Tunnel::noLeftConnections(void)
{
	Threads::MutexVar mv(m_mtx);
	return m_leftConn.size();
}

// Get the number of right side connections for this tunnel
unsigned int Tunnel::noRightConnections(void)
{
	Threads::MutexVar mv(m_mtx);
	return m_rightConn.size();
}


// Get a single connection matching the provided criteria
Connection* Tunnel::getConnFromSel(std::vector<Selector*>& crit)
{
	// Get all matching connections
	std::vector<Connection*> allConns;
	getConnsFromSel(crit, allConns);

	// Resulting connection is the one at the head of the remaining list
	if(allConns.size() > 0)
		return allConns[0];

	return 0;
}

// Get all connections matching the provided criteria
void Tunnel::getConnsFromSel(std::vector<Selector*>& crit,
	std::vector<Connection*>& res)
{
	// Get list of all connections
	std::back_insert_iterator<std::vector<Connection*> > dest(res);

	Threads::MutexVar mv(m_mtx);
	std::vector<Connection*>::iterator first = m_leftConn.begin();
	std::vector<Connection*>::iterator last = m_leftConn.end();
	std::copy(first, last, dest);
	first = m_rightConn.begin();
	last = m_rightConn.end();
	std::copy(first, last, dest);
	mv.unlock();

	// Apply all selection criteria (freeing them as we go)
	std::vector<Selector*>::iterator sit;
	for(sit = crit.begin(); sit != crit.end(); ++sit) {
		(*sit)->operator()(res);
		delete *sit;
	}
}


// Get current poll timeout
unsigned int Tunnel::getPollTimeout(void) const
{
	return m_pollTimeout;
}

// Get name of owning user
std::string Tunnel::getUserName(void) const
{
	return m_user;
}

// Get source addresses of connections for this tunnel
std::string Tunnel::getSrcConnectionAddresses(void)
{
	// Only know in the server
	return "Unknown";
}

// Get destination addresses of connections for this tunnel
std::string Tunnel::getDstConnectionAddresses(void)
{
	// Only know in the server
	return "Unknown";
}

// Get logging key
unsigned int Tunnel::getLogKey(void) const
{
	return m_logKey;
}

// Get our current state
Tunnel::TunnelState Tunnel::getState(void) const
{
	return m_tunnelState;
}

// Have we not received any raw data for a suspiciously long time?
bool Tunnel::rawRecvTimeout(void)
{
	Threads::MutexVar mv(m_mtx);
	Time quiet = Time::now() - m_lastRawRecvTime;
	return (quiet.getSecs() >= MAX_NO_RECV_TIME);
}

// Have we not received any tunnel data for a suspiciously long time?
bool Tunnel::tunRecvTimeout(void)
{
	Threads::MutexVar mv(m_mtx);
	Time quiet = Time::now() - m_lastTunRecvTime;
	return (quiet.getSecs() >= MAX_NO_RECV_TIME);
}


/************/
/* Mutators */
/************/


// Add connection to list of left side connections
void Tunnel::addLeftConnection(Connection* conn)
{
	Threads::MutexVar mv(m_mtx);

	// Already know about this connection?
	std::vector<Connection*>::iterator cit;
	cit = std::find(m_leftConn.begin(), m_leftConn.end(), conn);
	if(cit == m_leftConn.end()) {
		// Don't re-add old removed connections
		if(!oldConnection(conn)) {
			m_leftConn.push_back(conn);

			#ifdef DEBUG_TUN
			std::stringstream dbg;
			dbg << "Tunnel::addLeftConnection() - connection " << conn->getID();
			DPRINTV(dbg.str());
			#endif
		}
	}

	// Connected?
	m_leftEverConnected = conn->everConnected();
	mv.unlock();

	// Also inform connection about us
	conn->addCarriedTunnel(this);
}

// Add connection to list of right side connections
void Tunnel::addRightConnection(Connection* conn)
{
	Threads::MutexVar mv(m_mtx);

	// Already know about this connection?
	std::vector<Connection*>::iterator cit;
	cit = std::find(m_rightConn.begin(), m_rightConn.end(), conn);
	if(cit == m_rightConn.end()) {
		// Don't re-add old removed connections
		if(!oldConnection(conn)) {
			m_rightConn.push_back(conn);

			#ifdef DEBUG_TUN
			std::stringstream dbg;
			dbg << "Tunnel::addRightConnection() - connection " << conn->getID();
			DPRINTV(dbg.str());
			#endif
		}
	}

	// Connected?
	m_rightEverConnected = conn->everConnected();
	mv.unlock();

	// Also inform connection about us
	conn->addCarriedTunnel(this);
}


// Remove closing connection from this tunnel
void Tunnel::removeConnection(Connection* conn)
{
	#ifdef DEBUG_TUN
	std::stringstream dbg;
	dbg << "Tunnel::removeConnection() - connection " << conn->getID();
	if(leftConnection(conn)) dbg << " left";
	else if(rightConnection(conn)) dbg << " right";
	DPRINTV(dbg.str());
	#endif

	bool removed = false;

	Threads::MutexVar mv(m_mtx);

	// Left connection?
	std::vector<Connection*>::iterator it;
	it = std::find(m_leftConn.begin(), m_leftConn.end(), conn);
	if(it != m_leftConn.end()) {
		// Remove it
		m_leftConn.erase(it);
		removed = true;

		// Are we still a viable tunnel?
		if(m_leftConn.size() == 0) {
			DPRINTI(getTidStr() + " - Tunnel no longer viable");
			recordConnectionError(UNVIABLE_ERROR);
			m_tunneler->scheduleEvent(new TunnelUnviableEvent(m_tunneler, m_tid));
		}
	}
	else {
		// Right connection?
		it = std::find(m_rightConn.begin(), m_rightConn.end(), conn);
		if(it != m_rightConn.end()) {
			// Remove it
			m_rightConn.erase(it);
			removed = true;

			// Are we still a viable tunnel?
			if(m_rightConn.size() == 0) {
				DPRINTI(getTidStr() + " - Tunnel no longer viable");
				recordConnectionError(UNVIABLE_ERROR);
				m_tunneler->scheduleEvent(new TunnelUnviableEvent(m_tunneler, m_tid));
			}
		}
	}

	// Remember connections we've dropped
	if(removed && !oldConnection(conn)) {
		m_oldConns.push_back(conn->getID());
	}

#if 0
	// Are we still a viable tunnel?
	if(m_leftConn.size() == 0 || m_rightConn.size() == 0) {
		DPRINTI(getTidStr() + " - Tunnel no longer viable");
		recordConnectionError(UNVIABLE_ERROR);
		m_tunneler->scheduleEvent(new TunnelUnviableEvent(m_tunneler, m_tid));
	}
#endif
}


// A connection that carries us has connected
void Tunnel::connectionConnected(Connection* conn)
{
	#ifdef DEBUG_TUN
	std::stringstream dbg;
	dbg << "Tunnel::connectionConnected() - connection " << conn->getID();
	DPRINTV(dbg.str());
	#endif

	Threads::MutexVar mv(m_mtx);

	// Remember that we have connected
	if(leftConnection(conn))
		m_leftEverConnected = true;
	else if(rightConnection(conn))
		m_rightEverConnected = true;

	// Maybe log this new connection
	if(m_tunnelState == CONNECTED) {
		Authenticator* authenticator = getAuthenticator();
		m_logKey = authenticator->logTunnel(this);
	}
}


// Decrease leaky bucket error count
void Tunnel::decreaseErrorCount(void)
{
	Threads::MutexVar mv(m_mtx);

	#ifdef DEBUG_TUN
	unsigned int oldCount = m_errorCount;
	#endif

	// Once we've reached the fatal error level, don't leak anymore
	if(m_errorCount < FATAL_ERROR_COUNT) {
		// Leak, but check for going negative
		if(m_errorCount > ERROR_COUNT_DECREASE)
			m_errorCount = m_errorCount - ERROR_COUNT_DECREASE;
		else
			m_errorCount = 0;
	}

	#ifdef DEBUG_TUN
	if(oldCount > 0) {
		std::stringstream nfo;
		nfo << getTidStr() << " old errorCount: " << oldCount << " new: " << m_errorCount;
		DPRINTI(nfo.str());
	}
	#endif
}

// Record connection error
void Tunnel::recordConnectionError(TunnelErrorType errType)
{
	// How much to increase error level by
	unsigned int errLevelInc = 0;

	// Do we consider this to be an error from the other tunnel end?
	bool otherEndError = false;

	// What type of error?
	switch(errType) {
		case(GENERAL_COMMS_ERROR): {
			errLevelInc = 5;
			break;
		}
		case(PROXY_AUTH_ERROR):
		case(SERVER_AUTH_ERROR): {
			// Fatal error
			errLevelInc = FATAL_ERROR_COUNT;
			otherEndError = true;

			// Let user know
			std::stringstream err;
			if(errType == PROXY_AUTH_ERROR) {
				err << "Proxy authentication error: Check proxy username & password ";
				err << "are correct";
			}
			else if(errType == SERVER_AUTH_ERROR) {
				err << "Server authentication error: Check cliauth username & password ";
				err << "are correct and you are not exceeding any quotas for your account";
			}
			g_pLogger->error(err.str());
			break;
		}
		case(BAD_RESPONSE_ERROR): {
			// Fatal error
			errLevelInc = FATAL_ERROR_COUNT;
			otherEndError = true;
			break;
		}
		case(MSG_RESEND_ERROR): {
			// Not an error until we exceed maximum resend count
			break;
		}
		case(MAX_RESEND_ERROR): {
			errLevelInc = FATAL_ERROR_COUNT;
			break;
		}
		case(CONNECTION_TIMEOUT_ERROR): {
			errLevelInc = 5;
			break;
		}
		case(CONNECTION_CLOSED_ERROR): {
			// Not necessarily an error
			break;
		}
		case(UNVIABLE_ERROR): {
			// Are we still managing to send data down the tunnel?
			Time connSendDiff = Time::now() - m_lastTunnelConnectedSendTime;
			Time anySendDiff = Time::now() - m_lastTunnelAnySendTime;
			if((unsigned int)connSendDiff.getSecs() >= UNVIABLE_ERROR_TIME) {
				errLevelInc = 2;
				if((unsigned int)anySendDiff.getSecs() >= UNVIABLE_ERROR_TIME)
					errLevelInc += 5;
			}
			break;
		}
		case(QUOTAS_EXCEEDED_ERROR): {
			errLevelInc = 2;
			break;
		}
		case(FATAL_QUOTAS_ERROR): {
			errLevelInc = FATAL_ERROR_COUNT;
			break;
		}
		default: {
			g_pLogger->warn("Unknown tunnel error type");
			break;
		}
	}

	// Increase error count
	Threads::MutexVar mv(m_mtx);
	m_errorCount += errLevelInc;

	#ifdef DEBUG_TUN
	std::stringstream dbg;
	dbg << "Tunnel::recordConnectionError() - " << getTidStr();
	dbg << " count " << m_errorCount;
	DPRINTV(dbg.str());
	#endif

	// Are we too broken to continue?
	if(errorLevelExceeded() && !m_shutdownScheduled) {
		// Schedule our own shutdown
		DPRINTV("Fatal error level exceeded, scheduling our own shutdown");

		#ifdef DEBUG_EVENTS
		std::stringstream nfo;
		nfo << "Scheduling ShutdownTunnelEvent for tunnel " << m_tid;
		nfo << " because tunnel fatal error level exceeded";
		g_pLogger->info("##########");
		g_pLogger->info(nfo.str());
		g_pLogger->info("##########");
		#endif

		m_tunneler->scheduleEvent(new ShutdownTunnelEvent(0, m_tunneler, m_tid, otherEndError));

		// Set flag so we don't schedule repeated shutdowns
		m_shutdownScheduled = true;
	}
}


// Notification we are closing
void Tunnel::closing(void)
{
	Threads::MutexVar mv(m_mtx);
	m_tunnelState = CLOSING;
}

// Notification that we are trying to reconnect this tunnel
void Tunnel::reconnecting(void)
{
	Threads::MutexVar mv(m_mtx);
	m_tunnelState = RECONNECT;
}


/***********/
/* Methods */
/***********/


// Process some newly arrived tunnel msg from one of our connections
void Tunnel::processTunnelMsg(const HTTPMsg* msg)
{
	// Get msg parameters
	const HTTPMsgParms* parms = msg->getParms();

	#ifdef DEBUG_TUN
	std::stringstream dbg;
	dbg << "Processing msg seq no: ";
	dbg << parms->getPayload()->getSeqNo() << " state: " << stateAsString(parms->getState());
	if(parms->getAck().size() > 0)
		dbg << " ack: " << parms->getAck();
	DPRINTV(dbg.str());
	#endif

	// Get payload sequence number
	SeqNo seqno = parms->getPayload()->getSeqNo();

	Threads::MutexVar mv(m_mtx);

	// If we actually have any payload data, update data received time
	if(parms->getPayload()->getPayloadSize() > 0)
		m_lastTunRecvTime = Time::now();

	// Check this isn't a duplicate msg
	if(!duplicateTunnelMsg(seqno)) {
		// Take a copy of the payload
		Payload* payload = new Payload(*parms->getPayload());

		// Check quotas
		Authenticator* authenticator = getAuthenticator();
		if(!authenticator->getUserCanRecvBytes(m_user, payload->getPayloadSize())) {
			DPRINTV("Quota exceeded for received msg");
			recordConnectionError(QUOTAS_EXCEEDED_ERROR);
		}

		// Insert payload into queue
		std::vector<Payload*>::iterator pos;
		pos = std::upper_bound(m_tunnelDataQueue.begin(), m_tunnelDataQueue.end(),
			payload, CmpPayloadsSeqNo());
		m_tunnelDataQueue.insert(pos, payload);
	}

	// Process ack from this msg
	std::string ack = parms->getAck();
	if(ack.size() > 0)
		processAck(ack);

	// Check whether we are using header fields for tunneling
	if(!parms->getUseHeaderFields()) {
		m_tunCfg->setUseHeaders(false);
	}

	// Advance our state machine according to this msg
	advanceState(parms);

	// Try sending any received data on to the destination
	sendRawMsgs();
}


// Process some newly extracted raw data from one of our connections
void Tunnel::processRawData(const RawBuff& raw)
{
	Threads::MutexVar mv(m_mtx);

	// Got any data to save?
	unsigned int rawLen = raw.getLength();
	if(rawLen > 0) {
		try {
			// Take a copy of raw data packet
			RawBuff* rawCopy = new RawBuff(raw);

			// Append data to queue
			m_rawDataQueue.push_back(rawCopy);
			m_rawDataQueueSize += rawLen;

			// Check if we need to block any connections
			checkBlockedConnections();
		}
		catch(MemoryError& e) {
			g_pLogger->error("Failed to process raw data arrival: " + e.what());
		}
	}

	// Update data received time
	m_lastRawRecvTime = Time::now();
}


// Determine whether we have experienced too many errors to continue
bool Tunnel::errorLevelExceeded(Direction dir)
{
	Threads::MutexVar mv(m_mtx);

	// Give up if we failed to connect out in the indicated direction
	bool res = !((dir == LEFT)? m_leftEverConnected : m_rightEverConnected);

	if(!res) {
		// Check direction agnostic errors
		res = errorLevelExceeded();
	}
	else
		DPRINTV("Too many errors; tunnel never connected");

	return res;
}

// Determine whether we have experienced too many errors, ignoring direction
bool Tunnel::errorLevelExceeded(void)
{
	// Leaky error count exceeded?
	bool res = (m_errorCount >= FATAL_ERROR_COUNT);
	return res;
}


// Send a message down the tunnel
void Tunnel::sendTunnelMsg(bool mgmt, bool retry, Connection* conn)
{
	DPRINTV("Tunnel::sendTunnelMsg() - " + m_tid.toString());

	Threads::MutexVar mv(m_mtx);

	// Check supplied connection for suitability
	conn = getTunnelConnection(conn);
	if(conn) {
		// Clear to send new message?
		if(clearToSendNewMsg(conn)) {
			try {
				// Build parameters for this message
				HTTPMsgParms* parms = makeTunnelMsg(0, MAX_NORMAL_HTTP_PAYLOAD,
					retry);

				// Try to send it (takes ownership of parms)
				sendTunnelMsgWithParms(conn, parms);
			}
			catch(TunnelError& e) {
				g_pLogger->error("Failed to send new tunnel msg: " + e.what());
			}
		}
	}
	else {
		DPRINTI("No suitable connection to send msg down found");
	}
}


// Send a response to a tunnel message
void Tunnel::respondToTunnelMsg(bool mgmt, Connection* conn, const HTTPMsgParms* recvd)
{
	DPRINTV("Tunnel::respondToTunnelMsg() - " + m_tid.toString());

	Threads::MutexVar mv(m_mtx);

	// Check supplied connection for suitability
	conn = getTunnelConnection(conn);
	if(conn) {
		// Clear to send new message?
		if(clearToSendNewMsg(conn)) {
			try {
				// Create parameters for new msg
				HTTPMsgParms* parms = makeTunnelMsg(recvd, MAX_NORMAL_HTTP_PAYLOAD,
					false);

				// Try to send it (takes ownership of parms)
				sendTunnelMsgWithParms(conn, parms);
			}
			catch(TunnelError& e) {
				g_pLogger->error("Failed to respond to tunnel msg: " + e.what());
			}
		}
	}
	else {
		DPRINTI("No suitable connection to send msg down found");
	}
}


// Poll the far end of the tunnel
void Tunnel::pollTunnel(void)
{
	DPRINTV("Tunnel::pollTunnel() - " + m_tid.toString());

	Threads::MutexVar mv(m_mtx);

	// Only send poll msg if we've been quiet
	Time quietTime = Time::now() - m_lastTunnelAnySendTime;
	if(quietTime.getTotalMillis() >= (m_pollTimeout - 5)) {
		#ifdef DEBUG_TUN
		std::stringstream dbg;
		dbg << "Last msg sent " << quietTime << " ago, timeout = ";
		dbg << m_pollTimeout << " - polling";
		DPRINTV(dbg.str());
		#endif

		// Force the sending of a msg
		sendTunnelMsg(false, false, 0);
	}

	// Schedule next poll timeout
	m_tunneler->scheduleEvent(new PollTimeoutEvent(Time(0, m_pollTimeout),
		m_tunneler, m_tid));
}


// Send out as many raw messages as we have queued tunnel payload for
void Tunnel::sendRawMsgs(void)
{
	DPRINTV("Tunnel::sendRawMsgs()");

	// Get a connection in this tunnel going the right way
	std::vector<Tunnel::Selector*> filt;
	if(m_tunnelDirection == LEFT)
		filt.push_back(new SelRightConn(this));
	else
		filt.push_back(new SelLeftConn(this));
	filt.push_back(new SelConnectedConn());

	Threads::MutexVar mv(m_mtx);
	Connection* conn = getConnFromSel(filt);

	try {
		// Send as much queued tunneled data as we can
		bool keepGoing = true;
		while(keepGoing) {
			// Get next raw message payload
			Payload* raw = makeRawPayload();
			if(raw) {
				// Create message parameters
				MsgParms parms(getTidStr(), raw, false);

				// Do we have a connection to send this out on?
				if(conn) {
					#ifdef DEBUG_TUN
					std::stringstream dbg;
					dbg << "Sending raw msg down connection id " << conn->getID();
					DPRINTV(dbg.str());
					#endif

					// Write data to connection
					conn->write(&parms);

					// Successfully sent message
					sentRawPayload(true);
				}
				else {
					DPRINTI("No suitable connection to send raw msg down found");
					keepGoing = false;

					// If we're just waiting to close down there's no point keeping payload
					if(m_tunnelState == CLOSING) {
						sentRawPayload(false);
					}
				}
			}
			else {
				// No more tunneled data to send
				keepGoing = false;
			}
		}
	}
	catch(WriteError& e) {
		// Error on socket, schedule error on this connection
		DPRINTW("Write failure sending raw msg: " + e.what());
		if(!conn->closing()) {
			m_tunneler->scheduleEvent(new ConnectionErrorEvent(conn->getID(),
				GENERAL_COMMS_ERROR, m_tunneler));
		}
	}
	catch(TunnelError& e) {
		// Failed to send raw msg
		DPRINTW("General failure sending raw msg: " + e.what());
	}
}


// Check ack timeout on all unacknowledged messages
bool Tunnel::checkAckTimeout(void)
{
	bool res = false;

	Threads::MutexVar mv(m_mtx);

	#ifdef DEBUG_TUN
	std::stringstream dbg;
	dbg << "Tunnel::checkAckTimeout() - Ack timeout currently: " << m_ackTimeout;
	DPRINTV(dbg.str());
	#endif

	// If this tunnel has been silent for too long, start worrying
	Time quietTime = Time::now() - m_lastTunnelAnySendTime;
	if(quietTime.getSecs() > MAX_SILENCE_TIME) {
		DPRINTV(m_tid.toString() + " - very quiet");
		recordConnectionError(GENERAL_COMMS_ERROR);
	}

	// Iterate over unacked msg queue
	time_t now = time(0);
	std::vector<SentMsgParms*>::iterator pit;
	for(pit = m_unackedMsgQueue.begin(); pit != m_unackedMsgQueue.end(); ++pit) {
		SentMsgParms* sentMsgParms = *pit;

		// How long has this msg been unacknowledged?
		time_t unackTime = now - sentMsgParms->m_sentTime;
		if(unackTime >= m_ackTimeout) {
			// Ack timeout exceeded
			SeqNo seqNo = sentMsgParms->m_parms->getPayload()->getSeqNo();

			#ifdef DEBUG_TUN
			std::stringstream nfo;
			nfo << "Need to resend msg seq no " << seqNo << " for tunnel " << m_tid;
			DPRINTI(nfo.str());
			#endif

			// Remember to resend this msg
			needToResendMsg(seqNo);
			res = true;
		}
	}

	return res;
}


// Are we in a state where we can be shutdown?
bool Tunnel::canBeShutdown(void)
{
	Threads::MutexVar mv(m_mtx);

	#ifdef DEBUG_TUN
	std::stringstream cbstr;
	#endif

	bool res = false;

	if(errorLevelExceeded(LEFT) || errorLevelExceeded(RIGHT)) {
		// Tunnel is errored, so close
		res = true;

		#ifdef DEBUG_TUN
		cbstr << "Errored";
		#endif
	}
	else if(m_sentCloseMsg) {
		// We've sent final close msg to other end, so close
		res = true;

		#ifdef DEBUG_TUN
		cbstr << "SentClose";
		#endif
	}
	else if(time(0) - m_lastTunnelSendTime >= MAX_ACK_TIMEOUT) {
		// Time since last msg sent exceeds limit, so close
		res = true;

		#ifdef DEBUG_TUN
		cbstr << "SendTimeout";
		#endif
	}

	#ifdef DEBUG_TUN
	std::stringstream dbg;
	dbg << "Tunnel::canBeShutdown() - " << getTidStr() << ": " << res;
	if(res) dbg << " " << cbstr.str();
	DPRINTV(dbg.str());
	#endif

	return res;
}


/***********************/
/* Protected Interface */
/***********************/


// Create complete HTTP msg parameters for sending down tunnel - Caller hold mutex
HTTPMsgParms* Tunnel::makeTunnelMsg(const HTTPMsgParms* recvd,
	unsigned int maxPayload, bool retry) throw(MemoryError)
{
	// Tunnel state we will report in this msg
	TunnelState reportedState = m_tunnelState;

	// Get authenticator
	Authenticator* authenticator = getAuthenticator();

	// Can we send any actual payload yet?
	if(m_tunnelState == CONNECTED || m_tunnelState == RECONNECT ||
	m_tunnelState == CLOSING) {
		// Check quotas for allowed payload size
		unsigned int quotaMaxPayload = authenticator->getUserBytesCanSend(m_user);
		if(quotaMaxPayload < maxPayload)
			maxPayload = quotaMaxPayload;
	}
	else {
		// Can't send any payload just now
		maxPayload = 0;
	}

	// Get some queued raw data to make up payload
	RawBuff payloadData;
	bool splitPayload = false;
	makeTunnelPayload(payloadData, splitPayload, maxPayload);

	// Create actual payload
	Payload* payload = new Payload(m_nextSendSeqNo, payloadData, splitPayload);

	// Are we trying to shutdown?
	if(m_tunnelState == CLOSING) {
		// Is this the final msg we will send before closing?
		if(payload->getPayload().getLength() == m_rawDataQueueSize) {
			m_tunnelState = CLOSED;
			reportedState = CLOSED;
		}
		else {
			// Don't let the other side know just yet
			reportedState = CONNECTED;

			#ifdef DEBUG_TUN
			std::stringstream dbg;
			dbg << "Not moving to CLOSED, payload len = ";
			dbg << payload->getPayload().getLength() << " remaining raw data len = ";
			dbg << m_rawDataQueueSize;
			DPRINTV(dbg.str());
			#endif
		}
	}

	// Make msg parms
	User userDetails = authenticator->getUserDetails(m_user);
	HTTPMsgParms* parms = new HTTPMsgParms(getTidStr(), payload, retry,
		getProxySettings(), getServerAuthSettings(), userDetails, reportedState,
		m_tunCfg->getHTTPVersion(), m_nextRecvSeqNo.toString(),
		m_tunCfg->getDestAddress(), m_tunCfg->getUseHeaders(), this);

	// Do we have a URI to respond to?
	if(recvd) {
		parms->setURI(recvd->getURI());
	}

	return parms;
}


// Successfully sent a new message down the tunnel - Caller hold mutex
void Tunnel::sentNewTunnelMsg(HTTPMsgParms* parms, bool remember)
{
	// Free sent raw data
	sentTunnelPayload(parms);

	// Update quota tracking
	unsigned int payloadLen = parms->getPayload()->getPayloadSize();
	Authenticator* authenticator = getAuthenticator();
	authenticator->setUserSentBytes(m_user, payloadLen);

	// Update last sent tunnel msg time
	m_lastTunnelSendTime = time(0);
	m_lastTunnelAnySendTime = Time::now();
	TunnelState msgTunState = parms->getState();
	if(msgTunState == CONNECTED)
		m_lastTunnelConnectedSendTime = Time::now();

	// Do we want to remember about this?
	if(remember) {
		// Did this msg contain a final close for this tunnel?
		if(msgTunState == CLOSED) {
			m_sentCloseMsg = true;

			// We won't be requiring an ack for the final close
			delete parms;
		}
		else {
			// Add paramaters to queue of paramaters waiting for an ack
			m_unackedMsgQueue.push_back(new SentMsgParms(parms));
		}

		// Increment next sent sequence number
		incSendSeqNo();
	}
	else {
		// We're going to pretend this msg hasn't been sent at all
		delete parms;
	}
}


// Scale poll timeout - Caller hold mutex
void Tunnel::scalePollTimeout(void)
{
	#ifdef DEBUG_TUN
	unsigned int oldPollTimeout = m_pollTimeout;
	#endif

	float scale = (float)(m_ackTimeout - m_minAckTimeout) /
		(float)(m_maxAckTimeout - m_minAckTimeout);
	float range = (MAX_POLL_TIMEOUT - MIN_POLL_TIMEOUT) * scale;
	m_pollTimeout = (int)(MIN_POLL_TIMEOUT + range);

	#ifdef DEBUG_TUN
	if(oldPollTimeout != m_pollTimeout) {
		std::stringstream nfo;
		nfo << "Tunnel::scalePollTimeout() - Now " << m_pollTimeout;
		nfo << " for tunnel " << getTidStr();
		DPRINTV(nfo.str());
	}
	#endif
}


// Get authenticator
Authenticator* Tunnel::getAuthenticator(void)
{
	return m_tunneler->getServerMutable()->getConfigMutable()->getAuthenticator();
}


// Have we got a stable HTTP connection to other end?
bool Tunnel::httpStable(void) const
{
	return true;
}


// Look for unacked management msgs we want to try resending - Caller holds mutex
void Tunnel::checkUnackedMgmtMsgs(void)
{
	// Check all currently unacked msgs
	std::vector<SentMsgParms*>::iterator sit;
	for(sit = m_unackedMsgQueue.begin(); sit != m_unackedMsgQueue.end(); ++sit) {
		// Get msg details
		SentMsgParms* sentMsgParms = *sit;
		HTTPMsgParms* msgParms = sentMsgParms->m_parms;

		// If the HTTP session is not yet stable, allow resending of mgmt msgs early
		if(isMgmtMsg(msgParms) && !httpStable()) {
			// Resend this mgmt msg
			SeqNo seqNo = msgParms->getPayload()->getSeqNo();

			#ifdef DEBUG_TUN
			std::stringstream nfo;
			nfo << "Allowing resend of unacked mgmt msg seq no " << seqNo << " ";
			nfo << "for tunnel " << m_tid;
			DPRINTI(nfo.str());
			#endif

			// Update server auth settings incase they have changed
			msgParms->setServerAuthSettings(getServerAuthSettings());
			Authenticator* authenticator = getAuthenticator();
			msgParms->setUser(authenticator->getUserDetails(m_user));

			// Remember to resend this msg
			needToResendMsg(seqNo);
		}
	}
}

// Resend all flagged unacknowledged msgs - Caller holds mutex
bool Tunnel::resendUnackedMsgs(Connection* conn) throw(ResendError)
{
	bool resentMsg = false;

	// Got any msgs to resend?
	while(m_resendQueue.size() > 0) {
		// Get sequence number of next msg to resend
		std::vector<SeqNo>::iterator sit = m_resendQueue.begin();
		SeqNo seqNo = *sit;

		#ifdef DEBUG_TUN
		std::stringstream dbg;
		dbg << "Tunnel::resendUnackedMsgs() - Considering resend of seq no ";
		dbg << seqNo << " for tunnel " << m_tid;
		dbg << " down connection " << conn->getID();
		DPRINTV(dbg.str());
		#endif

		// Look for msg paramaters in unacked msg queue
		std::vector<SentMsgParms*>::iterator pit;
		pit = std::find_if(m_unackedMsgQueue.begin(), m_unackedMsgQueue.end(),
			CmpSentParmsSeqNo(seqNo));

		// Ack still outstanding?
		if(pit != m_unackedMsgQueue.end()) {
			// Try to resend msg
			SentMsgParms* sentMsgParms = *pit;
			HTTPMsgParms* msgParms = sentMsgParms->m_parms;
			resentMsg = true;

			// Check we haven't already tried to resend this msg too many times
			if(sentMsgParms->m_resentCount >= MAX_MSG_RESEND) {
				recordConnectionError(MAX_RESEND_ERROR);
				throw ResendError("Resend count exceeded for msg");
			}

			// Update acks to current value
			msgParms->setAck(m_nextRecvSeqNo.toString());

			#ifdef DEBUG_TUN
			std::stringstream dbg;
			dbg << "Tunnel::resendUnackedMsgs() - Attempting resend";
			dbg << " with ack: " << msgParms->getAck();
			DPRINTV(dbg.str());
			#endif

			try {
				// Record this as an error on this tunnel
				if(httpStable())
					recordConnectionError(MSG_RESEND_ERROR);

				// Write data to connection
				conn->write(msgParms);

				// Remove resent sequence number from queue
				m_resendQueue.erase(sit);

				// Update sent time
				m_lastTunnelAnySendTime = Time::now();
				sentMsgParms->m_sentTime = time(0);
				sentMsgParms->m_resentCount++;
			}
			catch(TunnelError& e) {
				// Error resending msg
				recordConnectionError(MSG_RESEND_ERROR);
				throw ResendError("Failed to resend msg: " + e.what());
			}
		}
		else {
			// Fatal error
			recordConnectionError(MAX_RESEND_ERROR);
			throw ResendError("Can't find msg parameters in unacked msg list");
		}
	}

	return resentMsg;
}

// Is some msg a management msg?
bool Tunnel::isMgmtMsg(const HTTPMsgParms* parms) const
{
	TunnelState msgState = parms->getState();
	return (msgState == CONNECT || msgState == AUTHREQ || msgState == AUTHRESP ||
		msgState == RECONNECT);
}


// Record we need to resend a msg - Caller holds mutex
void Tunnel::needToResendMsg(const SeqNo& seqNo)
{
	// Check we haven't already recorded this msg
	std::vector<SeqNo>::iterator pos;
	pos = std::find(m_resendQueue.begin(), m_resendQueue.end(), seqNo);
	if(pos == m_resendQueue.end()) {
		// Record msg in resend queue
		pos = std::lower_bound(m_resendQueue.begin(), m_resendQueue.end(), seqNo);
		m_resendQueue.insert(pos, seqNo);
	}
}


/*********************/
/* Private Interface */
/*********************/


// Increment next send packet sequence numbers - Caller holds mutex
void Tunnel::incSendSeqNo(void)
{
	++m_nextSendSeqNo;
}

// Increase next received packet sequence numbers - Caller holds mutex
void Tunnel::incRecvSeqNo(void)
{
	// Always increment
	++m_nextRecvSeqNo;

	#ifdef DEBUG_TUN
	if(m_tunnelDataQueue.size() >= SeqNo::getMaxSeqNo() - 5) {
		std::stringstream wrn;
		wrn << "Tunnel data queue size approaching seqno limit: ";
		wrn << m_tunnelDataQueue.size() << "," << SeqNo::getMaxSeqNo();
		DPRINTW(wrn.str());
	}
	#endif

	// Have we already received a msg with this sequence no.?
	for(unsigned int i = 0; i < m_tunnelDataQueue.size(); ++i) {
		if(m_tunnelDataQueue[i]->getSeqNo() != m_nextRecvSeqNo) {
			// Still waiting for msg with this sequence no.
			break;
		}

		// Already received this msg, so increment again
		++m_nextRecvSeqNo;
	}

	#ifdef DEBUG_TUN
	std::stringstream dbg;
	dbg << "Next recv/ack seq no = " << m_nextRecvSeqNo;
	DPRINTV(dbg.str());
	#endif
}


// Create payload for tunneled msg to send to the destination - Caller holds mutex
void Tunnel::makeTunnelPayload(RawBuff& data, bool& split, unsigned int max)
	throw(MemoryError)
{
	split = false;

	// Iterate over queued raw packets
	std::vector<RawBuff*>::iterator pit;
	for(pit = m_rawDataQueue.begin(); pit != m_rawDataQueue.end(); ++pit) {
		// Get next queued raw data packet
		RawBuff* queuedData = *pit;

		// Check type of queued packets
		if(m_tid.getProtocol() == Enums::UDP) {
			// Maintain UDP datagram boundaries
			data.append(*queuedData, max);
			if(data.getLength() > 0 && data.getLength() < queuedData->getLength())
				split = true;
			break;
		}
		else {
			// Pack TCP data upto the maximum
			unsigned int appendBytes = max - data.getLength();
			data.append(*queuedData, appendBytes);
			if(data.getLength() == max) {
				// Reached maximum size
				break;
			}
		}
	}
}

// Clear up raw packet queue after sending tunneled msg - Caller holds mutex
void Tunnel::sentTunnelPayload(HTTPMsgParms* parms)
{
	// Get size of sent payload
	unsigned int payloadLen = parms->getPayload()->getPayloadSize();

	// Iterate over queued raw packets
	while(payloadLen > 0 && m_rawDataQueue.size() > 0) {
		// Get next queued raw data packet
		RawBuff* queuedData = m_rawDataQueue[0];
		unsigned int queuedDataLen = queuedData->getLength();

		// Whole of this packet sent?
		if(queuedDataLen <= payloadLen) {
			// Keep track of bytes erased
			payloadLen -= queuedDataLen;
			m_rawDataQueueSize -= queuedDataLen;

			// Erase this packet from queue
			m_rawDataQueue.erase(m_rawDataQueue.begin());
			delete queuedData;
		}
		else {
			// Partial packet sent
			queuedData->erase(payloadLen);
			m_rawDataQueueSize -= payloadLen;
			payloadLen = 0;
		}
	}

	// Check if we need to unblock any connections
	checkBlockedConnections();
}


// Create raw data for sending to the destination - Caller hold mutex
Payload* Tunnel::makeRawPayload(void) throw(MemoryError)
{
	Payload* raw = 0;

	// Got any payload data we can make the raw msg from?
	if(m_tunnelDataQueue.size() > 0) {
		// Are we missing any payload?
		Payload* payload = m_tunnelDataQueue[0];
		if(m_nextRecvSeqNo >= payload->getSeqNo()) {

			// Got split packets to deal with?
			if(payload->getSplit()) {
				bool buildSplit = false;
				// Do we have all payloads for split packet?
				std::vector<Payload*>::iterator pit;
				for(pit = m_tunnelDataQueue.begin(); pit != m_tunnelDataQueue.end(); ++pit) {
					payload = *pit;
					if(m_nextRecvSeqNo >= payload->getSeqNo()) {
						// Member of split set
						payload->setSent(true);

						// Final packet from split set?
						if(!payload->getSplit()) {
							// Got all splits, can build payload
							buildSplit = true;
							break;
						}
					}
					else {
						// Missing packet from split payload
						break;
					}
				}

				// Can we go ahead and recombine split packets?
				if(buildSplit) {
					RawBuff recombined;
					for(pit = m_tunnelDataQueue.begin(); pit != m_tunnelDataQueue.end(); ++pit) {
						payload = *pit;
						if(payload->getSent()) {
							recombined.append(payload->getPayload());
						}
						else {
							break;
						}
					}

					raw = new Payload(0, recombined, false);
				}
			}

			else {
				// Single non-split payload, so just copy it
				raw = new Payload(*payload);
				payload->setSent(true);
			}
		}
	}

	return raw;
}

// Successfully sent raw data to the destination - Caller hold mutex
void Tunnel::sentRawPayload(bool sentOk)
{
	// Free all sent payload data
	while(m_tunnelDataQueue.size() > 0) {
		Payload* payload = m_tunnelDataQueue[0];
		if(payload->getSent()) {
			m_tunnelDataQueue.erase(m_tunnelDataQueue.begin());
			delete payload;
		}
		else {
			// No more sent payload packets
			break;
		}
	}
}


// Process an acknowledgement - Mutex held by caller
void Tunnel::processAck(std::string& ack)
{
	#ifdef DEBUG_TUN
	time_t oldAckTimeout = m_ackTimeout;
	#endif

	// Convert to numeric sequence number
	SeqNo seqNo(strtoul(ack.c_str(), 0, 10));

	// This acknowledges everything less than that sequence number
	std::vector<SentMsgParms*>::iterator pit;
	while((pit = std::find_if(m_unackedMsgQueue.begin(), m_unackedMsgQueue.end(),
	CmpSentParmsLessSeqNo(seqNo))) != m_unackedMsgQueue.end()) {
		// Acknowledges this msg
		SentMsgParms* sentMsgParms = *pit;
		SeqNo acked = sentMsgParms->m_parms->getPayload()->getSeqNo();

		// Store round trip time for this ack
		time_t rtt = time(0) - sentMsgParms->m_sentTime;
		if(m_ackRTTs.size() >= MAX_UNACKED_MSG  + 1)
			m_ackRTTs.erase(m_ackRTTs.begin());
		m_ackRTTs.push_back(rtt);

		// Calculate average round trip time for acks
		time_t totRTT = 0;
		for(unsigned int i = 0; i < m_ackRTTs.size(); ++i)
			totRTT += m_ackRTTs[i];
		m_avgAckRTT = totRTT / m_ackRTTs.size();

		// Update ack timeout
		m_ackTimeout = m_avgAckRTT * 3;
		if(m_ackTimeout < m_minAckTimeout)
			m_ackTimeout = m_minAckTimeout;
		else if(m_ackTimeout > m_maxAckTimeout)
			m_ackTimeout = m_maxAckTimeout;

		// Free up sent parms
		delete sentMsgParms;
		m_unackedMsgQueue.erase(pit);

		// Is this a late arrival for a msg we've flagged as needing a resend?
		std::vector<SeqNo>::iterator rit;
		rit = std::find(m_resendQueue.begin(), m_resendQueue.end(), acked);
		if(rit != m_resendQueue.end()) {
			// No longer need to resend
			m_resendQueue.erase(rit);
		}
	}

	// Does it look like the sender is requesting we resend an unacked msg?
	checkForResendRequest(seqNo);

	#ifdef DEBUG_TUN
	if(m_ackTimeout != oldAckTimeout) {
		std::stringstream nfo;
		nfo << "Ack timeout for tunnel " << m_tid << " now " << m_ackTimeout;
		DPRINTV(nfo.str());
	}
	#endif

	// Scale connection poll periods in line with ack timeout values
	scalePollTimeout();
}


// Check some new payload isn't from a duplicate message - Caller holds Mutex
bool Tunnel::duplicateTunnelMsg(SeqNo& seqNo)
{
	// Assume it is not a duplicate until proved otherwise
	bool res = false;

	// Is this payload out of sequence?
	if(seqNo != m_nextRecvSeqNo) {
		// If it's earlier then it must be a duplicate
		if(seqNo < m_nextRecvSeqNo) {
			res = true;
		}
		else {
			// Look for a duplicate in the queue of received msgs
			std::vector<Payload*>::iterator pit;
			pit = std::find_if(m_tunnelDataQueue.begin(), m_tunnelDataQueue.end(),
				CmpPayloadSeqNo(seqNo));
			if(pit != m_tunnelDataQueue.end()) {
				// Found a duplicate
				res = true;
			}
		}

		#ifdef DEBUG_TUN
		std::stringstream dbg;
		dbg << "Got out of sequence msg " << seqNo << ", expected next " << m_nextRecvSeqNo;
		if(res) dbg << " (Duplicate)";
		DPRINTV(dbg.str());
		#endif
	}
	else {
		// Increase next expected received sequence no.
		incRecvSeqNo();
	}

	return res;
}


// Check & get a suitable connection for sending a tunnel msg out of -
// Caller holds Mutex
Connection* Tunnel::getTunnelConnection(Connection* conn)
{
	Connection* res = conn;

	// Already supplied a candidate connection to use?
	if(res) {
		// Check supplied connection is actually one of ours going the right way
		if(m_tunnelDirection == LEFT) {
			// Check connection
			if(!leftConnection(res) || !conn->connected()) {
				#ifdef DEBUG_TUN
				std::stringstream dbg;
				dbg << "Tunnel::getTunnelConnection() - Overriding connection " << conn->getID();
				if(!leftConnection(res))
					dbg << " not-left";
				if(!conn->connected())
					dbg << " not-connected";
				DPRINTI(dbg.str());
				#endif

				res = 0;
			}
		}
		else {
			// Check connection
			if(!rightConnection(res) || !conn->connected()) {
				#ifdef DEBUG_TUN
				std::stringstream dbg;
				dbg << "Tunnel::getTunnelConnection() - Overriding connection " << conn->getID();
				if(!rightConnection(res))
					dbg << " not-right";
				if(!conn->connected())
					dbg << " not-connected";
				DPRINTI(dbg.str());
				#endif

				res = 0;
			}
		}
	}

	// Do we need to look a connection up?
	if(res == 0) {
		// Get a recent connection in this tunnel going the right way
		std::vector<Tunnel::Selector*> filt;
		if(m_tunnelDirection == LEFT)
			filt.push_back(new SelLeftConn(this));
		else
			filt.push_back(new SelRightConn(this));
		filt.push_back(new SelConnectedConn());
		filt.push_back(new SelRecentConn());
		res = getConnFromSel(filt);

		#ifdef DEBUG_TUN
		if(res) {
			std::stringstream nfo;
			nfo << "Selected connection id " << res->getID();
			DPRINTV(nfo.str());
		}
		#endif
	}

	return res;
}


// Advance our internal state machine - Caller holds mutex
void Tunnel::advanceState(const HTTPMsgParms* parms)
{
	// Save current state
	TunnelState oldState = m_tunnelState;

	// Once we are closing or closed, don't change out of those states
	if(m_tunnelState != CLOSED && m_tunnelState != CLOSING) {
		switch(parms->getState()) {
			case(CONNECT): {
				// Now we need to authenticate
				m_tunnelState = AUTHREQ;
				break;
			}
			case(AUTHREQ): {
				// Send authentiation response
				m_tunnelState = AUTHRESP;
				break;
			}
			case(AUTHRESP): {
				// Now we are connected
				m_tunnelState = CONNECTED;
				break;
			}
			case(RECONNECT): {
				// Reconnected
				m_tunnelState = CONNECTED;
				break;
			}
			case(CLOSING): {
				// Closing
				m_tunnelState = CLOSING;
				break;
			}
			case(CLOSED): {
				// Closed
				m_tunnelState = CLOSED;
				break;
			}
			case(CONNECTED): {
				// Stay connected
				m_tunnelState = CONNECTED;
				break;
			}
			case(UNKNOWNSTATE): {
				// Assume connected
				m_tunnelState = CONNECTED;
				g_pLogger->error("Tunnel state is UNKOWN");
				break;
			}
		}

		// Have we changed state?
		if(oldState != m_tunnelState) {
			// Let controlling tunneler know
			m_tunneler->scheduleEvent(new TunnelStateEvent(m_tunneler,
				m_tid, m_tunnelState));

			// If we've become CONNECTED, log details
			if(m_tunnelState == CONNECTED) {
				Authenticator* authenticator = getAuthenticator();
				m_logKey = authenticator->logTunnel(this);
			}
		}
	}

	#ifdef DEBUG_TUN
	if(oldState != m_tunnelState) {
		std::stringstream nfo;
		nfo << "Tunnel moved into state " << stateAsString(m_tunnelState);
		DPRINTI(nfo.str());
	}
	#endif
}


// Get tunnel state as string
std::string Tunnel::stateAsString(TunnelState state) const
{
	const char* stateStr[] = { "CONNECT", "AUTHREQ", "AUTHRESP", "RECONNECT",
		"CLOSING", "CLOSED", "CONNECTED", "UNKNOWNSTATE" };
	return stateStr[state];
}


// Is a connection an old removed one? - Caller holds mutex
bool Tunnel::oldConnection(Connection* conn) const
{
	std::vector<unsigned int>::const_iterator ocit;
	ocit = std::find(m_oldConns.begin(), m_oldConns.end(), conn->getID());
	return (ocit != m_oldConns.end());
}


// Send a tunnel msg with the given paramaters - Caller holds mutex
void Tunnel::sendTunnelMsgWithParms(Connection* conn, HTTPMsgParms* parms)
	throw(TunnelError)
{
	bool sentOk = false;

	try {
		#ifdef DEBUG_TUN
		std::stringstream dbg;
		dbg << "Sending HTTP msg seq no: " << parms->getPayload()->getSeqNo();
		dbg << " state: " << stateAsString(parms->getState());
		if(parms->getAck().size() > 0)
			dbg << " ack: " << parms->getAck();
		dbg << " on connection " << conn->getID();
		DPRINTV(dbg.str());
		#endif

		// Write data to connection
		conn->write(parms);
		sentOk = true;

		// Save msg parameters incase we need to resend later
		sentNewTunnelMsg(parms);
	}
	catch(CommsError& e) {
		g_pLogger->error("CommsError sending tunnel msg: " + e.what());
		if(!conn->closing()) {
			// Schedule error reporting event for connection
			m_tunneler->scheduleEvent(new ConnectionErrorEvent(conn->getID(),
				GENERAL_COMMS_ERROR, m_tunneler));
		}
	}
	catch(TunnelError& e) {
		g_pLogger->error("General failure sending tunnel msg: " + e.what());
	}

	// Error sending?
	if(!sentOk && parms) {
		// Free parameters from failed message
		delete parms;

		// Let caller know
		throw TunnelError("Failed to send tunnel msg");
	}
}


/**************/
/* Predicates */
/**************/


// CmpSentParmsSeqNo
bool Tunnel::CmpSentParmsSeqNo::operator()(const SentMsgParms* p)
{
	return (p->m_parms->getPayload()->getSeqNo() == m_seqNo);
}

// CmpSentParmsLessSeqNo
bool Tunnel::CmpSentParmsLessSeqNo::operator()(const SentMsgParms* p)
{
	return (p->m_parms->getPayload()->getSeqNo() < m_seqNo);
}


/**********************/
/* SentMsgParms Class */
/**********************/


// Construction/destruction
Tunnel::SentMsgParms::SentMsgParms(HTTPMsgParms* parms)
: m_parms(parms), m_sentTime(time(0)), m_ackCount(0), m_resentCount(0)
{
}
Tunnel::SentMsgParms::~SentMsgParms()
{
	// Free up HTTP message parameters
	delete m_parms;
}


}	// namespace

