/* (C) 2005 Richard Mills / The Linux Academy Ltd.
 *
 * Somewhere to store all information about a unique Tunnel.
 *
 * $Id: Tunnel.hh,v 1.69 2007-08-04 11:21:43 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#ifndef _TUNNEL_HH_
#define _TUNNEL_HH_

#include <sys/time.h>
#include <vector>
#include <functional>
#include <sstream>
#include <time.h>
#include "Mutex.hh"
#include "Payload.hh"
#include "Exceptions.hh"
#include "TunnelID.hh"
#include "Time.hh"


namespace HTTPTunnel {

class TunnelConfig;
class Tunneler;
class RawBuff;
class Connection;
class HTTPMsg;
class HTTPMsgParms;
class ProxySettings;
class ServerAuthSettings;
class Authenticator;


class Tunnel {
public:

	/*********/
	/* Types */
	/*********/

	// Left and right
	enum Direction { LEFT, RIGHT, UNKNOWNDIR };

	// Tunnel states
	enum TunnelState { CONNECT, AUTHREQ, AUTHRESP, RECONNECT, CLOSING,
		CLOSED, CONNECTED, UNKNOWNSTATE };

	// Enumerate all communication error type
	enum TunnelErrorType {
		GENERAL_COMMS_ERROR, PROXY_AUTH_ERROR, SERVER_AUTH_ERROR, BAD_RESPONSE_ERROR,
		MSG_RESEND_ERROR, MAX_RESEND_ERROR, CONNECTION_TIMEOUT_ERROR, CONNECTION_CLOSED_ERROR,
		UNVIABLE_ERROR, QUOTAS_EXCEEDED_ERROR, FATAL_QUOTAS_ERROR
	};


	/*************/
	/* Selectors */
	/*************/

	// Base class
	struct Selector {
		Selector() {}
		virtual ~Selector() {}
		virtual void operator()(std::vector<Connection*>& sel) = 0;
	};

	// Select only left side connections
	struct SelLeftConn : public Selector {
		SelLeftConn(Tunnel* tun) : m_tun(tun) {}
		void operator()(std::vector<Connection*>& sel);
		Tunnel* m_tun;
	};

	// Select only right side connections
	struct SelRightConn : public Selector {
		SelRightConn(Tunnel* tun) : m_tun(tun) {}
		void operator()(std::vector<Connection*>& sel);
		Tunnel* m_tun;
	};

	// Sort according to most recently first connected connection
	struct SelRecentConn : public Selector {
		void operator()(std::vector<Connection*>& sel);
	};

	// Exclude the specified connection
	struct SelNotConn : public Selector {
		SelNotConn(Connection* conn) : m_conn(conn) {}
		void operator()(std::vector<Connection*>& sel);
		Connection* m_conn;
	};

	// Select only connected connections
	struct SelConnectedConn : public Selector {
		void operator()(std::vector<Connection*>& sel);
	};


	/**************/
	/* Predicates */
	/**************/

	// Determine whether connection is from left side
	struct PredIsLeft : public std::unary_function<Connection*, bool> {
		PredIsLeft(Tunnel* tun) : m_tun(tun) {}
		bool operator()(const Connection* const& c) {
			return (m_tun->leftConnection(c));
		}
		Tunnel* m_tun;
	};

	// Determine whether connection is from right side
	struct PredIsRight : public std::unary_function<Connection*, bool> {
		PredIsRight(Tunnel* tun) : m_tun(tun) {}
		bool operator()(const Connection* const& c) {
			return (m_tun->rightConnection(c));
		}
		Tunnel* m_tun;
	};


	/***********/
	/* Methods */
	/***********/

	// Construction/destruction
	Tunnel(TunnelConfig* tcfg, Tunneler* tunneler, Direction tunDir, TunnelState state);
	virtual ~Tunnel();


	/*************/
	/* Accessors */
	/*************/

	// Get our config
	const TunnelConfig* getConfig(void) const;
	ProxySettings* getProxySettings(void);
	ServerAuthSettings* getServerAuthSettings(void);

	// Get our tunnel ID as a string
	std::string getTidStr(void) const;

	// Get our owning tunneler
	Tunneler* getTunneler(void);

	// Does the specified connection belong to this Tunnel?
	bool ownsConnection(const Connection* conn);

	// Is the specified connection from the left or right side?
	bool leftConnection(const Connection* conn);
	bool rightConnection(const Connection* conn);

	// Get the number of left/right side connections for this tunnel
	unsigned int noLeftConnections(void);
	unsigned int noRightConnections(void);

	// Get connections matching the provided criteria
	Connection* getConnFromSel(std::vector<Selector*>& crit);
	void getConnsFromSel(std::vector<Selector*>& crit, std::vector<Connection*>& res);

	// Get current poll timeout
	unsigned int getPollTimeout(void) const;

	// Get name of owning user
	std::string getUserName(void) const;

	// Get source/destination addresses of connections for this tunnel
	virtual std::string getSrcConnectionAddresses(void);
	virtual std::string getDstConnectionAddresses(void);

	// Get logging key
	unsigned int getLogKey(void) const;

	// Get our current state
	TunnelState getState(void) const;

	// Have we not received any data for a suspiciously long time?
	bool rawRecvTimeout(void);
	bool tunRecvTimeout(void);


	/************/
	/* Mutators */
	/************/

	// Add new connections to this tunnel
	void addLeftConnection(Connection* conn);
	void addRightConnection(Connection* conn);

	// Remove closing connection from this tunnel
	void removeConnection(Connection* conn);

	// A connection that carries us has connected
	void connectionConnected(Connection* conn);

	// Decrease leaky bucket error count
	void decreaseErrorCount(void);

	// Record connection error
	void recordConnectionError(TunnelErrorType errType);

	// State change notifications
	void closing(void);
	void reconnecting(void);


	/*************/
	/* Utilities */
	/*************/

	// Process some newly extracted tunnel/raw data from one of our connections
	virtual void processTunnelMsg(const HTTPMsg* msg);
	void processRawData(const RawBuff& raw);

	// Determine whether we have experienced too many errors to continue
	bool errorLevelExceeded(Direction dir);
	bool errorLevelExceeded(void);

	// Send messages in the appropriate direction
	void sendTunnelMsg(bool mgmt, bool retry, Connection* conn);
	void respondToTunnelMsg(bool mgmt, Connection* conn, const HTTPMsgParms* recvd);
	void pollTunnel(void);
	void sendRawMsgs(void);

	// Check ack timeout on all unacknowledged messages
	bool checkAckTimeout(void);

	// Are we in a state where we can be shutdown?
	bool canBeShutdown(void);


protected:

	/*********/
	/* Types */
	/*********/

	// HTTPMsgParms for a sent message, plus other details
	struct SentMsgParms {
		// Construction/destruction
		SentMsgParms(HTTPMsgParms* parms);
		~SentMsgParms();

		// Data
		HTTPMsgParms*			m_parms;
		time_t					m_sentTime;
		unsigned int			m_ackCount;
		unsigned int			m_resentCount;
	};


	/*************/
	/* Constants */
	/*************/

	// Maximum payload length for created HTTP messages
	static const unsigned int		MAX_NORMAL_HTTP_PAYLOAD		= 8000;
	static const unsigned int		MAX_NOPROXY_HTTP_PAYLOAD	= 20000;
	static const unsigned int		MAX_URL_HTTP_PAYLOAD			= 1200;

	// Time between poll timeout events (milli-seconds)
	static const unsigned int		MIN_POLL_TIMEOUT				= 250;
	static const unsigned int		MAX_POLL_TIMEOUT				= 1000;

	// Maximum no. of outstanding unacked msgs
	static const unsigned int		MAX_UNACKED_MSG				= 5;

	// Maximum no. of times we will try resending a msg
	static const unsigned int		MAX_MSG_RESEND					= 3;

	// Raw data buffer fill threshhold
	static const unsigned int		MAX_RAW_DATA_FILL				= 1024 * 1024;
	static const unsigned int		RAW_DATA_UNBLOCK_LEVEL		= 90;


	/**************/
	/* Predicates */
	/**************/

	// Find msg parameter by payload sequence number
	struct CmpSentParmsSeqNo {
		CmpSentParmsSeqNo(const SeqNo& seqNo) : m_seqNo(seqNo) {}
		bool operator()(const SentMsgParms* p);
		SeqNo m_seqNo;
	};

	// Find msg parameter less than payload sequence number
	struct CmpSentParmsLessSeqNo {
		CmpSentParmsLessSeqNo(const SeqNo& seqNo) : m_seqNo(seqNo) {}
		bool operator()(const SentMsgParms* p);
		SeqNo m_seqNo;
	};


	/***********/
	/* Methods */
	/***********/

	// Create a new payload msg for sending down the tunnel
	virtual HTTPMsgParms* makeTunnelMsg(const HTTPMsgParms* recvd,
		unsigned int maxPayload, bool retry) throw(MemoryError);

	// Record that we've sent a new tunnel msg
	virtual void sentNewTunnelMsg(HTTPMsgParms* msgParms) = 0;
	void sentNewTunnelMsg(HTTPMsgParms* msgParms, bool remember);

	// Scale poll timeout
	virtual void scalePollTimeout(void);

	// Get authenticator
	virtual Authenticator* getAuthenticator(void);

	// Have we got a stable HTTP connection to other end?
	virtual bool httpStable(void) const;

	// Can we send a new tunnel msg?
	virtual bool clearToSendNewMsg(Connection* conn) = 0;

	// Determine whether an ack implies we should resend a msg
	virtual void checkForResendRequest(const SeqNo& ackNo) = 0;

	// Check whether any connections need to be blocked/unblocked
	virtual void checkBlockedConnections(void) = 0;

	// Handle resend of old msgs
	void checkUnackedMgmtMsgs(void);
	bool resendUnackedMsgs(Connection* conn) throw(ResendError);
	bool isMgmtMsg(const HTTPMsgParms* parms) const;

	// Record we need to resend a msg - Caller holds mutex
	void needToResendMsg(const SeqNo& seqNo);


	/********/
	/* Data */
	/********/

	// A Mutex
	Threads::Mutex							m_mtx;

	// Lists of connections that comprise this Tunnel
	std::vector<Connection*>			m_leftConn;
	std::vector<Connection*>			m_rightConn;

	// Configuration for this Tunnel
	TunnelConfig*							m_tunCfg;
	TunnelID									m_tid;

	// Min/max and current ack timeout values
	time_t									m_minAckTimeout;
	time_t									m_maxAckTimeout;
	time_t									m_ackTimeout;

	// User of this tunnel
	std::string								m_user;

	// Time between poll timeout events
	unsigned int							m_pollTimeout;

	// Current tunnel state
	TunnelState								m_tunnelState;

	// Queue of parameters for sent but unacknowledged messages
	std::vector<SentMsgParms*>			m_unackedMsgQueue;

	// Time we last sent ANY msg down the tunnel
	Time										m_lastTunnelAnySendTime;

	// Queues of received tunnel/raw data
	std::vector<Payload*>				m_tunnelDataQueue;
	std::vector<RawBuff*>				m_rawDataQueue;
	unsigned int							m_rawDataQueueSize;


private:

	/*************/
	/* Constants */
	/*************/

	// Amount to decrease leaky bucket error count by each period
	static const unsigned int		ERROR_COUNT_DECREASE			= 1;

	// Level for leaky bucket error count which we consider fatal
	static const unsigned int		FATAL_ERROR_COUNT				= 20;

	// Ack timeout in seconds
	static const time_t				MIN_ACK_TIMEOUT				= 5;
	static const time_t				MAX_ACK_TIMEOUT				= 30;

	// Seconds tunnel can stay completely silent before we call it an error
	static const time_t				MAX_SILENCE_TIME				= 60;

	// Seconds we go with no data received before considering tunnel silent
	static const time_t				MAX_NO_RECV_TIME				= 60 * 2;

	// Time after sending last tunnel msg we act on unviable errors (seconds)
	static const unsigned int		UNVIABLE_ERROR_TIME			= 5;


	/**************/
	/* Predicates */
	/**************/

	// Order message payloads by sequence number
	struct CmpPayloadsSeqNo {
		bool operator()(const Payload* p1, const Payload* p2) {
			return (*p1 < *p2);
		}
	};

	// Find message payload by sequence number
	struct CmpPayloadSeqNo {
		CmpPayloadSeqNo(const SeqNo& seqNo) : m_seqNo(seqNo) {}
		bool operator()(const Payload* p) {
			return (p->getSeqNo() == m_seqNo);
		}
		SeqNo m_seqNo;
	};


	/***********/
	/* Methods */
	/***********/

	// Increment next sent/received packet sequence numbers
	void incSendSeqNo(void);
	void incRecvSeqNo(void);

	// HTTP payload handling
	void makeTunnelPayload(RawBuff& data, bool& split, unsigned int max)
		throw(MemoryError);
	void sentTunnelPayload(HTTPMsgParms* parms);

	// Handle raw data for sending to the destination
	Payload* makeRawPayload(void) throw(MemoryError);
	void sentRawPayload(bool sentOk);

	// Handle some received acknowledgements/non-acknowledgements
	void processAck(std::string& ack);

	// Check some new payload isn't from a duplicate message
	bool duplicateTunnelMsg(SeqNo& seqNo);

	// Check & get a suitable connection for sending a tunnel msg out of
	Connection* getTunnelConnection(Connection* conn);

	// Advance our internal state machine
	void advanceState(const HTTPMsgParms* parms);

	// Get tunnel state as string
	std::string stateAsString(TunnelState state) const;

	// Is a connection an old removed one?
	bool oldConnection(Connection* conn) const;

	// Send a tunnel msg with the given paramaters
	void sendTunnelMsgWithParms(Connection* conn, HTTPMsgParms* parms)
		throw(TunnelError);


	/********/
	/* Data */
	/********/

	// Remember IDs of connections we've explicitly dropped
	std::vector<unsigned int>			m_oldConns;

	// Owning tunneler
	Tunneler*								m_tunneler;

	// Sequence numbers for msgs sent/received
	SeqNo										m_nextSendSeqNo;
	SeqNo										m_nextRecvSeqNo;

	// Queue of sequence nos. for messages we must resend
	std::vector<SeqNo>					m_resendQueue;

	// Leaky bucket error counter
	unsigned int							m_errorCount;

	// Flag to indicate if the last dead connections ever connected
	bool										m_leftEverConnected;
	bool										m_rightEverConnected;

	// In which direction does the tunnel lie?
	Direction								m_tunnelDirection;

	// Flag to indicate whether we have scheduled a shutdown
	bool										m_shutdownScheduled;

	// Round trip ack times
	std::vector<time_t>					m_ackRTTs;
	time_t									m_avgAckRTT;

	// Things to determine whether we can be shutdown
	time_t									m_lastTunnelSendTime;
	bool										m_sentCloseMsg;

	// Time we last sent a standard CONNECTED msg down the tunnel
	Time										m_lastTunnelConnectedSendTime;

	// Key to use for logging
	unsigned int							m_logKey;

	// Time we last received data
	Time										m_lastRawRecvTime;
	Time										m_lastTunRecvTime;

};


}	// namespace

#endif

