/* (C) 2005 Richard Mills / The Linux Academy Ltd.
 *
 * Uniquely identify an individual tunneled connection.
 *
 * $Id: TunnelID.cc,v 1.11 2007-04-29 11:56:38 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <string>
#include <sstream>
#include <iostream>
#include <vector>
#include "TunnelID.hh"
#include "Exceptions.hh"
#include "Utils.hh"

namespace HTTPTunnel {


// Construction/destruction
TunnelID::TunnelID()
: m_id(0), m_svrIP(), m_svrPort(0), m_cliName(), m_cliProcID(0),
  m_protocol(Enums::TCP)
{
}

TunnelID::TunnelID(unsigned int id, std::string serverip, unsigned int serverport,
	std::string clientname, unsigned int proc, Enums::Protocol prot)
: m_id(id), m_svrIP(serverip), m_svrPort(serverport), m_cliName(clientname),
  m_cliProcID(proc), m_protocol(prot)
{
}

TunnelID::TunnelID(std::string id) throw(InvalidTunnelID)
: m_id(0), m_svrIP(), m_svrPort(0), m_cliName(), m_cliProcID(0),
  m_protocol(Enums::TCP)
{
	// Split composite ID
	std::vector<std::string> parts;
	Utils::SplitString(id, ":", parts);

	// Got everything we need?
	if(parts.size() < 5) {
		std::stringstream err;
		err << "Invalid tunnel ID string: " << id;
		throw InvalidTunnelID(err.str());
	}

	// Construct individual ID values
	m_id = strtoul(parts[0].c_str(), 0, 10);
	m_svrIP = parts[1];
	m_svrPort = atoi(parts[2].c_str());
	m_cliName = parts[3];
	m_cliProcID = atoi(parts[4].c_str());
	if(parts.size() == 6)
		m_protocol = (Enums::Protocol)(atoi(parts[5].c_str()));
}

TunnelID::TunnelID(const TunnelID& that)
: m_id(that.m_id), m_svrIP(that.m_svrIP), m_svrPort(that.m_svrPort),
  m_cliName(that.m_cliName), m_cliProcID(that.m_cliProcID),
  m_protocol(that.m_protocol)
{
}

TunnelID::~TunnelID()
{
}

// Assignment
TunnelID& TunnelID::operator=(const TunnelID& that)
{
	// Self assignment?
	if(this != &that) {
		m_id = that.m_id;
		m_svrIP = that.m_svrIP;
		m_svrPort = that.m_svrPort;
		m_cliName = that.m_cliName;
		m_cliProcID = that.m_cliProcID;
		m_protocol = that.m_protocol;
	}

	return *this;
}


// Operators
bool TunnelID::operator==(const TunnelID& that) const
{
	// ID, IP addresses, port, proc ID and protocol must match
	bool res = (that.m_id == m_id);
	res &= (that.m_svrIP == m_svrIP);
	res &= (that.m_svrPort == m_svrPort);
	res &= (that.m_cliName == m_cliName);
	res &= (that.m_cliProcID == m_cliProcID);
	res &= (that.m_protocol == m_protocol);

	return res;
}

bool TunnelID::operator<(const TunnelID& that) const
{
	// Just create a string of all members and compare that
	return (toString() < that.toString());
}


// Accessors
unsigned int TunnelID::getID(void) const
{
	return m_id;
}

std::string TunnelID::getServerIP(void) const
{
	return m_svrIP;
}

unsigned int TunnelID::getServerPort(void) const
{
	return m_svrPort;
}

std::string TunnelID::getClientName(void) const
{
	return m_cliName;
}

unsigned int TunnelID::getProcID(void) const
{
	return m_cliProcID;
}

Enums::Protocol TunnelID::getProtocol(void) const
{
	return m_protocol;
}


// Create a representation of ourselves as a string
std::string TunnelID::toString(void) const
{
	std::stringstream val;
	val << m_id << ":" << m_svrIP << ":" << m_svrPort << ":" << m_cliName;
	val << ":" << m_cliProcID << ":" << (unsigned int)(m_protocol);
	return val.str();
}


/******************/
/* Helper Methods */
/******************/


// Output to a stream
std::ostream& operator<<(std::ostream& str, const TunnelID& tid)
{
	str << tid.toString();
	return str;
}


}	// namespace

