/* (C) 2005 Richard Mills / The Linux Academy Ltd.
 *
 * Base class for client and server tunnel controllers.
 *
 * $Id: Tunneler.cc,v 1.73 2007-07-07 18:05:11 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <vector>
#include <iostream>
#include <sstream>
#include <algorithm>
#include "Tunneler.hh"
#include "Debug.hh"
#include "TunnelID.hh"
#include "Tunnel.hh"
#include "MutexVar.hh"
#include "Exceptions.hh"
#include "Logger.hh"
#include "Connection.hh"
#include "Payload.hh"
#include "RawBuff.hh"
#include "HTTPMsg.hh"
#include "ConnectResultEvent.hh"
#include "ConnectionErrorEvent.hh"
#include "LeakTunnelErrorEvent.hh"
#include "ShutdownTunnelEvent.hh"
#include "PollTimeoutEvent.hh"
#include "CheckAckEvent.hh"
#include "TunnelStateEvent.hh"
#include "TunnelUnviableEvent.hh"
#include "ConnectionManagementEvent.hh"
#include "TunnelManagementEvent.hh"
#include "AuditEvent.hh"
#include "UDPSilentEvent.hh"
#include "SendMsgEvent.hh"
#include "ConnectionManager.hh"
#include "Server.hh"
#include "ClientSideConfig.hh"
#include "ServerSideConfig.hh"


// Global references
extern HTTPTunnel::Logger* g_pLogger;
extern HTTPTunnel::GlobalConfig* g_pConfig;


namespace HTTPTunnel {


// Initialise static data
unsigned int Tunneler::c_tunCount = 0;
Threads::Mutex Tunneler::c_tunCountMtx;


// Construction/destruction
Tunneler::Tunneler(Server* server)
: EventReceiver(), m_tunMtx(), m_server(server), m_tunnels(), m_tunnelsMtx(),
  m_closingTunnels(), m_closingTunnelsMtx(), m_connections(), m_connectionsMtx(),
  m_closingConnections(), m_closingConnectionsMtx(), m_connectionMgr(0)
{
	DPRINTV("Tunneler::Tunneler()");

	// Create connection manager
	m_connectionMgr = new ConnectionManager();

	// Schedule initial events
	scheduleEvent(new CheckAckEvent(ACK_CHECK_PERIOD, this));
	scheduleEvent(new ConnectionManagementEvent(CONN_MANAGE_PERIOD, this));
	scheduleEvent(new TunnelManagementEvent(TUN_MANAGE_PERIOD, this));
	scheduleEvent(new LeakTunnelErrorEvent(LEAK_ERROR_PERIOD, this));

	#ifdef DEBUG_TUN
	scheduleEvent(new AuditEvent(AUDIT_PERIOD, this));
	#endif
}

Tunneler::~Tunneler()
{
	DPRINTV("Tunneler::~Tunneler()");

	// Clear up list of Tunnels
	Threads::MutexVar tmv(m_tunnelsMtx);
	std::vector<Tunnel*>::iterator tit;
	for(tit = m_tunnels.begin(); tit != m_tunnels.end(); ++tit) {
		delete *tit;
		*tit = 0;
	}
	m_tunnels.clear();
	tmv.unlock();

	// Destroy connection manager
	Threads::MutexVar cmv(m_connectionsMtx);
	Threads::MutexVar ccmv(m_closingConnectionsMtx);
	delete m_connectionMgr;
	m_connectionMgr = 0;

	// Clear up list of connections
	std::vector<Connection*>::iterator cit;
	for(cit = m_connections.begin(); cit != m_connections.end(); ++cit) {
		delete *cit;
		*cit = 0;
	}
	m_connections.clear();

	// Clear up list of closing connections
	std::vector<Connection*>::iterator ccit;
	for(ccit = m_closingConnections.begin(); ccit != m_closingConnections.end(); ++ccit) {
		delete *ccit;
		*ccit = 0;
	}
	m_closingConnections.clear();
}


/*************/
/* Accessors */
/*************/


// Get owning server
const Server* Tunneler::getServer(void) const
{
	return m_server;
}

Server* Tunneler::getServerMutable(void)
{
	return m_server;
}


// Get owning server configuration
const ClientSideConfig* Tunneler::getClientSideConfig(void) const
	throw(ConfigError)
{
	// Get basic config and cast it
	const ClientSideConfig* cfg = dynamic_cast<const ClientSideConfig*>
		(m_server->getConfig());

	// Got it ok?
	if(cfg == 0)
		throw ConfigError("Unable to retrieve client side config");
	return cfg;
}

ClientSideConfig* Tunneler::getClientSideConfigMutable(void)
	throw(ConfigError)
{
	// Get basic config and cast it
	ClientSideConfig* cfg = dynamic_cast<ClientSideConfig*>
		(m_server->getConfigMutable());

	// Got it ok?
	if(cfg == 0)
		throw ConfigError("Unable to retrieve client side config");
	return cfg;
}

const ServerSideConfig* Tunneler::getServerSideConfig(void) const
	throw(ConfigError)
{
	// Get basic config and cast it
	const ServerSideConfig* cfg = dynamic_cast<const ServerSideConfig*>
		(m_server->getConfig());

	// Got it ok?
	if(cfg == 0)
		throw ConfigError("Unable to retrieve server side config");
	return cfg;
}

ServerSideConfig* Tunneler::getServerSideConfigMutable(void)
	throw(ConfigError)
{
	// Get basic config and cast it
	ServerSideConfig* cfg = dynamic_cast<ServerSideConfig*>
		(m_server->getConfigMutable());

	// Got it ok?
	if(cfg == 0)
		throw ConfigError("Unable to retrieve server side config");
	return cfg;
}


// Get number of active managed tunnels
unsigned int Tunneler::getNoActiveTunnels(void)
{
	Threads::MutexVar tmv(m_tunnelsMtx);
	return m_tunnels.size();
}


/**********/
/* Events */
/**********/


// Notification of a new UDP 'connection'
void Tunneler::newUDPConnectionArrived(Sock* insock) throw(CommsError)
{
	DPRINTI("Tunneler::newUDPConnectionArrived()");

	// Error to receive this in a non-overriden class
	throw CommsError("UDP tunneling not supported in this tunnel type");
}


// Notification that a tunnel needs to be cleared down
void Tunneler::shutdownTunnel(const TunnelID& tid, bool otherEndInitiated)
{
	#ifdef DEBUG_TUN
	std::stringstream nfo;
	nfo << "Tunneler::shutdownTunnel() - " << tid.toString();
	nfo << " OEI = " << otherEndInitiated;
	DPRINTI(nfo.str());
	#endif

	try {
		// Lookup tunnel
		Threads::MutexVar tmv(m_tunnelsMtx);
		Tunnel* tun = getTunFromPred(PredTunID(&tid));
		if(tun) {
			if(otherEndInitiated) {
				// Other end not expecting anything else from us, so just destroy tunnel
				destroyTunnel(tun);
			}
			else {
				// Try to shutdown tunnel cleanly
				tunnelClosing(tun);
			}
		}
		else {
			std::stringstream err;
			err << "Tunneler::shutdownTunnel() - Unknown tunnel " << tid;
			g_pLogger->warn(err.str());
		}
	}
	catch(LookupManagementError& e) {
		g_pLogger->warn("Error shutting down tunnel: " + e.what());
	}
}


// Notification of a poll timeout on a tunnel
void Tunneler::pollTimeout(const TunnelID& tid)
{
	// Not interested in base class
}


// Check UDP datagram timeout
void Tunneler::checkDatagramTimeout(const TunnelID& tid)
{
	DPRINTI("Tunneler::checkDatagramTimeout() - " + tid.toString());

	try {
		// Lookup tunnel
		Threads::MutexVar tmv(m_tunnelsMtx);
		Tunnel* tun = getTunFromPred(PredTunID(&tid));
		if(tun) {
			// UDP silence threshold exceeded?
			if(tun->rawRecvTimeout() && tun->tunRecvTimeout()) {
				DPRINTV("UDP silence threshold exceeded for tunnel " + tid.toString());

				// Try to shutdown tunnel cleanly
				tunnelClosing(tun);
			}
			else {
				// Schedule next UDP silent check event
				scheduleEvent(new UDPSilentEvent(this, tid));
			}
		}
		else {
			std::stringstream err;
			err << "Tunneler::checkDatagramTimeout() - Unknown tunnel " << tid;
			g_pLogger->warn(err.str());
		}
	}
	catch(LookupManagementError& e) {
		g_pLogger->warn("Error processing UDP datagram timeout: " + e.what());
	}
}


// Notification that a connection has changed state
void Tunneler::connectionStateChange(unsigned int connID,
	Connection::ConnectionState state)
{
	#ifdef DEBUG_TUN
	std::stringstream nfo;
	nfo << "Tunneler::connectionStateChange () - connection " << connID;
	nfo << " state: " << state;
	DPRINTI(nfo.str());
	#endif

	try {
		// Lookup connection
		Threads::MutexVar cmv(m_connectionsMtx);
		Threads::MutexVar tmv(m_tunnelsMtx);
		std::vector<Connection*>::iterator cit;
		cit = getConnection(connID);
		Connection* conn = *cit;

		// Do whatever the state change requires
		if(state == Connection::CONNECTED) {
			conn->tellTunnelsAboutConnect();
		}
	}
	catch(LookupManagementError& e) {
		g_pLogger->warn("Error handling connection state change: " + e.what());
	}
}


// Notification that a tunnel has changed state
void Tunneler::tunnelStateChanged(const TunnelID& tid, Tunnel::TunnelState state)
{
	// Not interested in base class
}


// Notification of a socket connect failure
void Tunneler::connectResult(ConnectResultEvent* result)
{
	DPRINTI("Tunneler::connectResult()");

	try {
		// Lookup connection
		Threads::MutexVar cmv(m_connectionsMtx);
		Threads::MutexVar tmv(m_tunnelsMtx);
		std::vector<Connection*>::iterator cit;
		cit = getConnection(result->getConnectionID());
		Connection* conn = *cit;

		// Let connection manager clear up after connect attempt
		m_connectionMgr->delProxyConnect(conn);

		// Connect failed?
		ConnectResultEvent::ConnectResultType resType = result->getType();
		if(resType != ConnectResultEvent::SUCCESS) {
			// Record failed connect attempt
			conn->incConnectFailedCount();

			// Shall we try again?
			bool retry = true;
			if(resType == ConnectResultEvent::FIREWALLED)
				retry = false;
			else if(conn->getConnectFailedCount() >= MAX_CONNECT_ATTEMPTS)
				retry = false;
			if(retry) {
				// Schedule reconnect attempt
				m_connectionMgr->addProxyConnect(conn, RECONNECT_DELAY,
					result->getHost(), result->getPort(), result->getProtocol(),
					result->getSSL(), result->getSSLCert());
			}
			else {
				// Clear up connection
				destroyConnection(cit);
			}
		}
	}
	catch(LookupManagementError& e) {
		g_pLogger->warn("Error handling connect result: " + e.what());
	}
}


// Notification of a fatal error on a socket connection
void Tunneler::connectionError(ConnectionErrorEvent* error)
{
	// Get values for error
	unsigned int connID = error->getConnectionID();
	Tunnel::TunnelErrorType errType = error->getErrorType();

	#ifdef DEBUG_TUN
	std::stringstream nfo;
	nfo << "Tunneler::connectionError() - Connection id " << connID;
	nfo << ", type " << errType;
	DPRINTI(nfo.str());
	#endif

	try {
		// Lookup connection (also lock tunnels to prevent race with tunnel shutdown)
		Threads::MutexVar cmv(m_connectionsMtx);
		Threads::MutexVar tmv(m_tunnelsMtx);
		std::vector<Connection*>::iterator cit = getConnection(connID);

		// Record error
		(*cit)->tellTunnelsAboutError(errType);

		// Move connection to the closing list
		connectionClosing(cit);
	}
	catch(LookupManagementError& e) {
		g_pLogger->warn("Error handling connection error: " + e.what());
	}
}


// Notification that a connection needs to be cleared down
void Tunneler::shutdownConnection(unsigned int connID)
{
	#ifdef DEBUG_TUN
	std::stringstream nfo;
	nfo << "Tunneler::shutdownConnection() - Connection id " << connID;
	DPRINTI(nfo.str());
	#endif

	try {
		// Lookup connection
		Threads::MutexVar cmv(m_connectionsMtx);
		Threads::MutexVar tmv(m_tunnelsMtx);
		std::vector<Connection*>::iterator cit = getConnection(connID);

		// Move connection to the closing list
		connectionClosing(cit);
	}
	catch(LookupManagementError& e) {
		g_pLogger->warn("Error shutting down connection: " + e.what());
	}
}


// Notification we should force a send of a message down a tunnel/connection
void Tunneler::sendTunnelMsg(const TunnelID& tid, unsigned int connID, bool retry,
	bool mgmt)
{
	DPRINTV("Tunneler::sendTunnelMsg() - " + tid.toString());

	try {
		// Lookup connection
		Threads::MutexVar cmv(m_connectionsMtx);
		std::vector<Connection*>::iterator cit = getConnection(connID);
		Connection* conn = *cit;

		// Lookup tunnel
		Threads::MutexVar tmv(m_tunnelsMtx);
		Tunnel* tun = getTunFromPred(PredTunID(&tid));
		if(tun) {
			// Force send
			tun->sendTunnelMsg(mgmt, retry, conn);
		}
		else {
			std::stringstream wrn;
			wrn << "Failed to force send of tunnel msg down unknown tunnel ";
			wrn << tid;
			g_pLogger->warn(wrn.str());
		}
	}
	catch(LookupManagementError& e) {
		std::stringstream wrn;
		wrn << "Error forcing send of tunnel msg down connection " << connID;
		wrn << ": " << e.what();
		g_pLogger->warn(wrn.str());
	}
}


// Notification we should try to flush a connection
void Tunneler::flushConnection(unsigned int connID, bool retry)
{
	#ifdef DEBUG_TUN
	std::stringstream nfo;
	nfo << "Tunneler::flushConnection() - Connection id " << connID;
	DPRINTI(nfo.str());
	#endif

	try {
		// Lookup connection
		Threads::MutexVar cmv(m_connectionsMtx);
		std::vector<Connection*>::iterator cit = getConnection(connID);
		Connection* conn = *cit;

		try {
			// Flush connection
			conn->flush(retry);
		}
		catch(CommsError& e) {
			std::stringstream wrn;
			wrn << "Error flushing connection: " << e.what();
			g_pLogger->warn(wrn.str());
		}
	}
	catch(LookupManagementError& e) {
		std::stringstream wrn;
		wrn << "Error flushing connection: " << e.what();
		g_pLogger->warn(wrn.str());
	}
}


// Notification to decrease tunnels leaky error counts
void Tunneler::decreaseTunnelsErrorCount(void)
{
	Threads::MutexVar tmv(m_tunnelsMtx);

	#ifdef DEBUG_TUN
	if(m_tunnels.size() > 0)
		DPRINTV("Tunneler::decreaseTunnelsErrorCount()");
	#endif

	// Iterate over all tunnels
	std::vector<Tunnel*>::iterator tit;
	for(tit = m_tunnels.begin(); tit != m_tunnels.end(); ++tit) {
		// Decrease error count
		Tunnel* tun = *tit;
		tun->decreaseErrorCount();
	}
	tmv.unlock();

	// Schedule next error leak event
	scheduleEvent(new LeakTunnelErrorEvent(LEAK_ERROR_PERIOD, this));
}


// Notification that we should perform connection manangement
void Tunneler::performConnectionManagement(void)
{
	Threads::MutexVar ccmv(m_closingConnectionsMtx);

	#ifdef DEBUG_TUN
	if(m_closingConnections.size() > 0) {
		std::stringstream dbg;
		dbg << "Tunneler::performConnectionManagement() - Managing ";
		dbg << m_closingConnections.size() << " connections";
		DPRINTV(dbg.str());
	}
	#endif

	// Delete any connections that can be shutdown
	std::transform(m_closingConnections.begin(), m_closingConnections.end(),
		m_closingConnections.begin(), PredConnShutdown(m_connectionMgr));
	std::vector<Connection*>::iterator cit = std::remove_if(m_closingConnections.begin(),
		m_closingConnections.end(), PredPtrNull<Connection>());
	m_closingConnections.erase(cit, m_closingConnections.end());

	// Schedule next event
	scheduleEvent(new ConnectionManagementEvent(CONN_MANAGE_PERIOD, this));
}


// Notification that we should perform tunnel manangement
void Tunneler::performTunnelManagement(void)
{
	Threads::MutexVar tmv(m_tunnelsMtx);
	Threads::MutexVar ctmv(m_closingTunnelsMtx);

	#ifdef DEBUG_TUN
	if(m_closingTunnels.size() > 0) {
		std::stringstream dbg;
		dbg << "Tunneler::performTunnelManagement() - Managing ";
		dbg << m_closingTunnels.size() << " tunnels";
		DPRINTV(dbg.str());
	}
	#endif

	// Find all tunnels from closing list that can be shutdown
	std::vector<Tunnel*> sdTunnels;
	std::vector<Tunnel*>::iterator ctit;
	for(ctit = m_closingTunnels.begin(); ctit != m_closingTunnels.end(); ++ctit) {
		// Can this tunnel be shutdown now?
		Tunnel* tun = *ctit;
		if(tun->canBeShutdown())
			sdTunnels.push_back(tun);
	}

	// Destroy all tunnels that can be shutdown
	for(ctit = sdTunnels.begin(); ctit != sdTunnels.end(); ++ctit) {
		try {
			destroyTunnel(*ctit);
		}
		catch(LookupManagementError& e) {
			g_pLogger->warn("Error destroying tunnel: " + e.what());
		}
	}

	// Schedule next event
	scheduleEvent(new TunnelManagementEvent(TUN_MANAGE_PERIOD, this));
}


// Perform some audit
void Tunneler::audit(void)
{
	DPRINTV("Tunneler::audit()");

	std::stringstream str;

	// Currently active connections
	Threads::MutexVar acmv(m_connectionsMtx);
	int acCount = m_connections.size();
	str << "Active connection count: " << acCount;
	DPRINTV(str.str());
	str.str("");
	if(acCount > 0) {
		for(int i = 0; i < acCount; ++i) {
			str << "\tConnection id " << m_connections[i]->getID();
			DPRINTV(str.str());
			str.str("");
		}
	}
	acmv.unlock();

	// Closing connections
	Threads::MutexVar ccmv(m_closingConnectionsMtx);
	int ccCount = m_closingConnections.size();
	str << "Closing connection count: " << ccCount;
	DPRINTV(str.str());
	str.str("");
	if(ccCount > 0) {
		for(int i = 0; i < ccCount; ++i) {
			str << "\tConnection id " << m_closingConnections[i]->getID();
			DPRINTV(str.str());
			str.str("");
		}
	}
	ccmv.unlock();

	// Currently active tunnels
	Threads::MutexVar tmv(m_tunnelsMtx);
	int tCount = m_tunnels.size();
	str << "Active tunnels count: " << tCount;
	DPRINTV(str.str());
	str.str("");
	if(tCount > 0) {
		for(int i = 0; i < tCount; ++i) {
			str << "\tTunnel id " << m_tunnels[i]->getTidStr();
			DPRINTV(str.str());
			str.str("");
		}
	}
	tmv.unlock();

	// Closing tunnels
	Threads::MutexVar ctmv(m_closingTunnelsMtx);
	int ctCount = m_closingTunnels.size();
	str << "Closing tunnels count: " << ctCount;
	DPRINTV(str.str());
	str.str("");
	if(ctCount > 0) {
		for(int i = 0; i < ctCount; ++i) {
			str << "\tTunnel id " << m_closingTunnels[i]->getTidStr();
			DPRINTV(str.str());
			str.str("");
		}
	}
	ctmv.unlock();

	// ConnectionManager audit
	acmv.lock();
	ccmv.lock();
	m_connectionMgr->audit();
	ccmv.unlock();
	acmv.unlock();

	// Schedule next audit
	scheduleEvent(new AuditEvent(AUDIT_PERIOD, this));
}


/***********************/
/* Protected Interface */
/***********************/


/**************/
/* Predicates */
/**************/


// Find connections that can be shutdown and delete them
Connection* Tunneler::PredConnShutdown::operator()(Connection* conn) const
{
	Connection* res = conn;
	bool shutdown = false;

	try {
		// Ask connection
		shutdown = conn->canBeShutdown();
	}
	catch(CommsError& e) {
		// If we can't even check this connection, go ahead & kill it
		g_pLogger->warn("Error checking connection shutdown state: " + e.what());
		shutdown = true;
	}

	if(shutdown) {
		// Let connection manager know
		m_mgr->delConnection(conn);

		// Delete connection
		delete conn;
		res = 0;
	}

	return res;
}


/***********/
/* Methods */
/***********/


// Increment tunnels created count and return a new random ID
unsigned int Tunneler::incTunCount(void)
{
	Threads::MutexVar mv(c_tunCountMtx);
	c_tunCount += 1;
	return c_tunCount;
}


// Add a new managed connection
void Tunneler::addConnection(Connection* conn)
{
	// Maintain an ordered list
	std::vector<Connection*>::iterator pos;
	Threads::MutexVar cmv(m_connectionsMtx);
	pos = std::lower_bound(m_connections.begin(), m_connections.end(),
		conn, Connection::PredID());
	m_connections.insert(pos, conn);

	// Let connection manager know
	m_connectionMgr->addConnection(conn);
}


// Destroy a managed connection - Caller holds mutex
void Tunneler::destroyConnection(std::vector<Connection*>::iterator& cit)
{
	// Get connection
	Connection* conn = *cit;

	#ifdef DEBUG_TUN
	std::stringstream dbg;
	dbg << "Tunneler::destroyConnection() - Destroying connection id ";
	dbg << conn->getID();
	DPRINTV(dbg.str());
	#endif	

	// Let connection manager know
	m_connectionMgr->delConnection(conn);

	// Remove from list
	m_connections.erase(cit);

	// Cancel any events that might still be pending for this connection
	cancelEvents<ConnectionErrorEvent, unsigned int>(conn->getID());
	cancelEvents<SendMsgEvent, unsigned int>(conn->getID());

	#ifdef DEBUG_EVENTS
	std::stringstream nfo;
	nfo << "Cancelled any events for connection id " << conn->getID();
	nfo << " because tunneler destroying connection";
	g_pLogger->info("**********");
	g_pLogger->info(nfo.str());
	g_pLogger->info("**********");
	#endif

	// Now destroy connection itself
	delete conn;
}


// Make a managed connection a candidate for shutting down - Caller holds mutex
void Tunneler::connectionClosing(std::vector<Connection*>::iterator& cit)
{
	// Get connection
	Connection* conn = *cit;

	#ifdef DEBUG_TUN
	std::stringstream dbg;
	dbg << "Tunneler::connectionClosing() - Moving connection id ";
	dbg << conn->getID() << " to closing list";
	DPRINTV(dbg.str());
	#endif	

	// Remove from list
	m_connections.erase(cit);

	// Cancel any events that might still be pending for this connection
	cancelEvents<ConnectionErrorEvent, unsigned int>(conn->getID());
	cancelEvents<SendMsgEvent, unsigned int>(conn->getID());

	// Let connection know it is now inactive
	conn->setInactive();

	// Place connection on closing list
	Threads::MutexVar ccmv(m_closingConnectionsMtx);
	m_closingConnections.push_back(conn);
}


// Locate a managed connection with some ID - Caller holds Mutex
std::vector<Connection*>::iterator Tunneler::getConnection(unsigned int id)
	throw(LookupManagementError)
{
	// Find connection in managed list
	std::vector<Connection*>::iterator first = m_connections.begin();
	std::vector<Connection*>::iterator cit;

	int top = m_connections.size() - 1;
	int bottom = 0;
	bool found = false;

	while(!found && top >= bottom) {
		int mid = (top + bottom) / 2;
		cit = first + mid;
		unsigned int connID = (*cit)->getID();
		if(id == connID) {
			found = true;
		}
		else if(id < connID) {
			top = mid - 1;
		}
		else {
			bottom = mid + 1;
		}
	}

	// Found it?
	if(!found) {
		std::stringstream err;
		err << "Unknown connection " << id;
		throw LookupManagementError(err.str());
	}

	return cit;
}


// Destroy a managed tunnel - Caller holds tunnels mutex
void Tunneler::destroyTunnel(Tunnel* tun) throw(LookupManagementError)
{
	if(tun) {
		// Get tunnel ID
		TunnelID tid = *tun->getConfig()->getTunnelID();
		DPRINTV("Destroying tunnel " + tid.toString());

		// Find tunnel in managed list
		std::vector<Tunnel*>::iterator tit;
		tit = std::find(m_tunnels.begin(), m_tunnels.end(), tun);
		if(tit != m_tunnels.end()) {
			// Remove from list
			m_tunnels.erase(tit);
		}
		else {
			// Not found
			std::stringstream err;
			err << "Tunneler::destroyTunnel() - Unknown tunnel " << tid;
			throw LookupManagementError(err.str());
		}

		// Cancel any events that might still be pending for this tunnel
		cancelEvents<PollTimeoutEvent, TunnelID>(tid);
		cancelEvents<TunnelStateEvent, TunnelID>(tid);
		cancelEvents<ShutdownTunnelEvent, TunnelID>(tid);
		cancelEvents<TunnelUnviableEvent, TunnelID>(tid);
		cancelEvents<UDPSilentEvent, TunnelID>(tid);

		// Remove tunnel ID from closing list if it's there
		std::vector<Tunnel*>::iterator ctit;
		Threads::MutexVar ctmv(m_closingTunnelsMtx);
		ctit = std::find(m_closingTunnels.begin(), m_closingTunnels.end(), tun);
		if(ctit != m_closingTunnels.end()) {
			m_closingTunnels.erase(ctit);
		}

		#ifdef DEBUG_EVENTS
		std::stringstream nfo;
		nfo << "Cancelled any ShutdownTunnelEvent for tunnel id ";
		nfo << tun->getTidStr();
		nfo << " because tunneler destroying tunnel";
		g_pLogger->info("##########");
		g_pLogger->info(nfo.str());
		g_pLogger->info("##########");
		#endif

		// Now destroy tunnel itself
		delete tun;

		// Notify change in number of tunnels
		noTunnelsChanged();
	}
	else {
		throw LookupManagementError("Tunneler::destroyTunnel() - Null tunnel");
	}
}


// Tag a tunnel as closing - Caller holds tunnels mutex
void Tunneler::tunnelClosing(Tunnel* tun) throw(LookupManagementError)
{
	if(tun) {
		DPRINTV("Tunneler::tunnelClosing() - " + tun->getTidStr());

		// Find tunnel in managed list
		std::vector<Tunnel*>::iterator tit;
		tit = std::find(m_tunnels.begin(), m_tunnels.end(), tun);
		if(tit != m_tunnels.end()) {
			// Check we don't already know this tunnel is closing
			std::vector<Tunnel*>::iterator ctit;
			Threads::MutexVar ctmv(m_closingTunnelsMtx);
			ctit = std::find(m_closingTunnels.begin(), m_closingTunnels.end(), tun);
			if(ctit == m_closingTunnels.end()) {
				// Copy tunnel to closing list
				m_closingTunnels.push_back(tun);

				// Tell tunnel it is closing
				tun->closing();
			}
		}
		else {
			// Not found
			std::stringstream err;
			err << "Tunneler::tunnelClosing() - Unknown tunnel " << tun->getTidStr();
			throw LookupManagementError(err.str());
		}
	}
	else {
		throw LookupManagementError("Tunneler::tunnelClosing() - Null tunnel");
	}
}


}	// namespace

