/* (C) 2005 Richard Mills / The Linux Academy Ltd.
 *
 * Base class for client and server tunnel controllers.
 *
 * $Id: Tunneler.hh,v 1.64 2007-07-08 11:36:51 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#ifndef _TUNNELER_HH_
#define _TUNNELER_HH_

#include <vector>
#include "EventReceiver.hh"
#include "TunnelID.hh"
#include "Mutex.hh"
#include "Exceptions.hh"
#include "Tunnel.hh"
#include "Connection.hh"
#include "TunnelConfig.hh"


namespace HTTPTunnel {

class Server;
class Payload;
class RawBuff;
class HTTPMsg;
class ConnectResultEvent;
class ConnectionErrorEvent;
class ConnectionManager;
class ClientSideConfig;
class ServerSideConfig;
class Sock;


class Tunneler : public EventReceiver {
public:
	// Construction/destruction
	Tunneler(Server* server);
	virtual ~Tunneler();


	/*************/
	/* Accessors */
	/*************/

	// Get owning server
	const Server* getServer(void) const;
	Server* getServerMutable(void);

	// Get owning server configuration
	const ClientSideConfig* getClientSideConfig(void) const throw(ConfigError);
	const ServerSideConfig* getServerSideConfig(void) const throw(ConfigError);
	ClientSideConfig* getClientSideConfigMutable(void) throw(ConfigError);
	ServerSideConfig* getServerSideConfigMutable(void) throw(ConfigError);

	// Get number of active managed tunnels
	unsigned int getNoActiveTunnels(void);


	/**********/
	/* Events */
	/**********/

	// Notification of a new socket connection
	virtual void newConnectionArrived(Sock* insock) throw(CommsError) = 0;

	// Notification that a tunnel has insufficient left/right connections to be viable
	virtual void unviableTunnel(const TunnelID& tid) = 0;

	// Notification of new HTTP message arrival
	virtual void messageArrived(unsigned int connID, const HTTPMsg* msg)
		throw(LookupManagementError) = 0;

	// Notification of new tunneled/raw data arrival
	virtual void rawDataArrived(unsigned int connID, const TunnelID& tid,
		const RawBuff& raw) throw(LookupManagementError) = 0;

	// Check we have received an ack for some sent message
	virtual void checkAckReceived(void) = 0;


	// Notification of a new UDP 'connection'
	virtual void newUDPConnectionArrived(Sock* insock) throw(CommsError);

	// Notification that a tunnel needs to be cleared down
	virtual void shutdownTunnel(const TunnelID& tid, bool otherEndInitiated);

	// Notification of a poll timeout on a tunnel
	virtual void pollTimeout(const TunnelID& tid);

	// Check UDP datagram timeout
	virtual void checkDatagramTimeout(const TunnelID& tid);

	// Notification that connection has changed state
	virtual void connectionStateChange(unsigned int connID,
		Connection::ConnectionState state);

	// Notification that a tunnel has changed state
	virtual void tunnelStateChanged(const TunnelID& tid, Tunnel::TunnelState state);

	// Notification of a socket connect result
	virtual void connectResult(ConnectResultEvent* result);

	// Notification of a fatal error on a socket connection
	virtual void connectionError(ConnectionErrorEvent* error);

	// Notification that a connection needs to be cleared down
	virtual void shutdownConnection(unsigned int connID);

	// Notification we should force a send of a message down a tunnel/connection
	virtual void sendTunnelMsg(const TunnelID& tid, unsigned int connID,
		bool retry, bool mgmt);

	// Notification we should try to flush a connection
	virtual void flushConnection(unsigned int connID, bool retry);

	// Notification to decrease tunnels leaky error counts
	virtual void decreaseTunnelsErrorCount(void);

	// Notification that we should perform connection manangement
	virtual void performConnectionManagement(void);

	// Notification that we should perform tunnel manangement
	virtual void performTunnelManagement(void);

	// Perform some audit
	virtual void audit(void);


protected:

	/*************/
	/* Constants */
	/*************/

	// Number of connect attempts before giving up
	static const unsigned int				MAX_CONNECT_ATTEMPTS		= 3;

	// Time between reconnection attempts
	static const unsigned int				RECONNECT_DELAY			= 5;

	// Time between acknowledgement checks
	static const unsigned int				ACK_CHECK_PERIOD			= 1;

	// Time between management events
	static const unsigned int				CONN_MANAGE_PERIOD		= 1;
	static const unsigned int				TUN_MANAGE_PERIOD			= 2;

	// Period in seconds between leaky bucket error count decreases
	static const unsigned int				LEAK_ERROR_PERIOD			= 2;

	// Time between audits
	static const unsigned int				AUDIT_PERIOD				= 300;


	/***********/
	/* Methods */
	/***********/

   // The number of tunnels in existance has changed
   virtual void noTunnelsChanged(void) = 0;

	// Increment tunnels created count and return a new random ID
	unsigned int incTunCount(void);

	// Connection management
	void addConnection(Connection* conn);
	void destroyConnection(std::vector<Connection*>::iterator& cit);
	void connectionClosing(std::vector<Connection*>::iterator& cit);
	std::vector<Connection*>::iterator getConnection(unsigned int id)
		throw(LookupManagementError);

	// Tunnel management
	void destroyTunnel(Tunnel* tun) throw(LookupManagementError);
	void tunnelClosing(Tunnel* tun) throw(LookupManagementError);

	// Locate a tunnel matching some predicate
	template<class P> Tunnel* getTunFromPred(P pred);
	template<class P> void getAllTunsFromPred(P pred, std::vector<Tunnel*>& res);


	/**************/
	/* Predicates */
	/**************/

	// Match tunnel from an ID
	struct PredTunID {
		PredTunID(const TunnelID* tid) : m_tid(tid) {}
		bool operator()(const Tunnel* tun) const {
			return (*(tun->getConfig()->getTunnelID()) == *m_tid);
		}
		const TunnelID* m_tid;
	};

	// Match tunnel from a connection
	struct PredTunConnection {
		PredTunConnection(Connection* conn) : m_conn(conn) {}
		bool operator()(Tunnel* tun) const {
			return (tun->ownsConnection(m_conn));
		}
		Connection* m_conn;
	};

	// Match connections that can be shutdown
	struct PredConnShutdown {
		PredConnShutdown(ConnectionManager* mgr) : m_mgr(mgr) {}
		Connection* operator()(Connection* conn) const;
		ConnectionManager* m_mgr;
	};

	// Match null pointers
	template<class T> struct PredPtrNull {
		bool operator()(T* ptr) const { return (ptr == 0); }
	};


	/********/
	/* Data */
	/********/

	// A mutex
	Threads::Mutex									m_tunMtx;

	// Our owning server
	Server*											m_server;

	// All the currently active tunnels
	std::vector<Tunnel*>							m_tunnels;
	Threads::Mutex									m_tunnelsMtx;

	// Tunnels we are trying to shutdown
	std::vector<Tunnel*>							m_closingTunnels;
	Threads::Mutex									m_closingTunnelsMtx;

	// All active connections
	std::vector<Connection*>					m_connections;
	Threads::Mutex									m_connectionsMtx;

	// Connections we are trying to shutdown
	std::vector<Connection*>					m_closingConnections;
	Threads::Mutex									m_closingConnectionsMtx;

	// A connection manager
	ConnectionManager*							m_connectionMgr;


private:

	/********/
	/* Data */
	/********/

	// A count of all tunnels created
	static unsigned int							c_tunCount;
	static Threads::Mutex						c_tunCountMtx;

};


}	// namespace


// Include template methods
#include "TunnelerT.cc"


#endif

