/* (C) 2006 Richard Mills / The Linux Academy Ltd.
 *
 * Useful utility functions.
 *
 * $Id: Utils.cc,v 1.15 2007-07-07 12:59:39 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <netdb.h>
#include <arpa/inet.h>
#include <iostream>
#include "Utils.hh"
#include "MutexVar.hh"


namespace HTTPTunnel {


// Initialise static data
char Utils::c_base64Table[64] = {
	'A','B','C','D','E','F','G','H','I','J','K','L','M','N','O','P','Q',
	'R','S','T','U','V','W','X','Y','Z','a','b','c','d','e','f','g','h',
	'i','j','k','l','m','n','o','p','q','r','s','t','u','v','w','x','y',
	'z','0','1','2','3','4','5','6','7','8','9','+','/'
};

char Utils::c_modBase64Table[64] = {
	'A','B','C','D','E','F','G','H','I','J','K','L','M','N','O','P','Q',
	'R','S','T','U','V','W','X','Y','Z','a','b','c','d','e','f','g','h',
	'i','j','k','l','m','n','o','p','q','r','s','t','u','v','w','x','y',
	'z','0','1','2','3','4','5','6','7','8','9','-','.'
};

Threads::Mutex Utils::c_mtx;


/********************/
/* Public Interface */
/********************/


// Split a string into component parts
void Utils::SplitString(const std::string str, const std::string sep,
	std::vector<std::string>& res)
{
	res.clear();

	std::string::size_type lastsep = 0, nextsep = 0;
	while((nextsep = str.find(sep, nextsep)) != std::string::npos) {
		res.push_back(std::string(str, lastsep, nextsep - lastsep));
		nextsep++;
		lastsep = nextsep;
	}

	res.push_back(std::string(str, lastsep, nextsep - lastsep));
}


// Extract the value of a parm="value" pair
bool Utils::getParmValue(const std::vector<std::string>& args,
	const char* parm, std::string& res)
{
	bool ok = false;
	std::string reqParm(parm);

	for(unsigned int i = 0; i < args.size(); ++i) {
		std::string argParm = args[i];
		if(argParm.compare(0, reqParm.size(), reqParm) == 0) {
			// Split into id=value
			std::vector<std::string> idvalue;
			SplitString(argParm, "=", idvalue);
			if(idvalue.size() == 2) {
				res = idvalue[1];
				ok = true;

				// Strip any quotes if found
				if(res[0] == '"' && res[res.size() - 1] == '"')
					res = res.substr(1, res.size() - 2);

				break;
			}
		}
	}

	return ok;
}


// UUEncode some string
std::string Utils::UUEncode(const std::string& src)
{
	int length = src.size();
	char uue[length * 2];
	unsigned int ui = 0;

	for(int i = 0; i < length; i += 3, ui += 4) {
		if(i < length - 2) {
			// Got 3 bytes to encode
			register char a = src[i];
			register char b = src[i + 1];
			register char c = src[i + 2];
			uue[ui] = ((a >> 2) & 0x3F) + 35;
			uue[ui + 1] = (((a << 4) | ((b >> 4) & 0x0F)) & 0x3F) + 35;
			uue[ui + 2] = (((b << 2) | ((c >> 6) & 0x03)) & 0x3F) + 35;
			uue[ui + 3] = (c & 0x3F) + 35;
		}
		else if(i < length - 1) {
			// Only got 2 bytes to encode
			register char a = src[i];
			register char b = src[i + 1];
			uue[ui] = ((a >> 2) & 0x3F) + 35;
			uue[ui + 1] = (((a << 4) | ((b >> 4) & 0x0F)) & 0x3F) + 35;
			uue[ui + 2] = ((b << 2) & 0x3F) + 35;
			uue[ui + 3] = '~';
		}
		else {
			// Only got 1 byte to encode
			register char a = src[i];
			uue[ui] = ((a >> 2) & 0x3F) + 35;
			uue[ui + 1] = ((a << 4) & 0x3F) + 35;
			uue[ui + 2] = '~';
			uue[ui + 3] = '~';
		}
	}

	// Construct string
	uue[ui] = '\0';
	return std::string(uue);
}


// UUDecode some string
RawBuff Utils::UUDecode(const std::string& src)
{
	const char* uuestart = src.c_str();
	const char* uueend = uuestart + src.size();
	unsigned char uud[src.size()];
	unsigned int uudlength = 0;

	for(const char* uue = uuestart; uue < uueend; uue += 4) {
		// Allways need first 2 chars
		register char a = uue[0] - 35;
		register char b = uue[1] - 35;
		uud[uudlength++] = (a << 2) | (b >> 4);

		if(uue[2] != '~') {
			// Got at least another byte of encoded data
			register char c = uue[2] - 35;
			uud[uudlength++] = ((b & 0x0F) << 4) | ((c & 0x3C) >> 2);

			if(uue[3] != '~') {
				// Got the full 3 bytes of encoded data
				register char d = uue[3] - 35;
				uud[uudlength++] = ((c & 0x03) << 6) | d;
			}
		}
	}

	// Construct RawBuff
	return RawBuff(uud, uudlength);
}


// Base64 encode some string
std::string Utils::Base64Encode(const std::string& src)
{
	int length = src.size();
	char buff[length * 2];
	unsigned int bi = 0;

	for(int i = 0; i < length; i += 3, bi += 4) {
		if(i < length - 2) {
			// Got 3 bytes to encode
			register char a = src[i];
			register char b = src[i + 1];
			register char c = src[i + 2];
			buff[bi] = c_base64Table[((a >> 2) & 0x3F)];
			buff[bi + 1] = c_base64Table[(((a << 4) | ((b >> 4) & 0x0F)) & 0x3F)];
			buff[bi + 2] = c_base64Table[(((b << 2) | ((c >> 6) & 0x03)) & 0x3F)];
			buff[bi + 3] = c_base64Table[(c & 0x3F)];
		}
		else if(i < length - 1) {
			// Only got 2 bytes to encode
			register char a = src[i];
			register char b = src[i + 1];
			buff[bi] = c_base64Table[((a >> 2) & 0x3F)];
			buff[bi + 1] = c_base64Table[(((a << 4) | ((b >> 4) & 0x0F)) & 0x3F)];
			buff[bi + 2] = c_base64Table[((b << 2) & 0x3F)];
			buff[bi + 3] = '=';
		}
		else {
			// Only got 1 byte to encode
			register char a = src[i];
			buff[bi] = c_base64Table[((a >> 2) & 0x3F)];
			buff[bi + 1] = c_base64Table[((a << 4) & 0x3F)];
			buff[bi + 2] = '=';
			buff[bi + 3] = '=';
		}
	}

	// Construct string
	buff[bi] = '\0';
	return std::string(buff);
}


// Base64 decode some string
std::string Utils::Base64Decode(const std::string& src)
{
	const char* start = src.c_str();
	const char* end = start + src.size();
	char decoded[src.size()];
	unsigned int dlength = 0;

	for(const char* enc = start; enc < end; enc += 4) {
		// Allways need first 2 chars
		register char a = base64DecodeChar(enc[0]);
		register char b = base64DecodeChar(enc[1]);
		decoded[dlength++] = (a << 2) | (b >> 4);

		if(enc[2] != '=') {
			// Got at least another byte of encoded data
			register char c = base64DecodeChar(enc[2]);
			decoded[dlength++] = ((b & 0x0F) << 4) | ((c & 0x3C) >> 2);

			if(enc[3] != '=') {
				// Got the full 3 bytes of encoded data
				register char d = base64DecodeChar(enc[3]);
				decoded[dlength++] = ((c & 0x03) << 6) | d;
			}
		}
	}

	// Construct string
	decoded[dlength] = '\0';
	return std::string(decoded);
}


// Modified base64 encode some string
std::string Utils::ModBase64Encode(const std::string& src)
{
	char buff[src.size() * 2];
	unsigned int bi = 0;
	int length = src.size();

	for(int i = 0; i < length; i += 3, bi += 4) {
		if(i < length - 2) {
			// Got 3 bytes to encode
			register char a = src[i];
			register char b = src[i + 1];
			register char c = src[i + 2];
			buff[bi] = c_modBase64Table[((a >> 2) & 0x3F)];
			buff[bi + 1] = c_modBase64Table[(((a << 4) | ((b >> 4) & 0x0F)) & 0x3F)];
			buff[bi + 2] = c_modBase64Table[(((b << 2) | ((c >> 6) & 0x03)) & 0x3F)];
			buff[bi + 3] = c_modBase64Table[(c & 0x3F)];
		}
		else if(i < length - 1) {
			// Only got 2 bytes to encode
			register char a = src[i];
			register char b = src[i + 1];
			buff[bi] = c_modBase64Table[((a >> 2) & 0x3F)];
			buff[bi + 1] = c_modBase64Table[(((a << 4) | ((b >> 4) & 0x0F)) & 0x3F)];
			buff[bi + 2] = c_modBase64Table[((b << 2) & 0x3F)];
			buff[bi + 3] = '*';
		}
		else {
			// Only got 1 byte to encode
			register char a = src[i];
			buff[bi] = c_modBase64Table[((a >> 2) & 0x3F)];
			buff[bi + 1] = c_modBase64Table[((a << 4) & 0x3F)];
			buff[bi + 2] = '*';
			buff[bi + 3] = '*';
		}
	}

	// Construct string
	buff[bi] = '\0';
	return std::string(buff);
}


// Modified base64 decode some string
RawBuff Utils::ModBase64Decode(const std::string& src)
{
	const char* start = src.c_str();
	const char* end = start + src.size();
	unsigned char decoded[src.size()];
	unsigned int dlength = 0;

	for(const char* enc = start; enc < end; enc += 4) {
		// Allways need first 2 chars
		register char a = modBase64DecodeChar(enc[0]);
		register char b = modBase64DecodeChar(enc[1]);
		decoded[dlength++] = (a << 2) | (b >> 4);

		if(enc[2] != '*') {
			// Got at least another byte of encoded data
			register char c = modBase64DecodeChar(enc[2]);
			decoded[dlength++] = ((b & 0x0F) << 4) | ((c & 0x3C) >> 2);

			if(enc[3] != '*') {
				// Got the full 3 bytes of encoded data
				register char d = modBase64DecodeChar(enc[3]);
				decoded[dlength++] = ((c & 0x03) << 6) | d;
			}
		}
	}

	// Construct RawBuff
	return RawBuff(decoded, dlength);
}


// Resolve an address to a string IP address
std::string Utils::resolveIP(std::string addr)
{
	std::string ip;
	struct hostent host;
	char auxHostBuff[AUX_HOST_BUFF_SIZE];

	if(getHostByName(addr, &host, auxHostBuff, sizeof(auxHostBuff)) == 0) {
		// Turn IP address back into string
		char address[INET_ADDRSTRLEN];
		inet_ntop(AF_INET, host.h_addr_list[0], address, sizeof(address));
		ip = address;
	}

	return ip;
}

// Resolve an address to a host entry
int Utils::getHostByName(std::string addr, struct hostent* host,
	char* auxBuff, size_t auxBuffLen)
{
	#if !defined(PLATFORM_OSX) && !defined(PLATFORM_CYGWIN)
	int gethosterrno = 0;
	#endif

	#ifdef PLATFORM_SOLARIS
	struct hostent* gethostres = gethostbyname_r(addr.c_str(), host,
		auxBuff, auxBuffLen, &gethosterrno);
	return (gethostres == 0)? -1 : 0;

	#elif defined(PLATFORM_OSX) || defined(PLATFORM_CYGWIN)
	Threads::MutexVar mv(c_mtx);
	struct hostent *gethostres = gethostbyname(addr.c_str());
	if(gethostres)
		*host = *gethostres;
	return (gethostres == 0)? -1 : 0;

	#else
	struct hostent* phost = 0;
	int gethostres = gethostbyname_r(addr.c_str(), host, auxBuff,
		auxBuffLen, &phost, &gethosterrno);
	return (gethostres == 0)? 0 : -1;
	#endif
}


/*********************/
/* Private Interface */
/*********************/


// Decode a single base64 byte
char Utils::base64DecodeChar(char enc)
{
	// Find char in encoding array
	char* pos = (char*)memchr(c_base64Table, enc, sizeof(c_base64Table));
	return (char)(pos - c_base64Table);
}

// Decode a single modified base64 byte
char Utils::modBase64DecodeChar(char enc)
{
	// Find char in encoding array
	char* pos = (char*)memchr(c_modBase64Table, enc, sizeof(c_modBase64Table));
	return (char)(pos - c_modBase64Table);
}


}	// namespace


